#ifndef PLUGINHANDLER_H
#define PLUGINHANDLER_H

#include "PapayaList.h"
#include "Plugin.h"

#include <gmodule.h>

typedef PapayaList List;
typedef PapayaListElement ListElement;

struct handler_entry {
  struct handler_entry * next;
  GModule * handler;
};

class PluginData;

class PluginHandler {

 public:

  PluginHandler();
  ~PluginHandler();

  void unloadPlugins();
  void loadPlugins();
  void loadFile(PluginData *);
  void unloadFile(PluginData *);

#ifdef WIN32
  void win32();
#endif

  void registerPlugin(Plugin *, char *);
  void unregisterPlugin(Plugin *);
  void addHandlerEntry(GModule *);

  GModule * findPlugin(char *);
  void * findPlugin(char *, char *);

  void displayPlugins();

  // @@  GtkWidget * pluginList;

  void saveConfigFile();

  /* Shouldn't need to change these. */

  void inputFilters(Connection *, char *);
  void outputFilters(Connection *, char *);
  void promptFilters(Connection *, char *);
  void clientMessage(Connection *, char *);
  void pageSwitched();

  void fireEvent(Event *, Connection *);
  void promptExpose(GtkWidget *, GdkEventExpose *, gpointer);

  void updateAvailablePlugins();
  void displayAvailablePlugins();

  void loadConfigFile();
  void checkPlugins();
  void loadPluginData(); // Loads plugin data from the .so files.

  void getPrefsWidgets(GtkNotebook *, MUD *);
  void onPrefsApply(MUD *);
  void onPrefsOk(MUD *);
  void onPrefsCancel(MUD *);
  void savePrefs(FILE *, MUD *);

#ifdef USE_GTK_2
  void toggleEnabled(char * name);
#else
  void selection(GtkWidget *, gint, gint, GdkEventButton *);
#endif

 private:

  PapayaList * pluginList; // The currently loaded set of plugins.
  PapayaList * availablePlugins; // The available set of plugins.

  struct handler_entry * first_handler;
  GtkWidget * create_plugin_list();
#ifdef USE_GTK_2
  void populate_plugin_list();
#else
  void populate_plugin_clist(GtkWidget *);
#endif

  void checkToLoad(PluginData *, bool);
  void checkDirForPlugins(char *);
  void loadEnabledPlugins();

  PluginData * queryPlugin(char * file, char * path);

  char module_message[16384];

  /*
  GtkWidget * pluginList_clist;
  void create_pluginList();
  */
};

#endif
