#ifndef INCLUDE_PREFS_H
#define INCLUDE_PREFS_H

#include <stdio.h>
#include <gdk/gdk.h>

#define DEFAULT_FONT "-*-courier-medium-r-normal-*-*-140-*-*-*-*-*-*"

class SystemColour;
class MUD;

struct pref_data {
  char * name;
  char * data;
  struct pref_data * next;
};

class Prefs {

 public:

  Prefs(MUD *);
  ~Prefs();

  char * getBrowser();
  void setBrowser(char *);

  void setNegotiateCharmode(bool);
  bool getNegotiateCharmode();

  void setMultiLinePrompt(int);
  int getMultiLinePrompt();

  void setEcho(bool);
  bool getEcho();
  
  void setBeep(bool);
  bool getBeep();

  void setPromptOnInputLine(bool);
  bool getPromptOnInputLine();

  void setCompression(bool);
  bool getCompression();

  void setCompressionOffCommand(char *);
  char * getCompressionOffCommand();

  void setSocks4(bool);
  bool getSocks4();

  void setSocks(bool);
  bool getSocks();

  void setSocksUser(char *);
  char * getSocksUser();

  void setSocksPass(char *);
  char * getSocksPass();

  void setSocksHost(char *);
  char * getSocksHost();

  void setSocksPort(char *);
  char * getSocksPort();

  void setMCW(bool);
  bool getMCW();

  void setMudftp(bool);
  bool getMudftp();

  void setTriggers(bool);
  bool getTriggers();

  void setSystemTriggers(bool);
  bool getSystemTriggers();

  void set(char *, char *);
  bool parseLine(FILE *);

  void setScrollOnInput(bool);
  bool getScrollOnInput();

  void setHistoryUseLimit(bool);
  bool getHistoryUseLimit();

  void setHistoryLimit(int);
  int getHistoryLimit();

  void setHistoryHonourNull(bool);
  bool getHistoryHonourNull();

  void setHistoryIgnoreRepeats(bool);
  bool getHistoryIgnoreRepeats();

  void setHistoryIsCircular(bool);
  bool getHistoryIsCircular();

  void setSemiColon(bool);
  bool getSemiColon();

  void setUsePapayaVT(bool);
  bool getUsePapayaVT();

  void setSendVersion(bool);
  bool getSendVersion();

  void setSendStats(bool);
  bool getSendStats();

  void setColour(bool);
  bool getColour();

  void setDefaultForegroundColour(int);
  int getDefaultForegroundColour();

  void setDefaultBackgroundColour(int);
  int getDefaultBackgroundColour();

  void setColours(char *);
  void setColours(int, int, int, int);
  void getColours(int index, int * r, int * g, int * b);
  char * getColours();
  bool getLoadedColours();
  struct colour_table * getColourTable();
  SystemColour * getSysColour();

  int getScreenBackgroundRed();
  int getScreenBackgroundGreen();
  int getScreenBackgroundBlue();
  void setScreenBackground(int, int, int);
  void setScreenBackground(char *);

  bool load(MUD *);
  bool save(MUD *);
  bool save(char *);

  void setUseMaxLines(bool);
  bool getUseMaxLines();

  void setMaxLines(int);
  int getMaxLines();

  void setInputFont(char *);
  void setOutputFont(char *);
  void setBoldFont(char *);
  void setPromptFont(char *);

  char * getInputFont();
  char * getOutputFont();
  char * getBoldFont();
  char * getPromptFont();

  void setUsePromptTerminator(bool);
  bool getUsePromptTerminator();

  void setPromptTerminator(char *);
  char * getPromptTerminator();

#ifdef USE_GTK_2
  char * getPangoInputFont();
  char * getPangoOutputFont();
  char * getPangoBoldOutputFont();

  void setPangoInputFont(char *);
  void setPangoOutputFont(char *);
  void setPangoBoldOutputFont(char *);
#endif

  bool getUseBoldFont();
  void setUseBoldFont(bool);

  void setAutoReconnect(bool);
  bool getAutoReconnect();

  static void loadFont(char *, GdkFont **);

  bool getTurfHttpd();
  void setTurfHttpd(bool);

  bool getHttpConnect();
  void setHttpConnect(bool);

  bool getHttpProxy();
  void setHttpProxy(bool);

  char * getHttpProxyHost();
  void setHttpProxyHost(char *);
  
  char * getHttpProxyPort();
  void setHttpProxyPort(char *);

  void setFilteredOutput(bool);
  bool getFilteredOutput();

  void setWindowX(int);
  void setWindowY(int);

  int getWindowX();
  int getWindowY();

  bool getSendKeepalive();
  void setSendKeepalive(bool);

  int getKeepaliveTimer();
  void setKeepaliveTimer(int);

  void setUseToolbar(bool);
  bool getUseToolbar();

  GdkFont * getOutput();
  GdkFont * getBold();
  GdkFont * getInput();
  GdkFont * getPrompt();

  void setDoublePrompt(bool);
  bool getDoublePrompt();

  struct pref_data * findPreference(char *);
  void setPreference(char *, char *);
  char * getPreference(char *);
  void saveCustomPrefs(FILE *);

 private:

  struct pref_data * pref_list;

  char * browser;


  bool doublePrompt; // TRUE if prompt echoed to main window as well as snarfed.
  bool promptOnInputLine; // TRUE if prompt and input on same line.  Else FALSE.
  bool papayavt;
  bool autoreconnect;
  bool echo;
  bool beep;
  bool scroll;
  bool history_use_limit;
  int history_limit;
  bool history_honour_null;
  bool history_ignore_repeats;
  bool history_circular;
  bool compression;
  bool negotiate_charmode;
  char * compression_off_command;
  bool socks4;
  bool socks;
  char * socks4_pass;
  char * socks4_user;
  char * socks4_host;
  char * socks4_port;
  bool mcw_start;
  bool mudftp;
  bool triggers;
  bool system_triggers;

  bool colour;
  bool colours_loaded;
  int fg_colour;
  int bg_colour;
  int colours [16][3]; // The colours that syscolour will grab.

  int bg_r;
  int bg_g;
  int bg_b;

  bool send_version;
  bool send_stats;

  bool semicolon;
  bool useMaxLines;
  int maxLines;

  bool toolbar;

  char * input_font;
  char * output_font;
  char * bold_font;
  char * prompt_font;
  bool use_bold_font;

  int multilineprompt;
  SystemColour * colour_table;

  bool http_proxy;
  bool http_connect;
  char * http_proxy_port;
  char * http_proxy_host;

  bool turf_httpd;

  bool filtered;

  bool send_keepalive;
  int keepalive_timer;

  int window_x;
  int window_y;

  GdkFont * inputFont;
  GdkFont * outputFont;
  GdkFont * boldFont;
  GdkFont * promptFont;

  char * promptTerminator;
  bool usePromptTerminator;

};

#endif
