#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <string.h>
#include <stdio.h>

#include "system.h"

#include <gdk/gdkkeysyms.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"

#include "version.h"
#include "About.h"

static GtkWidget * about = (GtkWidget *)NULL;

static void on_about_ok_pressed(GtkButton * button, gpointer user_data) {
  if (about) {
    gtk_widget_hide(about);
    gtk_widget_destroy(about);
    about = (GtkWidget *)NULL;
  }
}

About::About() {
  about = create_about_window();
  gtk_widget_show(about);
}

About::~About() {
  if (about) {
    gtk_widget_hide(about);
    gtk_widget_destroy(about);
    about = (GtkWidget *)NULL;
  }
}

GtkWidget * About::create_about_window (void)
{
  GtkWidget *about_window;
  GtkWidget *vbox1;
  GtkWidget *table1;
  GtkWidget *about_version;
  GtkWidget *scrolledwindow1;
#ifdef USE_GTK
  GtkWidget *about_author;
#endif
  GtkWidget *about_pixmap;
  GtkWidget *hbuttonbox1;
  GtkWidget *about_ok;

  about_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
#ifndef USE_GTK_2
  gtk_object_set_data (GTK_OBJECT (about_window), "about_window", about_window);
#endif
  gtk_window_set_title (GTK_WINDOW (about_window), "About");
  gtk_window_set_modal (GTK_WINDOW (about_window), TRUE);
  gtk_window_set_default_size (GTK_WINDOW (about_window), 474, 360);
#ifndef USE_GTK_2
  gtk_window_set_policy (GTK_WINDOW (about_window), TRUE, TRUE, TRUE);
#endif
  vbox1 = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (vbox1);
#ifndef USE_GTK_2
  gtk_object_set_data_full (GTK_OBJECT (about_window), "vbox1", vbox1,
                            (GtkDestroyNotify) gtk_widget_unref);
#endif
  gtk_widget_show (vbox1);
  gtk_container_add (GTK_CONTAINER (about_window), vbox1);

  table1 = gtk_table_new (3, 1, FALSE);
  gtk_widget_ref (table1);
#ifndef USE_GTK_2
  gtk_object_set_data_full (GTK_OBJECT (about_window), "table1", table1,
                            (GtkDestroyNotify) gtk_widget_unref);
#endif
  gtk_widget_show (table1);
  gtk_box_pack_start (GTK_BOX (vbox1), table1, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (table1), 4);
  gtk_table_set_row_spacings (GTK_TABLE (table1), 4);

  char buf[1024];
  sprintf(buf, "Papaya version %s (build %d)", VERSION, RELEASE);

  about_version = gtk_label_new (buf);
  gtk_widget_ref (about_version);
#ifndef USE_GTK_2
  gtk_object_set_data_full (GTK_OBJECT (about_window), "about_version", about_version,
                            (GtkDestroyNotify) gtk_widget_unref);
#endif
  gtk_widget_show (about_version);
  gtk_table_attach (GTK_TABLE (table1), about_version, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_padding (GTK_MISC (about_version), 0, 7);

  scrolledwindow1 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_ref (scrolledwindow1);
#ifndef USE_GTK_2
  gtk_object_set_data_full (GTK_OBJECT (about_window), "scrolledwindow1", scrolledwindow1,
                            (GtkDestroyNotify) gtk_widget_unref);
#endif
  gtk_widget_show (scrolledwindow1);
  gtk_table_attach (GTK_TABLE (table1), scrolledwindow1, 0, 1, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), 0, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow1), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

#define COPYRIGHT "Copyright 2000-2001 Catherine Allen.\n\nAuthor: Catherine Allen (mango@turf.org)\n\nOther contributors: Chris Allen (fastjack@turf.org)\n                           Oliver Jowett (nemon@abandoned.org).\n                           Noah Gibbs (noah_gibbs@yahoo.com)\n\nBug Reporters: k0n, Mavetju\n"

#ifdef USE_GTK_2
  GtkTextBuffer * about_buffer = gtk_text_buffer_new(NULL);
  GtkWidget * about_view = gtk_text_view_new_with_buffer(about_buffer);

  gtk_widget_show(about_view);
  gtk_container_add(GTK_CONTAINER(scrolledwindow1), about_view);
  
  GtkTextIter start, end;
  gtk_text_buffer_get_bounds(about_buffer, &start, &end);

  gtk_text_buffer_insert(about_buffer, &end, COPYRIGHT, -1);
#else
  about_author = gtk_text_new (NULL, NULL);
  gtk_widget_ref (about_author);
  gtk_object_set_data_full (GTK_OBJECT (about_window), "about_author", about_author,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (about_author);
  gtk_container_add (GTK_CONTAINER (scrolledwindow1), about_author);
  gtk_text_insert (GTK_TEXT (about_author), NULL, NULL, NULL, COPYRIGHT, -1);
#endif


  about_pixmap = create_pixmap (about_window, "papaya.xpm");
  gtk_widget_ref (about_pixmap);
#ifndef USE_GTK_2
  gtk_object_set_data_full (GTK_OBJECT (about_window), "about_pixmap", about_pixmap,
                            (GtkDestroyNotify) gtk_widget_unref);
#endif
  gtk_widget_show (about_pixmap);
  gtk_table_attach (GTK_TABLE (table1), about_pixmap, 0, 1, 0, 1,
                    (GtkAttachOptions) (0),
                    (GtkAttachOptions) (0), 0, 0);

  hbuttonbox1 = gtk_hbutton_box_new ();
  gtk_widget_ref (hbuttonbox1);
#ifndef USE_GTK_2
  gtk_object_set_data_full (GTK_OBJECT (about_window), "hbuttonbox1", hbuttonbox1,
                            (GtkDestroyNotify) gtk_widget_unref);
#endif
  gtk_widget_show (hbuttonbox1);
  gtk_box_pack_start (GTK_BOX (vbox1), hbuttonbox1, FALSE, FALSE, 0);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox1), GTK_BUTTONBOX_END);

  about_ok = gtk_button_new_with_label (_("Ok"));
  gtk_widget_ref (about_ok);
#ifndef USE_GTK_2
  gtk_object_set_data_full (GTK_OBJECT (about_window), "about_ok", about_ok,
                            (GtkDestroyNotify) gtk_widget_unref);
#endif
  gtk_widget_show (about_ok);
  gtk_container_add (GTK_CONTAINER (hbuttonbox1), about_ok);
  GTK_WIDGET_SET_FLAGS (about_ok, GTK_CAN_DEFAULT);

#ifdef USE_GTK_2
  g_signal_connect_data(G_OBJECT(about_ok), "clicked",
			GTK_SIGNAL_FUNC(on_about_ok_pressed), NULL,
			NULL, (GConnectFlags) 0);
#else
  gtk_signal_connect (GTK_OBJECT (about_ok), "clicked",
                      GTK_SIGNAL_FUNC (on_about_ok_pressed),
                      NULL);
#endif
  return about_window;
}

