#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "AliasEntity.h"
#include "EntityHandler.h"

extern EntityHandler * entities;

#define ALIAS_FAIL 0
#define ALIAS_OK 1

AliasEntity::AliasEntity() : Entity() {
  body = NULL;
  command = NULL;
  macro_key = 0;
  setType(EntityAlias);
  alias_name = NULL;
}

AliasEntity::~AliasEntity() {
  if (body)
    free(body);
}

void AliasEntity::setBody(char * b) {
  if (body)
    free(body);
  body = strdup(b);
}

void AliasEntity::setCommand(char * c) {
  if (command)
    free(command);
  command = strdup(c);
}

char * AliasEntity::getCommand() {
  if (command)
    return command;
  else
    return "";
}

char * AliasEntity::getBody() {
  if (body)
    return body;
  return "";
}

void AliasEntity::setMacroKey(int k) {
  macro_key = k;
}

void AliasEntity::setMacroAlt(bool t) {
  macro_alt = t;
}

void AliasEntity::setMacroCtrl(bool t) {
  macro_ctrl = t;
}

bool AliasEntity::getMacroAlt() {
  return macro_alt;
}

bool AliasEntity::getMacroCtrl() {
  return macro_ctrl;
}

int AliasEntity::getMacroKey() {
  return macro_key;
}

bool AliasEntity::findExecute(EntityType t, char * field, char * value, void * data, bool multiple) {

  if (t != getType())
    return false;

  if (disabled)
    return false;

  if (!strcasecmp(field, "key")) {
    int key = *((int *)value);

    if (getMacroKey() != key)
      return false;

    execute(value, (char *)data);
    return true;
  }

  if (!strcasecmp(field, "command")) {
    if (strcmp(value, getCommand())) // no match
      return false;

    execute(value, (char *)data);
    return true;
  }

  return false;
}

static char * findArgument(char * text, int index) {

  int count = 1;
  char * arg = text;

  while (count < index) {
    while (isspace(*arg))
      arg++;
     arg = strchr(arg, ' ');
     if (!arg) {
       return NULL;
     }
     count++;
  }

  while (isspace(*arg))
    arg++;
  return arg;
}


static int append_argument(char * output, int len, int pos, char * input, char i) {
  
  int index;
  char index_str[2];
  int count = 0;

  if (!input)
    return 0;

  index_str[0] = i;
  index_str[1] = '\0';
  index = atoi(index_str);
  
  if (index == 0) {
    if (pos + strlen(input) >= (unsigned int)len)
      return 0;

    memcpy(output + pos, input, strlen(input));
    return strlen(input);
  }

  char * arg = findArgument(input, index);
  if (!arg)
    return 0;

  while (*arg != '\0' && !isspace(*arg)) {
    output[pos + count] = *arg;
    arg++;
    count++;
  }

  return count;
}

static int append_char(char * output, int len, int pos, char c) {

  if (pos >= len) {
    output[len-1] = '\0';
    return ALIAS_FAIL;
  }

  output[pos] = c;
  return ALIAS_OK;
}

void AliasEntity::execute(char * input, char * output) {

  char * pc = getBody();
  int out_pos = 0;
  int output_len = 16384;

  output[0] = '\0';

  if (output_len == 0)
    return;

  while (*pc != '\0' && out_pos < output_len - 1) {
    // Normal character.
    if (*pc != '$') {
      append_char(output, output_len, out_pos, *pc);
      pc++;
      out_pos++;
      continue;
    } else { // $ character.
      if (*(pc+1) == '\0') { // $ EOS
	break;
      }
      else if (*(pc+1) == '$') { // $$
	append_char(output, output_len, out_pos, '$');
	pc += 2;
	out_pos++;
	continue;
      } else { // $[0-9]
	out_pos += append_argument(output, output_len, out_pos, input, *(pc + 1));
	pc += 2;
	continue;
      }
    }
  }
  
  append_char(output, output_len, out_pos, '\0');
}

void AliasEntity::save(FILE * fp) {
  fprintf(fp, "Alias %s\n", getName());
  fprintf(fp, "Description\n");
  fprintf(fp, "%s", getDescription());
  if (getDescription()[strlen(getDescription()) - 1] != '\n')
    fprintf(fp, "\n");
  fprintf(fp, "EndDescription\n");
  fprintf(fp, "Command %s\n", getCommand());
  fprintf(fp, "MacroKey %d\n", getMacroKey());
  fprintf(fp, "Body\n");
  fprintf(fp, "%s", getBody());
  if (getBody()[strlen(getBody()) - 1] != '\n')
    fprintf(fp, "\n");

  fprintf(fp, "EndBody\n");
  fprintf(fp, "EndAlias\n");
}

bool AliasEntity::load(FILE * fp, char * name) {
  char buf[16384];
  char * ptr;

  setName(name);
  
  while (true) {
    if (!fgets(buf, 16384, fp)) {
      perror("fgets");
      return false;
    }

    if (!strncasecmp(buf, "Description", 11)) {
      ptr = loadBlock(fp, "EndDescription");
      if (!ptr)
	return false;
      setDescription(ptr);
      continue;
    }

    if (!strncasecmp(buf, "Body", 4)) {
      ptr = loadBlock(fp, "EndBody");
      if (!ptr)
	return false;
      setBody(ptr);
      continue;
    }

    if (!strncasecmp(buf, "Command", 7)) {
      char comm[16384];
      if (sscanf(buf, "Command %[^\n]\n", comm) == 1)
	setCommand(comm);
      else {
	perror("sscanf");
	return false;
      }
      continue;
    }

    if (!strncasecmp(buf, "MacroKey", 8)) {
      int tmp;
      if (sscanf(buf, "MacroKey %d\n", &tmp) == 1)
	setMacroKey(tmp);
      else {
	perror("sscanf");
	return false;
      }
      continue;
    }

    if (!strncasecmp(buf, "EndAlias", 8)) {
      return true;
    }

  }
}

void AliasEntity::apply() {

  if (!alias_name)
    return;

  const gchar * new_name = gtk_entry_get_text(GTK_ENTRY(alias_name));
  setName((gchar *)new_name);

#ifdef USE_GTK_2
  GtkTextIter start, end;
  gtk_text_buffer_get_bounds(description_buffer, &start, &end);
  gchar * new_desc = gtk_text_buffer_get_text(description_buffer, &start, &end, false);
#else
  gchar * new_desc = gtk_editable_get_chars(GTK_EDITABLE(alias_description), 0, -1);
#endif
  setDescription(new_desc);

  const gchar * new_command = gtk_entry_get_text(GTK_ENTRY(alias_command));
  setCommand((gchar *)new_command);

  const gchar * new_shortcut = gtk_entry_get_text(GTK_ENTRY(alias_shortcut));
  setMacroKey(atoi(new_shortcut));

  if (GTK_TOGGLE_BUTTON(alias_alt)->active)
    setMacroAlt(true);
  else
    setMacroAlt(false);

  if (GTK_TOGGLE_BUTTON(alias_ctrl)->active)
    setMacroCtrl(true);
  else
    setMacroCtrl(false);

#ifdef USE_GTK_2
  gtk_text_buffer_get_bounds(body_buffer, &start, &end);
  gchar * new_body = gtk_text_buffer_get_text(body_buffer, &start, &end, -1);
#else
  gchar * new_body = gtk_editable_get_chars(GTK_EDITABLE(alias_body), 0, -1);
#endif
  setBody(new_body);

#ifdef USE_GTK_2

  gint index = gtk_option_menu_get_history(GTK_OPTION_MENU(option_menu));
  printf ("Index: %d.\n", index);

  // Maybe we can get the list of currently available options, as the list
  // is updated when new groups are created/deleted.

  GList * items = entities->getGroupOptionMenuList();
  gchar * label = (gchar *)g_list_nth_data(items, index);
  printf ("Label: %s.\n", label);

  // Now find out the current group for this Entity
  entities->moveGroup(this, label);


#endif
}

GtkWidget * AliasEntity::getWidgets() {

  if (widget) {
    gtk_widget_hide(option_menu);
    gtk_widget_destroy(option_menu);

    option_menu = entities->getGroupOptionMenu(this);
    gtk_widget_show(option_menu);
    gtk_table_attach(GTK_TABLE(table1), option_menu, 1, 2, 1, 2,
		     (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
		     (GtkAttachOptions) (0), 0, 0);


    return widget;
  }

  GtkWidget *alias_box;
  GtkWidget *frame2;
  GtkWidget *label7;
  GtkWidget *scrolledwindow3;
  GtkWidget *label9;
  GtkWidget *frame3;
  GtkWidget *table2;
  GtkWidget *label10;
  GtkWidget *label11;
  GtkWidget *scrolledwindow4;

#ifdef USE_GTK_2
  GtkWidget * description_view, * body_view;
#endif
  GtkWidget *label12;

  alias_box = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (alias_box);
  gtk_widget_show (alias_box);

  frame2 = gtk_frame_new (_("General"));
  gtk_widget_ref (frame2);
  gtk_widget_show (frame2);
  gtk_box_pack_start (GTK_BOX (alias_box), frame2, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (frame2), 2);

  table1 = gtk_table_new (2, 3, FALSE);
  gtk_widget_ref (table1);
  gtk_widget_show (table1);
  gtk_container_add (GTK_CONTAINER (frame2), table1);
  gtk_container_set_border_width (GTK_CONTAINER (table1), 2);
  gtk_table_set_row_spacings (GTK_TABLE (table1), 2);
  gtk_table_set_col_spacings (GTK_TABLE (table1), 2);

  label7 = gtk_label_new (_("Name:"));
  gtk_widget_ref (label7);
  gtk_widget_show (label7);
  gtk_table_attach (GTK_TABLE (table1), label7, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label7), 0, 0.5);

  alias_name = gtk_entry_new ();
  gtk_widget_ref (alias_name);
  gtk_widget_show (alias_name);
  gtk_table_attach (GTK_TABLE (table1), alias_name, 1, 2, 0, 1,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  label7 = gtk_label_new(_("Group:"));
  gtk_widget_ref(label7);
  gtk_widget_show(label7);
  gtk_table_attach (GTK_TABLE (table1), label7, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label7), 0, 0.5);

  option_menu = entities->getGroupOptionMenu(this);
  gtk_widget_show(option_menu);
  gtk_table_attach(GTK_TABLE(table1), option_menu, 1, 2, 1, 2,
		   (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
		   (GtkAttachOptions) (0), 0, 0);

  scrolledwindow3 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_ref (scrolledwindow3);
  gtk_widget_show (scrolledwindow3);
  gtk_table_attach (GTK_TABLE (table1), scrolledwindow3, 1, 2, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), 0, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow3), GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);

#ifdef USE_GTK_2
  description_buffer = gtk_text_buffer_new(NULL);
  description_view = gtk_text_view_new_with_buffer(description_buffer);
  gtk_widget_ref(description_view);
  gtk_widget_show(description_view);
  gtk_container_add(GTK_CONTAINER(scrolledwindow3), description_view);
#else
  alias_description = gtk_text_new (NULL, NULL);
  gtk_widget_ref (alias_description);

  gtk_widget_show (alias_description);
  gtk_container_add (GTK_CONTAINER (scrolledwindow3), alias_description);
#endif
  label9 = gtk_label_new (_("Description:"));
  gtk_widget_ref (label9);
  gtk_widget_show (label9);
  gtk_table_attach (GTK_TABLE (table1), label9, 0, 1, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label9), 0, 7.45058e-09);

  frame3 = gtk_frame_new (_("Alias"));
  gtk_widget_ref (frame3);
  gtk_widget_show (frame3);
  gtk_box_pack_start (GTK_BOX (alias_box), frame3, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (frame3), 2);

  table2 = gtk_table_new (3, 4, FALSE);
  gtk_widget_ref (table2);
  gtk_widget_show (table2);
  gtk_container_add (GTK_CONTAINER (frame3), table2);
  gtk_container_set_border_width (GTK_CONTAINER (table2), 2);
  gtk_table_set_row_spacings (GTK_TABLE (table2), 2);
  gtk_table_set_col_spacings (GTK_TABLE (table2), 2);

  label10 = gtk_label_new (_("Command:"));
  gtk_widget_ref (label10);
  gtk_widget_show (label10);
  gtk_table_attach (GTK_TABLE (table2), label10, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label10), 0, 0.5);

  alias_command = gtk_entry_new ();
  gtk_widget_ref (alias_command);
  gtk_widget_show (alias_command);
  gtk_table_attach (GTK_TABLE (table2), alias_command, 1, 4, 0, 1,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  label11 = gtk_label_new (_("Shortcut Key:"));
  gtk_widget_ref (label11);
  gtk_widget_show (label11);
  gtk_table_attach (GTK_TABLE (table2), label11, 0, 1, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label11), 0, 0.5);

  alias_shortcut = gtk_entry_new ();
  gtk_widget_ref (alias_shortcut);
  gtk_widget_show (alias_shortcut);
  gtk_table_attach (GTK_TABLE (table2), alias_shortcut, 1, 2, 1, 2,
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  alias_alt = gtk_check_button_new_with_label (_("ALT"));
  gtk_widget_ref (alias_alt);
  gtk_widget_show (alias_alt);
  gtk_table_attach (GTK_TABLE (table2), alias_alt, 2, 3, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  alias_ctrl = gtk_check_button_new_with_label (_("CTRL"));
  gtk_widget_ref (alias_ctrl);
  gtk_widget_show (alias_ctrl);
  gtk_table_attach (GTK_TABLE (table2), alias_ctrl, 3, 4, 1, 2,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (0), 0, 0);

  scrolledwindow4 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_ref (scrolledwindow4);
  gtk_widget_show (scrolledwindow4);
  gtk_table_attach (GTK_TABLE (table2), scrolledwindow4, 1, 4, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), 0, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow4), GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);

#ifdef USE_GTK_2
  body_buffer = gtk_text_buffer_new(NULL);
  body_view = gtk_text_view_new_with_buffer(body_buffer);
  gtk_widget_ref(body_view);
  gtk_widget_show(body_view);
  gtk_container_add(GTK_CONTAINER(scrolledwindow4), body_view);
#else
  alias_body = gtk_text_new (NULL, NULL);
  gtk_widget_ref (alias_body);
  gtk_widget_show (alias_body);
  gtk_container_add (GTK_CONTAINER (scrolledwindow4), alias_body);
#endif

  label12 = gtk_label_new (_("Body:"));
  gtk_widget_ref (label12);
  gtk_widget_show (label12);
  gtk_table_attach (GTK_TABLE (table2), label12, 0, 1, 2, 3,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), 0, 0);
  gtk_misc_set_alignment (GTK_MISC (label12), 0, 7.45058e-09);

  if (getName())
    gtk_entry_set_text(GTK_ENTRY(alias_name), getName());

#ifdef USE_GTK_2
  if (getDescription()) {
    GtkTextIter start, end;
    gtk_text_buffer_get_bounds(description_buffer, &start, &end);
    gtk_text_buffer_insert(description_buffer, &end, getDescription(), -1);
  }
#else
  if (getDescription())
    gtk_text_insert(GTK_TEXT(alias_description), NULL, NULL, NULL, getDescription(), -1);
#endif

  if (getCommand())
    gtk_entry_set_text(GTK_ENTRY(alias_command), getCommand());

  char buf[1024];
  snprintf(buf, 1024, "%d", getMacroKey());
  gtk_entry_set_text(GTK_ENTRY(alias_shortcut), buf);

#ifdef USE_GTK_2
  if (getBody()) {
    GtkTextIter start, end;
    gtk_text_buffer_get_bounds(body_buffer, &start, &end);
    gtk_text_buffer_insert(body_buffer, &end, getBody(), -1);
  }
#else
  if (getBody())
    gtk_text_insert(GTK_TEXT(alias_body), NULL, NULL, NULL, getBody(), -1);
#endif

  widget = alias_box;
  return widget;
}

