#include <sys/stat.h>
#ifndef WIN32
#include <unistd.h>
#endif
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <gdk/gdkkeysyms.h>
#include <string.h>

#include "mudclient.h"

#include "Connection.h"
#include "Message.h"

struct auto_data {
  int countdown;
  gint timeout; // Need to remove me.
  GtkWidget * window; // Need to destroy me.
  GtkWidget * label; // Don't touch me.
  Connection * conn; // The connection this autoreconnect is from.
};

int auto_reconnect_timeout(gpointer d) {
  struct auto_data * data = (struct auto_data *)d;
  char buf[1024];
  
  if (data->countdown == 0) {
    gtk_timeout_remove(data->timeout);
    data->conn->reconnect();
    gtk_widget_hide(data->window);
    gtk_widget_destroy(data->window);
    free(data);
    return 0;
  }
  
  sprintf(buf, "Reconnecting to %.900s:%d in %d second%s.\n", data->conn->getMUD()->getHostname(), data->conn->getMUD()->getPort(), data->countdown, data->countdown == 1 ? "" : "s");
  gtk_label_set_text(GTK_LABEL(data->label), buf);
  data->countdown--;
  return 1;
}

void on_disconnect_button_clicked(GtkButton * button, void * d) {
  
  struct auto_data *data = (struct auto_data *)d;
  gtk_timeout_remove(data->timeout);
  gtk_widget_hide(data->window);
  gtk_widget_destroy(data->window);
  delete data->conn; // This should remove the item from the MainWindow.
  data->conn = NULL;
  free(data);
}

void on_reconnect_button_clicked(GtkButton * button, void * d) {

  struct auto_data *data = (struct auto_data *)d;
  gtk_timeout_remove(data->timeout);
  data->conn->reconnect();
  gtk_widget_hide(data->window);
  gtk_widget_destroy(data->window);
  free(data);
}

void create_autoconnect_dialog(struct auto_data * data) {

  GtkWidget *vbox1;
  GtkWidget *hbuttonbox1;
  GtkWidget *reconnect_button;
  GtkWidget *disconnect_button;

  char buf[1024];

  data->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

#ifndef USE_GTK_2
  gtk_object_set_data (GTK_OBJECT (data->window), "data->window", data->window);
#endif
  gtk_window_set_title (GTK_WINDOW (data->window), "Automatically Reconnect?");
  gtk_window_set_modal (GTK_WINDOW (data->window), TRUE);
#ifdef USE_GTK_2
  gtk_window_set_resizable(GTK_WINDOW(data->window), TRUE);
#else
  gtk_window_set_policy (GTK_WINDOW (data->window), TRUE, TRUE, TRUE);
#endif

  vbox1 = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (vbox1);

#ifndef USE_GTK_2
  gtk_object_set_data_full (GTK_OBJECT (data->window), "vbox1", vbox1,
                            (GtkDestroyNotify) gtk_widget_unref);
#endif
  gtk_widget_show (vbox1);
  gtk_container_add (GTK_CONTAINER (data->window), vbox1);

  sprintf(buf, "Reconnecting to %.900s:%d in %d second%s.\n", data->conn->getMUD()->getHostname(), data->conn->getMUD()->getPort(), data->countdown, data->countdown == 1 ? "" : "s");

  data->label = gtk_label_new (buf);
  gtk_widget_ref (data->label);
#ifndef USE_GTK_2
  gtk_object_set_data_full (GTK_OBJECT (data->window), "data->label", data->label,
                            (GtkDestroyNotify) gtk_widget_unref);
#endif
  gtk_widget_show (data->label);
  gtk_box_pack_start (GTK_BOX (vbox1), data->label, FALSE, FALSE, 0);
  gtk_label_set_justify (GTK_LABEL (data->label), GTK_JUSTIFY_LEFT);
  gtk_label_set_line_wrap (GTK_LABEL (data->label), TRUE);

  hbuttonbox1 = gtk_hbutton_box_new ();
  gtk_widget_ref (hbuttonbox1);
#ifndef USE_GTK_2
  gtk_object_set_data_full (GTK_OBJECT (data->window), "hbuttonbox1", hbuttonbox1,
                            (GtkDestroyNotify) gtk_widget_unref);
#endif
  gtk_widget_show (hbuttonbox1);
  gtk_box_pack_start (GTK_BOX (vbox1), hbuttonbox1, FALSE, FALSE, 0);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox1), GTK_BUTTONBOX_SPREAD);

#if defined USE_GTK_2
  reconnect_button = gtk_button_new_with_label_and_pixmap(hbuttonbox1, "Reconnect", "reconnect.xpm");
#else
  reconnect_button = gtk_button_new_with_label (_("Reconnect"));
#endif
  gtk_widget_ref (reconnect_button);
#ifndef USE_GTK_2
  gtk_object_set_data_full (GTK_OBJECT (data->window), "reconnect_button", reconnect_button,
                            (GtkDestroyNotify) gtk_widget_unref);
#endif
  gtk_widget_show (reconnect_button);
  gtk_container_add (GTK_CONTAINER (hbuttonbox1), reconnect_button);
  GTK_WIDGET_SET_FLAGS (reconnect_button, GTK_CAN_DEFAULT);

#if defined USE_GTK_2
  disconnect_button = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
#else
  disconnect_button = gtk_button_new_with_label (_("Cancel"));
#endif

  gtk_widget_ref (disconnect_button);
#ifndef USE_GTK_2
  gtk_object_set_data_full (GTK_OBJECT (data->window), "disconnect_button", disconnect_button,
                            (GtkDestroyNotify) gtk_widget_unref);
#endif
  gtk_widget_show (disconnect_button);
  gtk_container_add (GTK_CONTAINER (hbuttonbox1), disconnect_button);
  GTK_WIDGET_SET_FLAGS (disconnect_button, GTK_CAN_DEFAULT);

#ifdef USE_GTK_2
  g_signal_connect_data(reconnect_button, "clicked",
			GTK_SIGNAL_FUNC(on_reconnect_button_clicked),
			data,
			NULL,
			(GConnectFlags)0);
  g_signal_connect_data(disconnect_button, "clicked",
			GTK_SIGNAL_FUNC(on_disconnect_button_clicked),
			data,
			NULL,
			(GConnectFlags)0);

#else
  gtk_signal_connect (GTK_OBJECT (reconnect_button), "clicked",
                      GTK_SIGNAL_FUNC (on_reconnect_button_clicked),
                      (void *)data);
  gtk_signal_connect (GTK_OBJECT (disconnect_button), "clicked",
                      GTK_SIGNAL_FUNC (on_disconnect_button_clicked),
                      (void *)data);
#endif

  gtk_widget_show(data->window);
}

struct auto_dialog {
  GtkWidget * dialog;
  Connection * conn;
};

void on_autoreconnect_http_message_button_clicked(GtkButton * button, gpointer data) {

  struct auto_dialog * d = (struct auto_dialog *)data;

  gtk_widget_hide(d->dialog);
  gtk_widget_destroy(d->dialog);

  // Attempt to restart the HTTP connection.
  d->conn->getSocket()->restart(d->conn->getMUD()->getHostname(), d->conn->getMUD()->getPort());
  free(d);
}

void on_autoreconnect_message_button_clicked(GtkButton * button, gpointer data) {

  struct auto_dialog * d = (struct auto_dialog *)data;

  gtk_widget_hide(d->dialog);
  gtk_widget_destroy(d->dialog);

  delete d->conn;
  free(d);
}

void autoreconnect_message(char * title, char * text, Connection * conn) {

#ifdef USE_GTK_2

  struct auto_dialog * auto_d = (struct auto_dialog *)malloc(sizeof(struct auto_dialog));
  auto_d->conn = conn;

  auto_d->dialog = gtk_dialog_new();

  gtk_window_set_title(GTK_WINDOW(auto_d->dialog), title);
  gtk_window_set_resizable(GTK_WINDOW(auto_d->dialog), FALSE);

  // Set up the message and icon.

  GtkWidget * vbox = GTK_DIALOG(auto_d->dialog)->vbox;
  gtk_widget_show(vbox);

  GtkWidget * hbox = gtk_hbox_new(FALSE, 0);
  gtk_widget_show(hbox);

  GtkStockItem item;

  char * icon;
  if (!strcmp(title, "Error"))
    icon = GTK_STOCK_DIALOG_ERROR;
  else if (!strcmp(title, "Warning"))
    icon = GTK_STOCK_DIALOG_WARNING;
  else
    icon = GTK_STOCK_DIALOG_INFO;

  gtk_stock_lookup(icon, &item);
  GtkWidget * image = gtk_image_new_from_stock(icon, GTK_ICON_SIZE_DIALOG);

  gtk_widget_show(image);

  GtkWidget * label = gtk_label_new(text);
  gtk_widget_show(label);

  gtk_box_pack_start(GTK_BOX(hbox), image, FALSE, FALSE, 0);
  gtk_box_pack_end(GTK_BOX(hbox), label, FALSE, FALSE, 0);

  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

  // Set up the button area
  GtkWidget * action_area = GTK_DIALOG(auto_d->dialog)->action_area;
  gtk_widget_show(action_area);
  gtk_widget_show (action_area);
  gtk_container_set_border_width (GTK_CONTAINER (action_area), 10);
  
  GtkWidget * hbuttonbox1 = gtk_hbutton_box_new ();
  gtk_widget_ref (hbuttonbox1);
  gtk_widget_show (hbuttonbox1);
  gtk_box_pack_start (GTK_BOX (action_area), hbuttonbox1, TRUE, TRUE, 0);

  GtkWidget * button;

  /* Only display this widget if the connection is set up for HTTP connect. */
  if (conn->queryPreferences()->getTurfHttpd()) {
    button = gtk_button_new_with_label (_("Attempt HTTP Reconnect"));
    gtk_widget_ref(button);
    gtk_widget_show (button);
    gtk_container_add (GTK_CONTAINER (hbuttonbox1), button);

    g_signal_connect_data (button, "clicked",
			   GTK_SIGNAL_FUNC (on_autoreconnect_http_message_button_clicked),
			   (gpointer)auto_d,
			   NULL,
			   (GConnectFlags)0);
  }

  button = gtk_button_new_with_label (_("Close Window"));
  gtk_widget_ref (button);
  gtk_widget_show (button);
  gtk_container_add (GTK_CONTAINER (hbuttonbox1), button);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  g_signal_connect_data(button, "clicked",
			GTK_SIGNAL_FUNC(on_autoreconnect_message_button_clicked),
			auto_d,
			NULL,
			(GConnectFlags)0);

  gtk_widget_grab_focus(GTK_WIDGET (button));
  gtk_widget_show(auto_d->dialog);

#else

  struct auto_dialog * auto_d = (struct auto_dialog *)malloc(sizeof(struct auto_dialog));
  auto_d->conn = conn;

  GtkWidget *dialog_vbox1;
  GtkWidget *label;
  GtkWidget *dialog_action_area1;
  GtkWidget *hbuttonbox1;
  GtkWidget *button;

  auto_d->dialog = gtk_dialog_new ();
  gtk_object_set_data (GTK_OBJECT (auto_d->dialog), "dialog", auto_d->dialog);
  gtk_window_set_title (GTK_WINDOW (auto_d->dialog), title);
  gtk_window_set_policy (GTK_WINDOW (auto_d->dialog), TRUE, TRUE, FALSE);

  dialog_vbox1 = GTK_DIALOG (auto_d->dialog)->vbox;

  gtk_object_set_data (GTK_OBJECT (auto_d->dialog), "dialog_vbox1", dialog_vbox1);
  gtk_widget_show (dialog_vbox1);

  label = gtk_label_new (text);
  gtk_widget_ref (label);
  gtk_object_set_data_full (GTK_OBJECT (auto_d->dialog), "label", label,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label);
  gtk_box_pack_start (GTK_BOX (dialog_vbox1), label, FALSE, FALSE, 0);
  gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_FILL);
  gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);

  gtk_misc_set_padding (GTK_MISC (label), 0, 13);

  dialog_action_area1 = GTK_DIALOG (auto_d->dialog)->action_area;
  gtk_object_set_data (GTK_OBJECT (auto_d->dialog), "dialog_action_area1", dialog_action_area1);
  gtk_widget_show (dialog_action_area1);
  gtk_container_set_border_width (GTK_CONTAINER (dialog_action_area1), 10);

  hbuttonbox1 = gtk_hbutton_box_new ();
  gtk_widget_ref (hbuttonbox1);
  gtk_object_set_data_full (GTK_OBJECT (auto_d->dialog), "hbuttonbox1", hbuttonbox1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (hbuttonbox1);
  gtk_box_pack_start (GTK_BOX (dialog_action_area1), hbuttonbox1, TRUE, TRUE, 0)
;

  /* Only display this widget if the connection is set up for HTTP connect. */
  if (conn->queryPreferences()->getTurfHttpd()) {
    button = gtk_button_new_with_label (_("Attempt HTTP Reconnect"));
    gtk_widget_ref(button);
    gtk_object_set_data_full (GTK_OBJECT (auto_d->dialog), "button", button,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (button);
    gtk_container_add (GTK_CONTAINER (hbuttonbox1), button);
    gtk_signal_connect (GTK_OBJECT (button), "clicked",
                        GTK_SIGNAL_FUNC (on_autoreconnect_http_message_button_clicked), (gpointer)auto_d);

  }

  button = gtk_button_new_with_label (_("Close Window"));

  gtk_widget_ref (button);
  gtk_object_set_data_full (GTK_OBJECT (auto_d->dialog), "button", button,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (button);
  gtk_container_add (GTK_CONTAINER (hbuttonbox1), button);
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
                      GTK_SIGNAL_FUNC (on_autoreconnect_message_button_clicked), (gpointer)auto_d);

  gtk_widget_grab_focus(GTK_WIDGET (button));

  gtk_widget_show(auto_d->dialog);
#endif
}

void auto_reconnect_init(Connection * c) {

  if (c->queryPreferences()->getAutoReconnect()) {
    struct auto_data * data = (struct auto_data *)malloc(sizeof(struct auto_data));
    data->conn = c;
    data->countdown = 10;
    create_autoconnect_dialog(data);
    data->timeout = gtk_timeout_add(1000, auto_reconnect_timeout, data);
  } else {
    char buf[1024];

    int written;
    int raw_read;
    int read;

    c->getSocket()->getStats(&written, &raw_read, &read);

    sprintf(buf, "Connection to %.900s port %d closed.\n"
	    "Total bytes sent: %d\n"
	    "Raw bytes received: %d\n"
	    "Uncompressed bytes received: %d\n",
	    c->getMUD()->getHostname(),
	    c->getMUD()->getPort(),
	    written,
	    raw_read,
	    read
	    );

    autoreconnect_message(_("Connection closed."), buf, c);
  }
}



