#ifdef WIN32
#include <algorithm>
#else
#include <algo.h>
#endif

#include "mudclient.h"

#define ENTITY_HANDLER_C
#include "EntityHandler.h"

#include "AliasEntity.h"
#include "TriggerEntity.h"

extern EntityHandler * entities;

EntityHandler::EntityHandler() {
}

EntityHandler::~EntityHandler() {
}

static int GroupCmp(EntityGroup * e1, EntityGroup * e2) {
  return (e1 < e2);
}

void EntityHandler::addEntity(char * group, Entity * e) {
  EntityGroup * eg = findGroup(group, true);
  eg->addEntity(e);
}

void EntityHandler::removeEntity(Entity * e) {
  for (GroupList::iterator i = groups.begin(); i != groups.end(); i++) {
    if ((*i)->removeEntity(e))
      return;
  }
}

void EntityHandler::removeEntityGroup(EntityGroup * e) {
  GroupList::iterator i = std::lower_bound(groups.begin(),
					   groups.end(),
					   e,
					   GroupCmp);

  if (i == groups.end() || (*i) != e)
    return;

  groups.erase(i);
}

EntityGroup * EntityHandler::findGroup(char * name, bool create) {

  for (GroupList::iterator i = groups.begin(); i != groups.end(); i++) {
    if (!strcmp((*i)->getName(), name))
      return (*i);
  }

  if (!create)
    return NULL;

  EntityGroup * eg = new EntityGroup();
  eg->setName(name);
  eg->setDescription(_("Dummy description."));
  addGroup(eg);
  return eg;
}

void EntityHandler::addGroup(EntityGroup * e) {
  GroupList::iterator i = std::lower_bound(groups.begin(),
					    groups.end(),
					    e,
					    GroupCmp);
  groups.insert(i, e);
}

void EntityHandler::moveGroup(Entity * e, char * group) {

  EntityGroup * new_group = findGroup(group, false);
  EntityGroup * old_group = NULL;
  
  for (GroupList::iterator i = groups.begin(); i != groups.end(); ++i) {
    if ((*i)->contains(e)) {
      old_group = *i;
      break;
    }
  }
 
  if (new_group == old_group)
    return;

  old_group->removeEntity(e);
  new_group->addEntity(e);
}

void EntityHandler::removeGroup(EntityGroup * e) {
  GroupList::iterator i = std::lower_bound(groups.begin(),
					    groups.end(),
					    e,
					    GroupCmp);
  if (i == groups.end() || (*i) != e)
    return;

  groups.erase(i);
}

void EntityHandler::findExecute(EntityType type, char * field, char * value, void * data, bool multiple) {
  for (GroupList::iterator i = groups.begin(); i != groups.end(); i++)
    if ((*i)->findExecute(type, field, value, data, multiple) && !multiple)
      return;
}

void EntityHandler::load() {

  printf ("Breakpoint on me.\n");

  char buf[16384];
  FILE * fp;

  fp = openFile("entities.dat", NULL, "r");
  if (!fp) {
    printf("EntityHandler:: missing entities.dat\n");
    return;
  }

  while (true) {
    if (!fgets(buf, 16384, fp)) {
      perror("fgets");
      fclose(fp);
      return;
    }

    if (!strncasecmp(buf, "EntityGroup", 11)) {
      char name[16384];
      if (sscanf(buf, "EntityGroup %[^\n]\n", name) != 1) {
	perror("sscanf");
	fclose(fp);
	return;
      }

      EntityGroup * tmp = new EntityGroup();
      if (!tmp->load(fp, name)) {
	delete tmp;
	return;
      }

      addGroup(tmp);
      continue;
    }

    return;
  }

}

void EntityHandler::save() {
  FILE * fp;

  fp = openFile("entities.dat", NULL, "w");
  if (!fp) {
    perror("openFile");
    return;
  }

  for (GroupList::iterator i = groups.begin(); i != groups.end(); i++)
    (*i)->save(fp);

  fclose(fp);
}

static void entity_apply_clicked(GtkButton * button, gpointer data) {
  Entity * e = entities->getSelectedEntity();
  ((EntityHandler *)data)->apply();
  entities->update();
  if (e)
    entities->selectEntity(e);
}

static void entity_ok_clicked(GtkButton * button, gpointer data) {
  ((EntityHandler *)data)->apply();
  ((EntityHandler *)data)->close();
}

static void entity_cancel_clicked(GtkButton * button, gpointer data) {
  ((EntityHandler *)data)->close();
}

void EntityHandler::close() {

  // Remove the entity widget from the vbox.
  if (entity_widget)
    gtk_container_remove(GTK_CONTAINER(vbox), entity_widget);

  // Cause all the plugins to destroy their widgets.
  for (GroupList::iterator i = groups.begin(); i != groups.end(); i++) {
    (*i)->destroyWidgets();
  }

  gtk_widget_hide(window);
  gtk_widget_destroy(window);
  window = NULL;
}

void EntityHandler::apply() {
  for (GroupList::iterator i = groups.begin(); i != groups.end(); i++) {
    (*i)->apply();
  }
  save();
}


#ifdef USE_GTK_2

void display_entity(GtkTreeModel * model, GtkTreePath * path, GtkTreeIter  * iter, gpointer user_data) {

  // Can we figure out the name of the item at iter?

  GValue value;
  gtk_tree_model_get_value(model, iter, 0, &value);

  g_value_unset(&value);
  

}

void tree_view_row_activated(GtkTreeSelection * selection, gpointer user_data) {
  entities->row_selected(selection, user_data);
}

void EntityHandler::row_selected (GtkTreeSelection *selection,
				  gpointer          data) {
  int id = getSelectedID();

  if (id >= 0) {
    Entity * e = findByID(id);
    if (!e) {
      printf ("FIXME: Entity in list selected, but not found!\n");
      return;
    }

    if (entity_widget) {
      gtk_container_remove(GTK_CONTAINER(vbox), entity_widget);
      //    gtk_widget_hide(entity_widget);
      //    gtk_widget_destroy(entity_widget);
    }
    entity_widget = e->getWidgets();
    gtk_box_pack_start(GTK_BOX(vbox), entity_widget, TRUE, TRUE, 0);
    gtk_widget_show(entity_widget);

  } else {
    if (entity_widget)
      gtk_container_remove(GTK_CONTAINER(vbox), entity_widget);
  }
}


#endif

void new_group_clicked(GtkWidget * widget, gpointer data) {

  char buf[1024];

  EntityGroup * eg = new EntityGroup();

  int i;
  for (i = 0; i < 100; i++) {
    snprintf(buf, 1024, "Group %d", i);
    if (!entities->findGroup(buf, false)) {
      eg->setName(buf);
      break;
    }
  }

  if (i == 100) {
    printf ("FIXME: 100 dummy entity groups are insufficient.\n");
    return;
  }

  entities->addGroup(eg);

  // Update the tree and ensure this is the selected item.
  entities->update();

  entities->selectEntity(eg);
}

#ifdef USE_GTK_2
gboolean select_by_id(GtkTreeModel * tree_model, GtkTreePath * path, GtkTreeIter * iter, gpointer data) {
  entities->selectByID(iter, data);
  return false;
}

void EntityHandler::selectByID(GtkTreeIter * iter, gpointer data) {
  int id = *(int *)data;
  char * id_buf;

  gtk_tree_model_get(GTK_TREE_MODEL(tree_model), iter, 1, &id_buf, -1);

  if (atoi(id_buf) == id) {
    gtk_tree_selection_select_iter(selection, iter);
    return;
  }

  g_free(id_buf);
}

void EntityHandler::selectEntity(Entity * e) {
  int id = e->getID();
  gtk_tree_model_foreach(GTK_TREE_MODEL(tree_model), select_by_id, (gpointer)&id);
}
#else
void EntityHandler::selectEntity(Entity * e) {

  int id = e->getID();

  // Figure out which tree item we're trying to select.
  printf("FIXME: implement EntityHandler::selectEntity() for GTK 1.2.\n");

}
#endif

void EntityHandler::newAlias() {

  GroupList::iterator i = groups.begin();
  if (!(*i)) {
    new Message(_("Error"), _("You need to create a group first."), true);
    return;
  }

  EntityGroup * group = getSelectedGroup();
  if (!group) {
    new Message("Error", _("Please select the group you wish the alias to be entered into and then choose 'New Alias'"), true);
    return;
  }

  AliasEntity * a = new AliasEntity();
  a->setName(_("EDIT ME"));
  addEntity(group->getName(), a);
  a->setCommand(_("EDIT ME"));

  update();

  selectEntity(a);
}

void EntityHandler::newTrigger() {
  GroupList::iterator i = groups.begin();
  if (!(*i)) {
    new Message(_("Error"), _("You need to create a group first."), true);
    return;
  }

  EntityGroup * group = getSelectedGroup();
  if (!group) {
    new Message("Error", _("Please select the group you wish the trigger to be entered into and then choose 'New Trigger'"), true);
    return;
  }

  TriggerEntity * t = new TriggerEntity();
  
  t->setName(_("EDIT ME"));
  addEntity(group->getName(), t);
  t->setText(_("EDIT ME"));
  update();

  selectEntity(t);
}

void new_alias_clicked(GtkWidget * widget, gpointer data) {
  entities->newAlias();
}

void new_trigger_clicked(GtkWidget * widget, gpointer data) {
  entities->newTrigger();
}

void delete_entity_clicked(GtkWidget * widget, gpointer data) {
  entities->deleteSelected();
}

Entity * EntityHandler::getSelectedEntity() {
  int id = getSelectedID();
  Entity * e = entities->findByID(id);
  return e;
}
EntityGroup * EntityHandler::getSelectedGroup() {

  int id = getSelectedID();
  if (id == -1)
    return NULL;

  Entity * e = findByID(id);
  if (!e || e->getType() != EntityGroupGroup)
    return NULL;
  return (EntityGroup *)e;
}

void EntityHandler::deleteSelected() {

  int id = getSelectedID();
  if (id == -1)
    return;
  
  Entity * e = findByID(id);
  if (!e) {
    printf ("FIXME: Entity in list selected, but not found!\n");
    return;
  }
  
  int count = 0;
  switch (e->getType()) {
  case EntityGroupGroup:
    printf ("FIXME: check for entities within group before deleting.\n");
    count = ((EntityGroup *)e)->count();
    if (count) {
      new Message("Error", _("A group must be empty before it can be deleted.  The selected group was not empty."), true);
      return;
    }

    removeEntityGroup((EntityGroup *)e);
    gtk_container_remove(GTK_CONTAINER(vbox), entity_widget);
    entity_widget = NULL;
    break;
    
  default:
    removeEntity(e);
    gtk_container_remove(GTK_CONTAINER(vbox), entity_widget);
    entity_widget = NULL;
    break;
    
  case EntitySystemTrigger:
    new Message("Error", _("This System Trigger is in use by a plugin or Papaya and may not be deleted."), true);
    break;
    
  }

  update();
}

#ifdef USE_GTK_2
int EntityHandler::getSelectedID() {
  GtkTreeModel * model;
  GtkTreeIter iter;

  if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
    gchar * id_buf;
    int id;

    gtk_tree_model_get(model, &iter, 1, &id_buf, -1);
    id = atoi(id_buf);
    g_free(id_buf);
    return id;
  }

  return -1;
}

#else
int EntityHandler::getSelectedID() {
  printf ("FIXME: EntityHandler::getSelectedID()\n");
  return -1;
}
#endif

void EntityHandler::update() {
#ifdef USE_GTK_2
  populate_tree_model(tree_model);
  gtk_tree_view_expand_all(GTK_TREE_VIEW(tree_view));
#else
  populate_tree(tree_widget);
#endif
}

void EntityHandler::edit() {
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(window), "Triggers and Aliases");
  gtk_widget_show(window);

  GtkWidget * hbox = gtk_hbox_new(FALSE, 3);
  gtk_widget_show(hbox);
  gtk_container_add(GTK_CONTAINER(window), hbox);

  GtkWidget * tree_vbox = gtk_vbox_new(FALSE, 3);
  gtk_widget_show(tree_vbox);
  gtk_box_pack_start(GTK_BOX(hbox), tree_vbox, TRUE, TRUE, 0);
  
  GtkWidget * tree_win = gtk_scrolled_window_new(NULL, NULL);
  gtk_widget_show(tree_win);
  gtk_container_set_border_width(GTK_CONTAINER(hbox), 3);
  gtk_box_pack_start(GTK_BOX(tree_vbox), tree_win, TRUE, TRUE, 0);

#ifdef USE_GTK_2
  GtkCellRenderer * tree_renderer;
  GtkTreeViewColumn * tree_column, *tree_data_column;

  tree_model = gtk_tree_store_new(2, G_TYPE_STRING, G_TYPE_STRING);
  //  populate_tree_model(tree_model);

  tree_view = gtk_tree_view_new_with_model(GTK_TREE_MODEL(tree_model));
  g_object_unref(tree_model);

  gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(tree_view), false);

  tree_renderer = gtk_cell_renderer_text_new();
  //  g_object_set(G_OBJECT(tree_renderer), "foreground", "green", NULL);
  
  tree_column = gtk_tree_view_column_new_with_attributes(_("Triggers and Aliases"),
						    tree_renderer,
						    "text", 0,
						    NULL);

  gtk_tree_view_append_column(GTK_TREE_VIEW(tree_view), tree_column);

  tree_data_column = gtk_tree_view_column_new();
  gtk_tree_view_column_set_visible(tree_data_column, false);

  gtk_tree_view_append_column(GTK_TREE_VIEW(tree_view), tree_data_column);

  gtk_widget_show(tree_view);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(tree_win), tree_view);

  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree_view));

  g_signal_connect_data(G_OBJECT(selection), "changed",
			GTK_SIGNAL_FUNC(tree_view_row_activated),
			this, NULL, (GConnectFlags) 0);

#else
  tree_widget = gtk_tree_new();
  gtk_widget_show(tree_widget);
  gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(tree_win), tree_widget);
#endif

  // Create hbuttonbox to hold New Group, New Alias, New Trigger, Delete
  GtkWidget * hbuttonbox = gtk_hbutton_box_new();
  gtk_widget_show(hbuttonbox);
  gtk_box_pack_start(GTK_BOX(tree_vbox), hbuttonbox, FALSE, FALSE, 0);

  GtkWidget * new_group = gtk_button_new_with_label(_("New Group"));
  GtkWidget * new_alias = gtk_button_new_with_label(_("New Alias"));
  GtkWidget * new_trigger = gtk_button_new_with_label(_("New Trigger"));
  GtkWidget * delete_entity = gtk_button_new_with_label(_("Delete"));

  gtk_widget_show(new_group);
  gtk_widget_show(new_alias);
  gtk_widget_show(new_trigger);
  gtk_widget_show(delete_entity);

  gtk_box_pack_start(GTK_BOX(hbuttonbox), new_group, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbuttonbox), new_alias, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbuttonbox), new_trigger, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbuttonbox), delete_entity, FALSE, FALSE, 0);

#ifdef USE_GTK_2
  g_signal_connect_data(G_OBJECT(new_group), "clicked",
			GTK_SIGNAL_FUNC(new_group_clicked),
			this, NULL, (GConnectFlags)0);

  g_signal_connect_data(G_OBJECT(new_alias), "clicked",
			GTK_SIGNAL_FUNC(new_alias_clicked),
			this, NULL, (GConnectFlags)0);

  g_signal_connect_data(G_OBJECT(new_trigger), "clicked",
			GTK_SIGNAL_FUNC(new_trigger_clicked),
			this, NULL, (GConnectFlags)0);

  g_signal_connect_data(G_OBJECT(delete_entity), "clicked",
			GTK_SIGNAL_FUNC(delete_entity_clicked),
			this, NULL, (GConnectFlags)0);
#else
  gtk_signal_connect(GTK_OBJECT(new_group), "clicked",
		     GTK_SIGNAL_FUNC(new_group_clicked),
		     this);

  gtk_signal_connect(GTK_OBJECT(new_alias), "clicked",
		     GTK_SIGNAL_FUNC(new_group_clicked),
		     this);

  gtk_signal_connect(GTK_OBJECT(new_trigger), "clicked",
		     GTK_SIGNAL_FUNC(new_group_clicked),
		     this);

  gtk_signal_connect(GTK_OBJECT(delete_entity), "clicked",
		     GTK_SIGNAL_FUNC(new_group_clicked),
		     this);

#endif

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(vbox);
  gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);
  GtkWidget * button_box = gtk_hbutton_box_new();
  gtk_widget_show(button_box);
  gtk_box_pack_end(GTK_BOX(vbox), button_box, FALSE, FALSE, 0);

#ifdef USE_GTK_2
  apply_button = gtk_button_new_from_stock(GTK_STOCK_APPLY);
#else
  apply_button = gtk_button_new_with_label(_("Apply"));
#endif
  gtk_widget_show(apply_button);
  gtk_box_pack_start(GTK_BOX(button_box), apply_button, TRUE, FALSE, 0);

#ifdef USE_GTK_2
  cancel_button = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
#else
  cancel_button = gtk_button_new_with_label(_("Cancel"));
#endif
  gtk_widget_show(cancel_button);
  gtk_box_pack_start(GTK_BOX(button_box), cancel_button, TRUE, FALSE, 0);

#ifdef USE_GTK_2
  ok_button = gtk_button_new_from_stock(GTK_STOCK_OK);
#else
  ok_button = gtk_button_new_with_label(_("Ok"));
#endif
  gtk_widget_show(ok_button);
  gtk_box_pack_start(GTK_BOX(button_box), ok_button, TRUE, FALSE, 0);

#ifdef USE_GTK_2
  g_signal_connect_data(apply_button, "clicked",
			GTK_SIGNAL_FUNC(entity_apply_clicked),
			this,
			NULL,
			(GConnectFlags)0);

  g_signal_connect_data(ok_button, "clicked",
			GTK_SIGNAL_FUNC(entity_ok_clicked),
			this,
			NULL,
			(GConnectFlags)0);

  g_signal_connect_data(cancel_button, "clicked",
			GTK_SIGNAL_FUNC(entity_cancel_clicked),
			this,
			NULL,
			(GConnectFlags)0);
#else
  gtk_signal_connect(GTK_OBJECT(apply_button), "clicked",
		     GTK_SIGNAL_FUNC(entity_apply_clicked),
		     this);

  gtk_signal_connect(GTK_OBJECT(ok_button), "clicked",
		     GTK_SIGNAL_FUNC(entity_ok_clicked),
		     this);

  gtk_signal_connect(GTK_OBJECT(cancel_button), "clicked",
		     GTK_SIGNAL_FUNC(entity_cancel_clicked),
		     this);
#endif

  entity_widget = NULL;

  update();
}

void entity_selected(GtkWidget * widget, gpointer data) {
  entities->entitySelected(data);
}

void EntityHandler::entitySelected(gpointer data) {

  Entity * e = (Entity *)data;

  if (entity_widget) {
    gtk_container_remove(GTK_CONTAINER(vbox), entity_widget);
    //    gtk_widget_hide(entity_widget);
    //    gtk_widget_destroy(entity_widget);
  }
  entity_widget = e->getWidgets();
  gtk_box_pack_start(GTK_BOX(vbox), entity_widget, TRUE, TRUE, 0);
  gtk_widget_show(entity_widget);

}

#ifdef USE_GTK_2
void EntityHandler::populate_tree_model(GtkTreeStore * model) {

  gtk_tree_store_clear(model);

  for (GroupList::iterator i = groups.begin(); i != groups.end(); i++) {
    GtkTreeIter iter;
    char id_buf[1024];

    gtk_tree_store_insert(GTK_TREE_STORE(model),
			  &iter,
			  NULL,
			  0);

    snprintf(id_buf, 1024, "%d", (*i)->getID());

    gtk_tree_store_set(GTK_TREE_STORE(model), &iter,
		       0, (*i)->getName(), 1, id_buf, -1);

    (*i)->populate_tree_model(model, &iter);
  }

}
#else
void EntityHandler::populate_tree(GtkWidget * widget) {
  for (GroupList::iterator i = groups.begin(); i != groups.end(); i++) {
    GtkWidget * sub_tree;
    GtkWidget * item;

    item = gtk_tree_item_new_with_label((*i)->getName());
    gtk_signal_connect (GTK_OBJECT(item), "select",
			GTK_SIGNAL_FUNC(entity_selected), (*i));

    gtk_tree_append(GTK_TREE(widget), item);
    //    gtk_tree_item_expand(GTK_TREE_ITEM(item));
    gtk_widget_show(item);

    sub_tree = gtk_tree_new();
    gtk_tree_item_set_subtree(GTK_TREE_ITEM(item), sub_tree);
    gtk_widget_show(sub_tree);

    (*i)->populate_tree(sub_tree);
  }

}
#endif

Entity * EntityHandler::findByID(int id) {

  for (GroupList::iterator i = groups.begin(); i != groups.end(); i++) {
    if ((*i)->getID() == id)
      return (*i);
    Entity * e = (*i)->findByID(id);
    if (e)
      return e;
  }

  return NULL;
}

EntityGroup * EntityHandler::getEntityGroupForEntity(Entity * e) {
  for (GroupList::iterator i = groups.begin(); i != groups.end(); i++) {
    if ((*i)->contains(e))
      return (*i);
  }
  return NULL;
}

GList * EntityHandler::getGroupOptionMenuList() {
  GList * list = NULL;
  for (GroupList::iterator i = groups.begin(); i != groups.end(); ++i) {
    list = g_list_append(list, (*i)->getName());
  }

  return list;
}

GtkWidget * EntityHandler::getGroupOptionMenu(Entity * e) {

  GtkWidget * option_menu;
  GtkWidget * menu;
  GtkWidget * item;
  int count = 0;
  int selected_item = 0;

  option_menu = gtk_option_menu_new();
  menu = gtk_menu_new();

  EntityGroup * eg = getEntityGroupForEntity(e);

  for (GroupList::iterator i = groups.begin(); i != groups.end(); ++i) {
    item = gtk_menu_item_new_with_label((*i)->getName());
    gtk_widget_show(item);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu), item);

    if ((*i) == eg)
      selected_item = count;

    count++;
  }

  gtk_option_menu_set_menu(GTK_OPTION_MENU(option_menu), menu);
  gtk_option_menu_set_history(GTK_OPTION_MENU(option_menu), selected_item);

  return option_menu;
}
