#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gdk/gdkkeysyms.h>

#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif

#include "mudclient.h"
#include "callbacks.h"
#include "Prefs.h"
#include "Message.h"
#include "SystemColour.h"
#include "PluginHandler.h"
#include "ColouredLabel.h"
#include "GTKTwoVT.h"
#include "BaseWindow.h"

#include "CommandInterpreter.h"

#include "TeloptFilter.h"

extern PluginHandler * phandler;
extern CommandInterpreter * commandInterpreter;
extern BaseWindow * mainWindow;

#define ESCAPE '\033'
#define PAGE_UP 65365
#define PAGE_DOWN 65366

gboolean prompt_expose(GtkWidget *, GdkEventExpose *, gpointer);

gboolean gtk_two_vt_changed(gpointer * widget, gpointer data) {
  GTKTwoVT * vt = (GTKTwoVT *)data;

  vt->timeout();
  return true;
}

/*
  Then, you just need to know that the Editable widgets don't even to
  try to paste when not editable. So you need to connect to
  "button_press_event", and do a gtk_selection_convert() yourself.
  Then, you need to connect to "selection_received" to get the data when
  it arrives. Your "selection_received" handler then needs to
  do a gtk_signal_emit_stop_by_name(), so that the GtkEditable
  default selection_received handler doesn't get called.
*/

void papaya_selection_received(GtkWidget * widget, GtkSelectionData * selection_data, gpointer data) {

  GtkWidget * input = mainWindow->getCurrentConnection()->getVT()->getInput();

  if (selection_data->length < 0)
    return;

  char buf[16384];
  buf[0] = '\0';
  
  if (gtk_entry_get_text(GTK_ENTRY(input)))
    strcat(buf, gtk_entry_get_text(GTK_ENTRY(input))); 

  if (selection_data->type == gdk_atom_intern("STRING", FALSE)) {
    selection_data->data[selection_data->length] = 0;
    strcat(buf, (char *)selection_data->data);
  } else {
    if (selection_data->type == gdk_atom_intern("COMPOUND_TEXT", FALSE)
      || selection_data->type == gdk_atom_intern("TEXT", FALSE)) {
      gchar **list;
      gint count;
      gint i;
      
      count = gdk_text_property_to_text_list (selection_data->type,
					      selection_data->format, 
					      selection_data->data,
					      selection_data->length,
					      &list);
      for (i = 0; i < count; i++)
	if (list[i])
	  strcat(buf, list[i]);
    }
  }
 
  gtk_entry_set_text(GTK_ENTRY(input), buf);
  
}

void output_size_changed(GtkWidget * widget, GdkEvent * event, gpointer data) {

  Connection * c = (Connection *)data;

  // Tell this connection to redo NAWS.
  TeloptFilter * filter = static_cast<TeloptFilter*>(c->getSocket()->inputFilters.findFilter("TelnetOptions"));
  if (filter)
    filter->do_naws();

  //  VT * vt = c->getVT();
  //  vt->scroll();

}

gboolean output_button_pressed(GtkWidget * widget, GdkEventButton * event, gpointer data) {

  GdkAtom ctext_atom = gdk_atom_intern("COMPOUND_TEXT", FALSE);

  switch (event->type) {
  case GDK_BUTTON_PRESS:
    switch (event->button) {
    case 2:
      g_signal_connect_data(GTK_OBJECT(widget), "selection_received", GTK_SIGNAL_FUNC(papaya_selection_received), data, NULL, (GConnectFlags)0);
      
      gtk_selection_convert(widget, GDK_SELECTION_PRIMARY, ctext_atom, GDK_CURRENT_TIME);
      
      return true;
    }
    return false;

  case GDK_2BUTTON_PRESS:
    switch (event->button) {
    case 1:

      
      return false;
    }

  /* k0n pointed out that some combinations of GDK/C++ spew warnings if
   * the default option is not here.  (switch (enumeration) requires
   * a case statement for each item in the enumeration.)
   */
  default:
    return false;
  }

}


/// Start conversion here.




GTKTwoVT::GTKTwoVT(Connection * c) : VT(c) {

  scrolled_window = (GtkWidget *)NULL;
  text_view = (GtkWidget *)NULL;
  buffer = (GtkTextBuffer *)NULL;
  marker = NULL;

  filtered_scrolled_window = (GtkWidget *)NULL;
  filtered_text_view = (GtkWidget *)NULL;
  filtered_buffer = (GtkTextBuffer *)NULL;
  filtered_marker = NULL;

  input = (GtkWidget *)NULL;

  appended_to_filtered = false;
  appended_to_output = false;
  filtered_lines = 0;
  output_lines = 0;

  // Initialise the colour state variables.
  bold = false;
  flashing = false;
  fg = 0;
  bg = 7;

  foreground_color_default = TRUE;
  background_color_default = TRUE;

  papaya_colour = false;
  flashing_text = false;
  should_scroll = TRUE;

  initTextArea();
}

GTKTwoVT::~GTKTwoVT() {
  delete prompt;
}

#if 0
  // Nuke the text tags.
  GtkTextTagTable * tag_table = gtk_text_buffer_get_tag_table(GTK_TEXT_BUFFER(buffer));

  for (int i = 0; i < 16; i++) {
    char buf[1024];
    snprintf(buf, 1024, "fg_colour:%d", i);
    GtkTextTag * tag = gtk_text_tag_table_lookup(tag_table, buf);
    if (tag)
      gtk_text_tag_table_remove(tag_table, tag);
    else
      printf ("Unable to find tag %s.\n", buf);

    snprintf(buf, 1024, "bg_colour:%d", i);
    tag = gtk_text_tag_table_lookup(tag_table, buf);
    if (tag)
      gtk_text_tag_table_remove(tag_table, tag);
    else
      printf ("Unable to find tag %s.\n", buf);
  }

  GtkTextTag * tag = gtk_text_tag_table_lookup(tag_table, "bold_font");
  if (tag)
    gtk_text_tag_table_remove(tag_table, tag);
  else
    printf ("Unable to find tag bold_font.\n");

  g_free(buffer);
#endif

void GTKTwoVT::destroyFilteredWidget() {
  gtk_container_remove(GTK_CONTAINER(output_vbox), filtered_text_view);
  gtk_widget_destroy(filtered_text_view);
  filtered_text_view = (GtkWidget *)NULL;
  filtered_buffer = (GtkTextBuffer *)NULL;
}

void GTKTwoVT::createFilteredWidget() {

  filtered_scrolled_window = gtk_scrolled_window_new(NULL, NULL);
  gtk_widget_show(filtered_scrolled_window);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(filtered_scrolled_window),
				 GTK_POLICY_ALWAYS,
				 GTK_POLICY_ALWAYS);


  // Create the text view and assign the buffer pointer
  filtered_buffer = gtk_text_buffer_new(NULL);
  filtered_text_view = gtk_text_view_new_with_buffer(filtered_buffer);
  gtk_widget_show(filtered_text_view);

  gtk_container_add(GTK_CONTAINER(filtered_scrolled_window),
		    filtered_text_view);


  // Set the default font.
  PangoFontDescription *font_desc;
  font_desc = pango_font_description_from_string (conn->queryPreferences()->getOutputFont());

  gtk_widget_modify_font (filtered_text_view, font_desc);
  pango_font_description_free (font_desc);
  
  gtk_text_view_set_editable(GTK_TEXT_VIEW(filtered_text_view), false);
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(filtered_text_view), GTK_WRAP_WORD);
  gtk_box_pack_start(GTK_BOX(output_vbox), filtered_scrolled_window, false, false, 0);

  g_signal_connect_data(GTK_OBJECT(filtered_text_view), "button_press_event",
		     GTK_SIGNAL_FUNC(output_button_pressed),
		     input, NULL, (GConnectFlags)0);
}

void GTKTwoVT::initTextArea() {

  vbox = gtk_vbox_new(false, 2);

  // Create the text view and assign the buffer pointer
  buffer = gtk_text_buffer_new(NULL);

  scrolled_window = gtk_scrolled_window_new(NULL, NULL);
  gtk_widget_show(scrolled_window);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
				 GTK_POLICY_ALWAYS,
				 GTK_POLICY_ALWAYS);
  //  g_object_set_data_full(G_OBJECT(vbox), "scrolled_window", scrolled_window,
  //  			 (GDestroyNotify) gtk_widget_unref);

  text_view = gtk_text_view_new_with_buffer(buffer);
  gtk_widget_show(text_view);

  //  g_object_set_data_full(G_OBJECT(vbox), "text_view", text_view,
  //			 (GDestroyNotify) gtk_widget_unref );

  createTagTable();

  gtk_container_add(GTK_CONTAINER(scrolled_window), text_view);

  // Set the default font.
  PangoFontDescription *font_desc;
  font_desc = pango_font_description_from_string (conn->queryPreferences()->getOutputFont());

  gtk_widget_modify_font (text_view, font_desc);
  pango_font_description_free (font_desc);

  gtk_text_view_set_editable(GTK_TEXT_VIEW(text_view), false);
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(text_view), GTK_WRAP_WORD);
  
  output_vbox = gtk_vbox_new(false, 2);
  gtk_widget_show(output_vbox);

  gtk_box_pack_start(GTK_BOX(output_vbox), scrolled_window, TRUE, TRUE, 0);
  if (conn->queryPreferences()->getFilteredOutput()) {
    createFilteredWidget();
  }

  prompt = new ColouredLabel(conn, "");
  prompt->createWidget();
  gtk_widget_show(prompt->getWidget());
  
  input = gtk_entry_new();
  gtk_widget_show (input);

  font_desc = pango_font_description_from_string(conn->queryPreferences()->getInputFont());
  gtk_widget_modify_font(input, font_desc);
  pango_font_description_free(font_desc);
  
  g_signal_connect_data (input, "activate",
			 GTK_SIGNAL_FUNC (papaya_input_entry_activate),
			 this, NULL, (GConnectFlags)0);
  
  g_signal_connect_data (input,
			 "key_press_event",
			 GTK_SIGNAL_FUNC(papaya_keypress),
			 NULL,
			 NULL,
			 G_CONNECT_AFTER);
			   
  g_signal_connect_data (text_view,
			 "key_press_event",
			 GTK_SIGNAL_FUNC(papaya_keypress),
			 (gpointer)NULL, /*optional gpointer for sig func*/
			 NULL,
			 G_CONNECT_AFTER);

  GTK_WIDGET_UNSET_FLAGS( GTK_WIDGET(text_view), GTK_CAN_FOCUS );

  gtk_box_pack_start(GTK_BOX(vbox), output_vbox, TRUE, TRUE, 0);

  input_vbox = gtk_vbox_new(false, 2);
  gtk_widget_show(input_vbox);

  if (conn->queryPreferences()->getPromptOnInputLine()) {
    GtkWidget * pbox = gtk_hbox_new(false, 2);
    gtk_box_pack_start(GTK_BOX(pbox), prompt->getWidget(), false, false, 0);
    gtk_box_pack_start(GTK_BOX(pbox), input, TRUE, TRUE, 0);
    gtk_widget_show(pbox);
    
    gtk_box_pack_start(GTK_BOX(input_vbox), pbox, false, false, 0);
  } else {
    gtk_box_pack_start(GTK_BOX(input_vbox), prompt->getWidget(), false, false, 0);
    gtk_box_pack_start(GTK_BOX(input_vbox), input, false, false, 0);
  }

  gtk_box_pack_start(GTK_BOX(vbox), input_vbox, false, false, 0);

  // Why does this center the widget in the box?

  //  gtk_widget_show(hbox);

  gtk_container_add(GTK_CONTAINER(box), vbox);
  gtk_widget_show(vbox);

  g_signal_connect_data(GTK_OBJECT(text_view), "button_press_event",
			GTK_SIGNAL_FUNC(output_button_pressed),
			input, NULL, (GConnectFlags)0);

  // Need to connect up a signal for when the size of this window is changed.
  g_signal_connect_data(GTK_OBJECT(text_view), "size_allocate",
			GTK_SIGNAL_FUNC(output_size_changed), conn,
			NULL, (GConnectFlags)0);

}

void GTKTwoVT::hideInput() {
  gtk_widget_hide(input_vbox);
}

void GTKTwoVT::showInput() {
  gtk_widget_show(input_vbox);
}

char * GTKTwoVT::zapChars(char * text) {

  char * pc;

  if (*text == '\r')
    text++;

  pc = text;

  while (*pc != '\0') {

	  if (*pc == '\r') { // || (unsigned char)*pc > 127) {
      memmove(pc, pc + 1, strlen(pc + 1) + 1);
      // Don't increment pc, as we just moved the next char to there
      continue;
    }

    pc++;
  }

  return text;
}

/**
 * Any text to be considered for the display will be sent to this method.
 * This includes prompts, which the VT must remove if it wishes to display
 * them in a separate area.
 */

void GTKTwoVT::append(char * text) {
  char * pc;
  char * last_marker;
  
  snarfPrompt(text); // note: modifies text in place & sets prompt

  pc = zapChars(text);
  last_marker = pc;

  if (strlen(pc) > 0)
    mainWindow->nextText(conn);

  while ((pc = strchr(pc, ESCAPE))) {
    // NUL terminate the last_marker string.
    *pc = '\0';

    // Write all stuff between last_marker and pc to the VT.
    printVT(last_marker);
    pc = parseVTCode(pc);
    last_marker = pc;
  }

  printVT(last_marker);

  if (appended_to_filtered)
    filtered_lines++;
  if (appended_to_output)
    output_lines++;

  appended_to_filtered = false;
  appended_to_output = false;
}

char * GTKTwoVT::getColourCodes(char * s) {

  char * pc;
  static char buf[16384];
  int count = 0;

  while ((pc = strchr(s, ESCAPE))) {

    buf[count++] = *pc++; // ESCAPE
    buf[count++] = *pc++; // [
    while (*pc && !isalpha(*pc))
      buf[count++] = *pc++;

    if (*pc)
	buf[count++] = *pc++; // alpha terminating character

    s = pc;
  }

  buf[count] = '\0';

  return buf;
}

void GTKTwoVT::snarfPrompt(char * s)
{
    // Find last prompt-end marker
    char *marker = strrchr(s, MAGIC_PROMPT);
    if (!marker) // no eol to be seen, no changes needed.
	return;

    // Find last line-end before marker
    char *pc;
    for (pc = marker - 1; pc >= s; --pc)
	if (*pc == '\n')
	    break;

    ++pc; // point to start of line (or buffer if no line-end found)

    // Take a copy of the prompt.
    int prompt_len = marker - pc;
    char *prompt = new char[prompt_len + 1];
    memcpy(prompt, pc, prompt_len);
    prompt[prompt_len] = 0;

    setPrompt(prompt);

    // Now meddle with the output buffer.
    if (conn->queryPreferences()->getDoublePrompt())
    {
	delete[] prompt;
	return; // no changes needed.
    }

    // Replace prompt with only the color codes from it.
    strcpy(pc, getColourCodes(prompt)); // guaranteed to be smaller
    strcpy(pc+strlen(pc), marker+1);
    delete[] prompt;
}

/**
 * Sets the prompt to 'text' after stripping colour codes.
 */

void GTKTwoVT::setPrompt(char * text) {
  
  phandler->promptFilters(conn, text);
  if (strlen(text) == 0)
    return;

  char * pr = strdup(text); // Assume it isn't safe to edit text directly.
  prompt->setText(pr);
  free(pr);
}

/**
 * 'code' is a pointer to the ESCAPE char at the start of a VT100 code.
 * The returned pointer points at the first character after the complete
 * VT100 code.
 */

char * GTKTwoVT::parseVTCode(char * code) {

  char *pc = code+1;

  switch (*pc) {
  case '[':
    pc = parseBracketVTCode(pc);
    break;

  default:
    pc = parsePlainVTCode(pc);
    break;
  }

  pc++;
  return pc;
}

char * GTKTwoVT::parseBracketVTCode(char * pc) {

  char * orig = pc;

  while (!isalpha(*pc))
    pc++;

  switch (*pc) {

  case 'm': // Colour code
    setColour(orig);
    break;

  case 'p': // Internal Papaya colour code.
    setPapayaColour(orig);
    break;

  case 'q': // Reset Papaya colour.
    resetPapayaColour();

    // Cursor Control

  case 'H': // Set cursor to x;y
  case 'f': // Set cursor to x;y
    setPosition(orig);
    break;

  case 'A': // Move the cursor up x rows
    break;

  case 'B': // Move cursor down x rows
    break;

  case 'C': // Move cursor forward x columns
    break;

  case 's': // Save current current position
  case '7':
    break;

  case 'u': // Restore cursor position to last saved.
  case '8':
    break;

    // Device Status

  case 'c': // Requests a "Report Device Code" response
    break;

  case 'n': // Various device response stuff.
    break;

    // Terminal Setup
   
  case 'h': // Line wrap enabled (7h)
    break;

  case 'l': // Line wrap disabled (7l)
    break;

    // Scrolling

  case 'r': // Enable scrolling "x,y" = rows x to y.  No args = whole display
    break;

    // Tab Control

  case 'g': // Clears tab at the current position (3g = all tabs)
    break;

    // Erasing Text

  case 'K': // Various line based erases
    break;

  case 'J': // Various row based erases
    break;

    // Printing

  case 'i': // Printing related commands
    break;

  }

  // Now at the character that marks the start of normal text or the ESCAPE
  // for a new code.
  return pc;
}

char * GTKTwoVT::parsePlainVTCode(char * pc) {

  while (!isalpha(*pc))
    pc++;

  switch (*pc) {

    // Scrolling
    
  case 'D': // Scroll down one line
    break;
    
  case 'M': // Scroll up one line
    break;

    // Tab Control
    
  case 'H': // Sets a tab at the current position
    break;
    
    // Terminal Setup

  case 'c': // Reset all terminal settings to default
    break;

    // Fonts

  case ')': // Set default font
    break;

  case '(': // Use alternate font
    break;

  }


  pc++;
  return pc;
}

void GTKTwoVT::setPapayaColour(char * buf) {

  int red;
  int green;
  int blue;

  // allocate color

  char * pc;

  pc = strchr(buf, 'p');
  if (!pc)
    return;

  if (pc - buf != 12) { // [rrr;ggg;bbb
    printf(_("Dud Papaya Colour code."));
    printf(CRLF);
    return;
  }

  *pc = '\0';
  if (sscanf(buf, "[%d;%d;%d", &red, &green, &blue) != 3) {
    printf(_("PapayaColour: unable to sscanf."));
    printf(CRLF);
    return;
  }

  papaya_colour = true;

  papaya_colour_red = red;
  papaya_colour_green = green;
  papaya_colour_blue = blue;

  // When text of this colour is inserted, it should use an anonymous tag.

}

void GTKTwoVT::resetPapayaColour() {
  papaya_colour = false;
}

void GTKTwoVT::setColour(char * buf) {
  char * pc = buf;
  char buf2[10];
  int pos = 0;

  buf2[0] = '\0';
  
  papaya_colour = false;

  while (1) {

    // If we haven't got a [ or ; indicating more codes, leave the loop.
    if (!(*pc == ';' || *pc == '['))
      break;

    // Rip out the bit until the next non-alphanumeric character.
    pos = 0;

    buf2[pos] = *pc++;

    while (1) {
      if (isalpha(*pc) || *pc == ';')
        break;

      buf2[pos++] = *pc++;
    }
    
    buf2[pos] = '\0';
    setActualColour(buf2);
  }
}

void GTKTwoVT::setActualColour(char * buf) {
  if (!conn->queryPreferences()->getColour())
    return;

  if (buf[0] == '0' || strlen(buf) == 0) {

    // Set everything to defaults.
    background_color_default = TRUE;
    foreground_color_default = TRUE;

    bold = false;
    flashing = false;
    return;
  }

  if (buf[0] == '1') {
    bold = TRUE;
    return;
  }

  if (buf[0] == '5') {
    flashing = TRUE;
    return;
  }

  /* Look up which colour we're to use. */

  /* buf[2] should contain the colour. */

  if (strlen(buf) >= 2) {
    for (int i = 0; i < 8; i++) {

      struct colour_table * table = conn->queryPreferences()->getColourTable();

      if (table[i].colour_character == buf[1]) {

        if (buf[0] == '3') {
	  foreground_color_default = false;
          fg = i;
          return;
        }
        
        if (buf[0] == '4') {
          background_color_default = false;
          bg = i;
          return;
        }
        
        return;

      }
    }
  }
}

void GTKTwoVT::beep(char * buf) {
  char * b;
  while ((b = strchr(buf, '\007'))) {
    gdk_beep();
    buf = b + 1;
  }
}

void GTKTwoVT::printVT(char * text) {
  char * start = text;

  if (conn->queryPreferences()->getBeep())
    beep(start);

  if (flashing_text && (start[0] == '\n' || start[0] == '\r')) {
    start++;
    flashing_text = false;
  }

  insertText(start);
}

void GTKTwoVT::insertText(char * text) {

  // Figure out which buffer we're writing to.
  GtkTextBuffer * target_buffer;
  gboolean free_it = false;

  gchar * output;
  gsize read, written;
  GError * error = NULL;

  gchar * in_charset = conn->queryPreferences()->getPreference("charset");
  if (!in_charset)
    in_charset = "ISO-8859-1";

  output = g_convert(text, strlen(text), "UTF-8", in_charset,
					&read, &written, &error);

  if (output) {
    free_it = true;
    text = output;
  }

  if (flashing && conn->queryPreferences()->getFilteredOutput()) {
    if (!filtered_text_view) {
      createFilteredWidget();
      scroll();
    }

    target_buffer = filtered_buffer;
    appended_to_filtered = true;
  } else {
    target_buffer = buffer;
    appended_to_output = true;
  }

  GtkTextIter start, end;

  gtk_text_buffer_get_bounds(GTK_TEXT_BUFFER(buffer),
			     &start,
			     &end);

  marker = gtk_text_buffer_create_mark(buffer, "", &end, false);

  if (filtered_buffer) {
    filtered_marker = gtk_text_buffer_create_mark(filtered_buffer, "", &end, false);
  }

  if (!papaya_colour && foreground_color_default && background_color_default) {
    if (bold && conn->queryPreferences()->getUseBoldFont())
      gtk_text_buffer_insert_with_tags_by_name(GTK_TEXT_BUFFER(target_buffer),
					       &end,
					       text,
					       -1,
					       "foreground_default",
					       "background_default",
					       "bold_font",
					       NULL);
    else
      gtk_text_buffer_insert_with_tags_by_name(GTK_TEXT_BUFFER(target_buffer),
					       &end,
					       text,
					       -1,
					       "foreground_default",
					       "background_default",
					       NULL);

#ifndef WIN32
    // For some reason WIN32 doesn't want to let us free(text), even after g_convert alloced it.
    if (free_it)
      free(text);
#endif
    return;
  }


  // Determine the appropriate colour tags.
  char fg_tag[1024];
  char bg_tag[1024];

  // Is this a Papaya colour connection.
  if (papaya_colour) {

    snprintf(fg_tag, 1024, "papaya_colour:%d:%d:%d",
	     papaya_colour_red,
	     papaya_colour_green,
	     papaya_colour_blue);

    // Does a tag already exist by this name?
    if (!gtk_text_tag_table_lookup(gtk_text_buffer_get_tag_table(GTK_TEXT_BUFFER(buffer)), fg_tag)) {

      // Create a tag for this colour.
      GdkColor color;
      color.red = papaya_colour_red * 0xffff / 255;
      color.green = papaya_colour_green * 0xffff / 255;
      color.blue = papaya_colour_blue * 0xffff / 255;
      gdk_color_alloc(gdk_colormap_get_system(), &color);

      gtk_text_buffer_create_tag(GTK_TEXT_BUFFER(buffer),
				 fg_tag,
				 "foreground-gdk",
				 &color,
				 NULL);
    }
  } else
    snprintf(fg_tag, 1024, "fg_colour:%d", bold ? fg + 8 : fg);

  snprintf(bg_tag, 1024, "bg_colour:%d", bold ? bg + 8 : bg);

  // Default foreground, MUD set background.
  if (foreground_color_default && !papaya_colour) {
    if (bold && conn->queryPreferences()->getUseBoldFont())
      gtk_text_buffer_insert_with_tags_by_name(GTK_TEXT_BUFFER(target_buffer),
					       &end,
					       text,
					       -1,
					       "bold_font",
					       bg_tag,
					       NULL);
    else
      gtk_text_buffer_insert_with_tags_by_name(GTK_TEXT_BUFFER(target_buffer),
					       &end,
					       text,
					       -1,
					       bg_tag,
					       NULL);
#ifndef WIN32
    if (free_it)
      free(text);
#endif
    return;
  }

  papaya_colour = false;

    
  // Default background, MUD set foreground.
  if (background_color_default) {
    if (bold && conn->queryPreferences()->getUseBoldFont())
      gtk_text_buffer_insert_with_tags_by_name(GTK_TEXT_BUFFER(target_buffer),
					       &end,
					       text,
					       -1,
					       "bold_font",
					       fg_tag,
					       NULL);
    else
      gtk_text_buffer_insert_with_tags_by_name(GTK_TEXT_BUFFER(target_buffer),
					       &end,
					       text,
					       -1,
					       fg_tag,
					       NULL);
#ifndef WIN32
    if (free_it)
      free(text);
#endif
    return;
  }

  // MUD set foreground and background.
  if (bold && conn->queryPreferences()->getUseBoldFont())
    gtk_text_buffer_insert_with_tags_by_name(GTK_TEXT_BUFFER(target_buffer),
					     &end,
					     text,
					     -1,
					     "bold_font",
					     fg_tag,
					     bg_tag,
					     NULL);
  else
    gtk_text_buffer_insert_with_tags_by_name(GTK_TEXT_BUFFER(target_buffer),
					     &end,
					     text,
					     -1,
					     fg_tag,
					     bg_tag,
					     NULL);

#ifndef WIN32
  if (free_it)
    free(text);
#endif
}

void GTKTwoVT::timeout() {

  bool scr = doScroll();

  if (!scr) {
    return;
  }

  if (text_view && marker) {
    gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(text_view), marker, 0.0, true, 0.0, 1.0);
    gtk_text_buffer_delete_mark(buffer, marker);
    marker = NULL;
  }

  if (filtered_text_view && filtered_marker) {
    gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(filtered_text_view), filtered_marker, 0.0, true,
				 0.0, 1.0);
    gtk_text_buffer_delete_mark(filtered_buffer, filtered_marker);
    filtered_marker = NULL;
  }
}

void GTKTwoVT::scroll() {

#if 0

  // Trim the amount of text to the maximum number of characters.
  if (conn->queryPreferences()->getUseMaxLines()
      && conn->queryPreferences()->getMaxLines() > 0) {


    int lines = gtk_text_buffer_get_line_count(GTK_TEXT_BUFFER(buffer));
    int remove_output = lines - conn->queryPreferences()->getMaxLines();

    // @@ Remove the first remove_output lines from the text buffer.


    // Are there any lines to remove.
    if (remove_output > 0) {
      gtk_text_freeze(GTK_TEXT(output));
      
      while (remove_output > 0) {
	// Remove the first line from the box.
	char * text = gtk_editable_get_chars(GTK_EDITABLE(output), 0, -1);
	// gtk_editable_get_chars does a strdup ...
	if (text) {
	  char * pc = strchr(text, '\n');
	  gtk_text_set_point(GTK_TEXT(output), 0);
	  gtk_text_forward_delete(GTK_TEXT(output), pc - text + 1);
	  output_lines--;
	  remove_output--;
	  free(text);
	}
      }
      gtk_text_set_point(GTK_TEXT(output), gtk_text_get_length(GTK_TEXT(output)));
      gtk_text_thaw(GTK_TEXT(output));
    }

    if (filtered_output) {
      int remove_filtered = filtered_lines - conn->queryPreferences()->getMaxLines();
      // Are there any lines to remove.
      if (remove_filtered > 0) {
	gtk_text_freeze(GTK_TEXT(filtered_output));
	
	while (remove_filtered > 0) {
	  
	  // Remove the first line from the box.
	  char * text = gtk_editable_get_chars(GTK_EDITABLE(output), 0, -1);
	  if (text) {
	    char * pc = strchr(text, '\n');
	    gtk_text_set_point(GTK_TEXT(filtered_output), 0);
	    gtk_text_forward_delete(GTK_TEXT(filtered_output), pc - text + 1);
	    filtered_lines--;
	    remove_filtered--;
	  }
	}
	gtk_text_set_point(GTK_TEXT(filtered_output), gtk_text_get_length(GTK_TEXT(filtered_output)));
	gtk_text_thaw(GTK_TEXT(filtered_output));
      }
    }
  }

#endif

  // Attempt to scroll text onscreen.
  timeout();

}

bool GTKTwoVT::doScroll() {
  if (conn->queryPreferences()->getScrollOnInput())
    return TRUE;

  return should_scroll;
}

/**
 * This function is called before text is inserted to determine
 * whether or not we should scroll to bottom after insertion.
 */

void GTKTwoVT::preInsert() {
  /*
  GtkAdjustment * adj = GTK_RANGE(vscrollbar)->adjustment;
  
  if (adj->value + adj->page_size == adj->upper)
    should_scroll = TRUE;
  else
    should_scroll = false;
  */
  should_scroll = true;
}

void GTKTwoVT::use() {

  GdkColor background_colour;

  background_colour.red = conn->queryPreferences()->getScreenBackgroundRed();
  background_colour.green = conn->queryPreferences()->getScreenBackgroundGreen();
  background_colour.blue = conn->queryPreferences()->getScreenBackgroundBlue();

  gdk_color_alloc(gdk_colormap_get_system(), &background_colour);

  GdkWindow * window = gtk_text_view_get_window(GTK_TEXT_VIEW(text_view),
						GTK_TEXT_WINDOW_TEXT);
  gdk_window_set_background(window, &background_colour);

  if (filtered_text_view) {
    window = gtk_text_view_get_window(GTK_TEXT_VIEW(filtered_text_view),
				      GTK_TEXT_WINDOW_TEXT);
    gdk_window_set_background(window, &background_colour);
  }
}

GtkWidget * GTKTwoVT::getInput() {
  return input;
}

GtkWidget * GTKTwoVT::getOutput() {
  return text_view;
}

GtkWidget * GTKTwoVT::getBox() {
  return vbox;
}

GtkWidget * GTKTwoVT::getPrompt() {
  return prompt->getWidget();
}

int GTKTwoVT::width() {
  int width, height;
  getScreenSize(&width, &height);
  return width;
}

int GTKTwoVT::height() {
  int width, height;
  getScreenSize(&width, &height);
  return height;
}

void GTKTwoVT::getScreenSize(int * width, int * height) {
  //  int h, w;
  //  int ww, wh;

  // @@ Rewrite this to use Pango
#if 0
  h = conn->queryPreferences()->getOutput()->ascent + conn->queryPreferences()->getOutput()->descent;
  w = gdk_char_width(conn->queryPreferences()->getOutput(), 'W');

  ww = output->allocation.width;
  wh = output->allocation.height;

  *width = (ww/w);
  *height = (wh/h);
#endif
}

void GTKTwoVT::setPosition(char * text) {
  
  // Returning immediately since this support is nowhere near complete.
  return;

  int row, column;
  char first[10];
  char second[10];
  int fc = 0, sc = 0;

  // Get past the [
  text++;

  while (isdigit(*text)) {
    first[fc++] = *text++;
  }
  first[fc] = '\0';

  // Get past the ;
  text++;

  while (isdigit(*text)) {
    second[sc++] = *text++;
  }
  second[sc] = '\0';

  row = atoi(first);
  column = atoi(second);

  setPosition(row, column);
}

void GTKTwoVT::setPosition(int r, int c) {
#if 0
  int rows, columns;
  getScreenSize(&columns, &rows);

  if (r > rows || c > columns) {
    return;
  }

  // Retrieve the text from the text box.
  char * text = gtk_editable_get_chars(GTK_EDITABLE(output), 0, -1);

  char * ptr = text + strlen(text);

  // Need to figure out which line is the right one.
  for (int i = rows; i >= 0; i--) {

    // Move to the char after '\n';
    ptr = findStartOfLine(ptr, text);

    if (rows - i == r) {

      char * eol = ptr;
      while (*eol != '\n')
	eol++;
      eol--; // Back off the \n

      while (1) {
	if (c <= eol - ptr) {
	  gtk_text_set_point(GTK_TEXT(output), (ptr - text) + c);
	  return;
	}

	gtk_text_set_point(GTK_TEXT(output), (int)(eol - text));
	insertText(" ");
	eol++;
      }
      
      // We might be off by one at this point.
      return;
    }
  }
#endif
}

char * GTKTwoVT::findStartOfLine(char * ptr, char * limit) {

  ptr--;
  ptr--;

  while (*ptr != '\n' && ptr >= limit)
    ptr--;

  ptr++; // Go past the \n

  if (ptr < limit)
    ptr = limit;

  return ptr;
}

void GTKTwoVT::moveLeft() {
}

void GTKTwoVT::moveRight() {
}

void GTKTwoVT::setInput(char * s) {
  gtk_entry_set_text(GTK_ENTRY(input), s);
  gtk_editable_set_position(GTK_EDITABLE(input), -1);
}

/**
 * The preferences have been updated, so let's check the settings.
 */

void GTKTwoVT::prefsUpdated() {

  // Update the colour and font tags.
  createTagTable();

  // Change the font on the font input

  PangoFontDescription * font = pango_font_description_from_string(conn->queryPreferences()->getInputFont());
  gtk_widget_modify_font(getInput(), font);
  pango_font_description_free(font);

  use();

  // Change where flashing text goes.
  if (conn->queryPreferences()->getFilteredOutput() && !filtered_text_view) {
    createFilteredWidget();
    scroll();
    return;
  }

  if (!conn->queryPreferences()->getFilteredOutput() && filtered_text_view) {
    destroyFilteredWidget();
    scroll();
    return;
  }
}

void GTKTwoVT::keypressEvent(GtkWidget * widget, GdkEventKey * event) {
  papaya_keypress(widget, event);

  // We need to detect which widget currently has focus.  If it's not the
  // input, then we need to switch focus and append the character.

  if (GTK_WIDGET_HAS_FOCUS(widget))
    return;

  focus();

  // append the character.
  gchar * current_input = (gchar *)gtk_entry_get_text(GTK_ENTRY(getInput()));
  gchar * new_input = (gchar *)g_malloc(strlen(current_input) + 2);
  snprintf(new_input, strlen(current_input) + 2, "%s%c", current_input, event->keyval);
  setInput(new_input);
  g_free(new_input);

}

int papaya_keypress(GtkWidget * widget, GdkEventKey * event) {

  Connection * c = mainWindow->getCurrentConnection();
  if (!c)
    return 0;

  /* If this is line mode, check for command history. */
  if (c->getVT()->getMode() == VTLineMode) {
    switch(event->keyval) {

    case P:
      if (!(event->state & GDK_CONTROL_MASK))
	break;
    case UP_ARROW:
      c->getVT()->setInput(c->getHistory()->getPrev());
      return 1;

    case N:
      if (!(event->state & GDK_CONTROL_MASK))
	break;
    case DOWN_ARROW:
      c->getVT()->setInput(c->getHistory()->getNext());
      return 1;

    /* Check for scroll up/down. */
  case PAGE_UP:
    c->getVT()->scrollUp();
    break;
  case PAGE_DOWN:
    c->getVT()->scrollDown();
    }

    return 1;
  }

  if (event->keyval == ENTER || event->keyval == KP_ENTER) {
    c->getSocket()->write("\r\0", 2);
    return 1;
  }

  if (event->string && strlen(event->string) > 0) {
    c->getSocket()->write(event->string, strlen(event->string));
    return 1;
  }

  return 0;
}

int papaya_input_entry_activate(GtkEditable * editable, gpointer data) {

  Connection * conn = mainWindow->getCurrentConnection();
  if (!conn)
    return 0;

  if (conn->getVT()->getMode() == VTCharacterMode) {
    conn->getVT()->setInput("");
    return 1;
  }

  GTKTwoVT * pvt = (GTKTwoVT *)conn->getVT();

  gchar * text = (gchar *)gtk_entry_get_text(GTK_ENTRY(pvt->getInput()));

  if (!commandInterpreter)
    commandInterpreter = new CommandInterpreter();

  commandInterpreter->interpret(conn, text);
  return 1;
}

void GTKTwoVT::setMode(VTMode m) {

  VT::setMode(m);

  if (getMode() == VTCharacterMode)
    hideInput();
  
  if (getMode() == VTLineMode)
    showInput();

  focus();
}

void GTKTwoVT::focus() {
  switch (getMode()) {
  case VTCharacterMode:
    gtk_widget_grab_focus(text_view);
    break;
  case VTLineMode:
    gtk_widget_grab_focus(input);
    break;
  }
}

void GTKTwoVT::hideCommands() {
  gtk_entry_set_visibility(GTK_ENTRY(input), false);
}

void GTKTwoVT::showCommands() {
  gtk_entry_set_visibility(GTK_ENTRY(input), true);
}

bool GTKTwoVT::showingCommands() {
  return GTK_WIDGET_VISIBLE(GTK_ENTRY(input)->editable);
}

void GTKTwoVT::scrollUp() {
  gint x, y;
  GtkTextIter iter;

  gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(text_view),
					gtk_text_view_get_window_type(GTK_TEXT_VIEW(text_view), text_view->window),
					0, 0,
					&x, &y);
  
  gtk_text_view_get_iter_at_location(GTK_TEXT_VIEW(text_view), &iter, x, y);

  gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(text_view), &iter, false, true,
			       0.0, 1.0);
}

void GTKTwoVT::scrollDown() {

  gint x, y;
  GtkTextIter iter;

  gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(text_view),
					gtk_text_view_get_window_type(GTK_TEXT_VIEW(text_view), text_view->window),
					text_view->allocation.width, text_view->allocation.height,
					&x, &y);

  gtk_text_view_get_iter_at_location(GTK_TEXT_VIEW(text_view), &iter, x, y);
  gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(text_view), &iter, false, true,
			       0.0, 0.0);
}

void GTKTwoVT::createTagTable() {
  int i;

  // Create the tag table for this connection.
  
  GtkTextTagTable * tag_table = gtk_text_buffer_get_tag_table(GTK_TEXT_BUFFER(buffer));
  if (!tag_table)
    abort();
  
  struct colour_table * table = conn->queryPreferences()->getColourTable();

  // Create a foreground_default tag.

  GtkTextTag * tag = gtk_text_tag_table_lookup(tag_table, "foreground_default");
  int index = conn->queryPreferences()->getDefaultForegroundColour();
  GdkColor color;

  if (index == -1) {
    color.red = conn->queryPreferences()->getScreenBackgroundRed();
    color.blue = conn->queryPreferences()->getScreenBackgroundBlue();
    color.green = conn->queryPreferences()->getScreenBackgroundGreen();
  } else {
    color.red = table[index].red;
    color.green = table[index].green;
    color.blue = table[index].blue;
  }
  
  if (tag)
    g_object_set(tag, "foreground-gdk", &color, NULL);
  else
    gtk_text_buffer_create_tag(GTK_TEXT_BUFFER(buffer),
			       "foreground_default",
			       "foreground-gdk",
			       &color,
			       NULL);
  
  tag = gtk_text_tag_table_lookup(tag_table, "background_default");
  index = conn->queryPreferences()->getDefaultBackgroundColour();
  
  if (index == -1) {
    color.red = conn->queryPreferences()->getScreenBackgroundRed();
    color.blue = conn->queryPreferences()->getScreenBackgroundBlue();
    color.green = conn->queryPreferences()->getScreenBackgroundGreen();
  } else {
    color.red = table[index].red;
    color.green = table[index].green;
    color.blue = table[index].blue;
  }

  if (tag)
    g_object_set(tag, "background-gdk", &color, NULL);
  else
    gtk_text_buffer_create_tag(GTK_TEXT_BUFFER(buffer),
			       "background_default",
			       "background-gdk",
			       &color,
			       NULL);

  // Create a tag for each of the foreground colours.
  for (i = 0; i < 16; i++) {
    char buf[1024];

    snprintf(buf, 1024, "fg_colour:%d", i);

    // Retrieve and allocate the colour.
    GdkColor color;
    color.red = table[i].red;
    color.green = table[i].green;
    color.blue = table[i].blue;
    gdk_color_alloc(gdk_colormap_get_system(), &color);

    tag = gtk_text_tag_table_lookup(tag_table, buf);
    if (tag)
      g_object_set(tag, "foreground-gdk", &color, NULL);
    else
      gtk_text_buffer_create_tag(GTK_TEXT_BUFFER(buffer),
				 buf,
				 "foreground-gdk",
				 &color,
				 NULL);
  }

  // Create a tag for each of the background colours.
  for (i = 0; i < 16; i++) {
    char buf[1024];

    snprintf(buf, 1024, "bg_colour:%d", i);

    // Retrieve and allocate the colour.
    GdkColor color;
    color.red = table[i].red;
    color.green = table[i].green;
    color.blue = table[i].blue;
    gdk_color_alloc(gdk_colormap_get_system(), &color);

    tag = gtk_text_tag_table_lookup(tag_table, buf);
    if (tag)
      g_object_set(tag, "background-gdk", &color, NULL);
    else
      gtk_text_buffer_create_tag(GTK_TEXT_BUFFER(buffer),
				 buf,
				 "background-gdk",
				 &color,
				 NULL);
  }

  PangoFontDescription * font = pango_font_description_from_string(conn->queryPreferences()->getOutputFont());

  gtk_widget_modify_font(text_view, font);
  pango_font_description_free(font);

  // Create the bold font tag.
  if (conn->queryPreferences()->getUseBoldFont())
    font = pango_font_description_from_string(conn->queryPreferences()->getBoldFont());
  else
    font = pango_font_description_from_string(conn->queryPreferences()->getOutputFont());

  tag = gtk_text_tag_table_lookup(tag_table, "bold_font");
  if (tag)
    g_object_set(tag, "font-desc", font, NULL);
  else
    gtk_text_buffer_create_tag(GTK_TEXT_BUFFER(buffer),
			       "bold_font",
			       "font-desc",
			       font,
			       NULL);
  pango_font_description_free(font);

  return;
}
