#include "system.h"

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/stat.h>
#ifndef WIN32
// @@ This isn't available under cygwin32
//#include <gdk/gdkx.h>
#include <unistd.h>
#endif
#include <string.h>
#include <stdlib.h>

#include "system.h"

#include <gdk/gdkkeysyms.h>

#include "mudclient.h"
#include "callbacks.h"
#include "Prefs.h"
#include "support.h"
#include "libusers.h"

#include "GTKTwoWindow.h"
#include "PluginHandler.h"

extern Prefs * globalPreferences;
extern BaseWindow * mainWindow;
extern PluginHandler * phandler;

GTKTwoWindow::GTKTwoWindow() {
  mud_menu = NULL;
  window = NULL;
  createInterface();
}

GTKTwoWindow::~GTKTwoWindow() {
  int width, height;

  width = window->allocation.width;
  height = window->allocation.height;

  globalPreferences->setWindowX(width);
  globalPreferences->setWindowY(height);

  globalPreferences->save((MUD *)NULL);

  gtk_widget_hide(window);
  gtk_widget_destroy(window);
}

int papaya_exit() {
	gtk_main_quit();
	return 1;
}

#ifdef USE_GTK
static GtkItemFactoryEntry menu_items[] = {
  { "/_File",            NULL,         0,                     0, "<Branch>" },
  { "/File/tearoff1",    NULL,         NULL,                  0, "<Tearoff>" },
  { "/File/Restart Papaya", NULL,      (GtkItemFactoryCallback)on_copyover_button_clicked, 0 },
  { "/File/sep1",        NULL,         NULL,                  0, "<Separator>" },
  { "/File/_Quit",       "<control>Q", (GtkItemFactoryCallback)papaya_exit,       0 },

  { "/_Connection",      NULL,         0, 0, "<Branch>" },
  { "/Connection/tearoff1", NULL, NULL, 0, "<Tearoff>" },
  { "/Connection/Connection Wizard...", "<control>W", (GtkItemFactoryCallback)on_connection_wizard_button_clicked, 0 },
  { "/Connection/Connect to...", "<control>C", (GtkItemFactoryCallback)on_connect_button_clicked, 0 },
  { "/Connection/Disconnect", "<control>D", (GtkItemFactoryCallback)on_disconnect_button_clicked, 0 },
  { "/Connection/Reconnect", "<control>R", (GtkItemFactoryCallback)on_reconnect_button_clicked, 0},
  { "/Connection/MUD List...", "<control>L", (GtkItemFactoryCallback)on_mudlist_button_clicked, 0},

  { "/_Settings", NULL, 0, 0, "<Branch>" },
  { "/Settings/tearoff1", NULL, NULL, 0, "<Tearoff>" },
  { "/Settings/_Preferences...", NULL, (GtkItemFactoryCallback)on_preferences_button_clicked, 0 },
  { "/Settings/Aliases and Triggers...", NULL, (GtkItemFactoryCallback)on_entities_button_clicked, 0 },

  { "/_Help", NULL, 0, 0, "<LastBranch>" },
  { "/Help/tearoff1", NULL, NULL, 0, "<Tearoff>" },
  { "/Help/User Guide", NULL, (GtkItemFactoryCallback)on_help_userguide_clicked, 0 },
  { "/Help/Plugins", NULL, (GtkItemFactoryCallback)on_help_plugins_clicked, 0 },
  { "/Help/_About", NULL, (GtkItemFactoryCallback)on_about_clicked, 0 },
};
#else
static GtkItemFactoryEntry menu_items[] = {
  { "/_File",            NULL,         0,                     0, "<Branch>" },
  { "/File/tearoff1",    NULL,         NULL,                  0, "<Tearoff>" },
  { "/File/Restart Papaya", NULL,      (GtkItemFactoryCallback)on_copyover_button_clicked, 0, "<StockItem>", GTK_STOCK_REFRESH },
  { "/File/sep1",        NULL,         NULL,                  0, "<Separator>" },
  { "/File/_Quit",       "<control>Q", (GtkItemFactoryCallback)papaya_exit,       0, "<StockItem>", GTK_STOCK_QUIT },

  { "/_Connection",      NULL,         0, 0, "<Branch>" },
  { "/Connection/tearoff1", NULL, NULL, 0, "<Tearoff>" },
  { "/Connection/Connection Wizard...", "<control>W", (GtkItemFactoryCallback)on_connection_wizard_button_clicked, 0 },
  { "/Connection/Connect to...", "<control>C", (GtkItemFactoryCallback)on_connect_button_clicked, 0 },
  { "/Connection/Disconnect", "<control>D", (GtkItemFactoryCallback)on_disconnect_button_clicked, 0 },
  { "/Connection/Reconnect", "<control>R", (GtkItemFactoryCallback)on_reconnect_button_clicked, 0},
  { "/Connection/MUD List...", "<control>L", (GtkItemFactoryCallback)on_mudlist_button_clicked, 0},

  { "/_Settings", NULL, 0, 0, "<Branch>" },
  { "/Settings/tearoff1", NULL, NULL, 0, "<Tearoff>" },
  { "/Settings/_Preferences...", NULL, (GtkItemFactoryCallback)on_preferences_button_clicked, 0, "<StockItem>", GTK_STOCK_PREFERENCES },
  { "/Settings/Aliases and Triggers...", NULL, (GtkItemFactoryCallback)on_entities_button_clicked, 0 },

  { "/_Help", NULL, 0, 0, "<LastBranch>" },
  { "/Help/tearoff1", NULL, NULL, 0, "<Tearoff>" },
  { "/Help/User Guide", NULL, (GtkItemFactoryCallback)on_help_userguide_clicked, 0 },
  { "/Help/Plugins", NULL, (GtkItemFactoryCallback)on_help_plugins_clicked, 0 },
  { "/Help/_About", NULL, (GtkItemFactoryCallback)on_about_clicked, 0 },
};
#endif

void GTKTwoWindow::createInterface() {
  char buf[1024];

  // Create the main application window
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
#ifdef USE_GTK_2
  g_object_set_data(G_OBJECT(window), "window", window);
  gtk_window_set_resizable(GTK_WINDOW(window), true);
#else
  gtk_window_set_policy(GTK_WINDOW(window), true, true, true);
#endif

  snprintf(buf, 1024, _("%s version %s (%s)"), CLIENT, VERSION, AUTHOR);
  gtk_window_set_title (GTK_WINDOW (window), buf);
 
#ifdef USE_GTK_2
  /* NULL window variable when window is closed */
  g_signal_connect (G_OBJECT (window), "destroy",
                    G_CALLBACK (gtk_widget_destroyed),
                    &window);
#else
  gtk_signal_connect(GTK_OBJECT(window), "destroy",
		     GTK_SIGNAL_FUNC(gtk_widget_destroyed), NULL);
#endif

  // Create a table to stick everything into
  GtkWidget * table = gtk_table_new(1, 4, FALSE);
  gtk_container_add(GTK_CONTAINER(window), table);
  //  g_object_set_data_full(G_OBJECT(window), "window", table,
  //                         (GDestroyNotify) g_object_unref);
  gtk_widget_show(table);
  
  item_factory;
  GtkAccelGroup *accel_group;
  gint nmenu_items = sizeof (menu_items) / sizeof (menu_items[0]);

  accel_group = gtk_accel_group_new ();
  item_factory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<main>", 
                                       accel_group);

  /* This function generates the menu items. Pass the item factory,
     the number of items in the array, the array itself, and any
     callback data for the the menu items. */
  gtk_item_factory_create_items (item_factory, nmenu_items, menu_items, NULL);

#ifndef WIN32
  /* Attach the new accelerator group to the window. */
  gtk_window_add_accel_group (GTK_WINDOW (window), accel_group);
#endif

  gtk_table_attach (GTK_TABLE (table),
                    gtk_item_factory_get_widget (item_factory, "<main>"),
                    0, 1,
                    0, 1,
                    (GtkAttachOptions)(GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions)0,
                    0, 0);

  gtk_widget_show(gtk_item_factory_get_widget(item_factory, "<main>"));

  // Create a toolbar
#ifdef USE_GTK_2
  toolbar = gtk_toolbar_new ();
  gtk_toolbar_set_icon_size (GTK_TOOLBAR (toolbar),
                             GTK_ICON_SIZE_SMALL_TOOLBAR);
#else
  toolbar = gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_BOTH);
#endif

  if (globalPreferences->getUseToolbar())
    gtk_widget_show(toolbar);
  //  g_object_set_data_full(G_OBJECT(window), "toolbar", toolbar,
  //                         (GDestroyNotify) g_object_unref);

#ifdef USE_GTK_2
  gtk_toolbar_insert_stock (GTK_TOOLBAR (toolbar),
                            GTK_STOCK_PREFERENCES,
                            _("Edit Preferences"),
                            NULL,
                            G_CALLBACK (on_preferences_button_clicked),
                            window, /* user data for callback */
                            -1);  /* -1 means "append" */
#else
  gtk_toolbar_append_item (GTK_TOOLBAR(toolbar),
			   _("Edit Preferences"),
			   _("Edit the current connection's preferences.  If there is no connection present, the global preferences will be edited."),
			   NULL,
			   NULL,
			   GTK_SIGNAL_FUNC(on_preferences_button_clicked),
			   window);
#endif


  gtk_table_attach (GTK_TABLE (table),
                    toolbar,
                    0, 1,
                    1, 2,
                    (GtkAttachOptions)(GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions)(GTK_FILL),
                    0, 0);

  splash_update("Loading Papaya logo");

  background_pixmap = create_pixmap (window, "papaya-logo-newbie.xpm");
  gtk_widget_ref (background_pixmap);
  //  g_object_set_data_full (G_OBJECT (window), "background_pixmap", background_pixmap,
  //			  (GDestroyNotify) g_object_unref);
  gtk_widget_show (background_pixmap);
  splash_increment();

  // attach to the table.
  gtk_table_attach(GTK_TABLE(table),
                   background_pixmap,
                   0, 1,
                   2, 3,
                   (GtkAttachOptions)(GTK_EXPAND | GTK_FILL),
                   (GtkAttachOptions)(GTK_EXPAND | GTK_FILL),
                   0, 0);

  notebook = gtk_notebook_new ();
  //  g_object_set_data_full (G_OBJECT (window), "notebook", notebook,
  //			  (GDestroyNotify) g_object_unref);
  GTK_WIDGET_UNSET_FLAGS(GTK_WIDGET(notebook), GTK_CAN_FOCUS);
  
  // attach to the table
  gtk_table_attach(GTK_TABLE(table),
                   notebook,
                   0, 1,
                   2, 3,
                   (GtkAttachOptions)(GTK_EXPAND | GTK_FILL),
                   (GtkAttachOptions)(GTK_EXPAND | GTK_FILL),
                   0, 0);

  // Connect the notebook signals
#ifdef USE_GTK_2
  printf ("Attaching switch-page signal to notebook for gtk 2.\n");
  g_signal_connect_data(G_OBJECT(notebook), "switch-page",
			GTK_SIGNAL_FUNC(on_notebook_page_switch),
			NULL, NULL, (GConnectFlags)0);
#else
  gtk_signal_connect(GTK_OBJECT(notebook), "switch-page",
		     GTK_SIGNAL_FUNC(on_notebook_page_switch),
		     NULL);
#endif

  // Create the statusbar
  statusbar = gtk_statusbar_new ();
  //  g_object_set_data_full(G_OBJECT(window), "statusbar", statusbar,
  //                         (GDestroyNotify) g_object_unref);
  gtk_widget_show(statusbar);

  gtk_table_attach (GTK_TABLE (table),
                    statusbar,
                    0, 1,
                    3, 4,
                    (GtkAttachOptions)(GTK_EXPAND | GTK_FILL),
                    (GtkAttachOptions)(GTK_FILL),
                    0, 0);

#ifdef USE_GTK_2
  // Connect the window events
  g_signal_connect_data(G_OBJECT(window), "key_press_event",
			GTK_SIGNAL_FUNC(on_window_keypress),
			NULL, NULL, (GConnectFlags)0);
  g_signal_connect_data(G_OBJECT(window), "delete_event",
			GTK_SIGNAL_FUNC(gtk_main_quit),
			NULL, NULL, (GConnectFlags)0);
#else
  gtk_signal_connect(GTK_OBJECT(window), "key_press_event",
		     GTK_SIGNAL_FUNC(on_window_keypress), NULL);
  gtk_signal_connect(GTK_OBJECT(window), "delete_event",
		     GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
#endif

  gtk_window_set_default_size(GTK_WINDOW(window), globalPreferences->getWindowX(), globalPreferences->getWindowY());

  configureUsers();
  
}

GtkItemFactory * GTKTwoWindow::getItemFactory() {
	return item_factory;
}

void GTKTwoWindow::display() {
  gtk_widget_show(window);
}

void GTKTwoWindow::setToolbarVisible(bool t) {
  if (t)
    gtk_widget_show(toolbar);
  else
    gtk_widget_hide(toolbar);
}

/**
 * Add a connection to the main interface.
 */

void GTKTwoWindow::addConnection(Connection * c) {
  int n = 1;
  char name[1024];

  // Figure out if this connection needs a different name...

  gtk_widget_hide(background_pixmap);
  gtk_widget_show(notebook);

  sprintf(name, "%s", c->getName());

  while (findConnection(name)) {
    sprintf(name, "%s (%d)", c->getName(), n++);
  }

  c->setName(name);

  gtk_widget_show(c->getVT()->getMainWindow());
  GtkWidget * label = gtk_label_new(c->getName());
  gtk_widget_show(label);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), c->getVT()->getMainWindow(), label);

  c->setNext(first_connection);
  first_connection = c;

  // Let us bring this to the front of the notebook.

  int page = gtk_notebook_page_num(GTK_NOTEBOOK(notebook), c->getVT()->getMainWindow());
#ifdef USE_GTK
  gtk_notebook_set_page(GTK_NOTEBOOK(notebook), page);
#else
  gtk_notebook_set_current_page(GTK_NOTEBOOK(notebook), page);
#endif
}

/**
 * Remove a connection from the main interface.
 */

void GTKTwoWindow::removeConnection(Connection * c) {
  // Find which page this connection's VT is situated on, and delete it.
  gint page = gtk_notebook_page_num(GTK_NOTEBOOK(notebook), c->getVT()->getMainWindow());
  
  gtk_notebook_remove_page(GTK_NOTEBOOK(notebook), page);

  if (numPages() == 0) {
	  gtk_widget_hide(notebook);
	  gtk_widget_show(background_pixmap);
  }

  if (first_connection == c) {
    first_connection = c->getNext();
    return;
  }
  
  for (Connection * tmp = first_connection; tmp; tmp = tmp->getNext()) {
    if (tmp->getNext() == c) {
      tmp->setNext(c->getNext());
      c->setNext((Connection *)NULL);
      return;
    }
  }

}

/**
 * Returns the Connection object associated with the currently open page.
 */

Connection * GTKTwoWindow::getCurrentConnection() {

  gint page = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));

  if (page == -1)
    return NULL;

  GtkWidget * label = gtk_notebook_get_tab_label(GTK_NOTEBOOK(notebook), gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), page));


#ifdef USE_GTK
  gchar * text;
  gtk_label_get(GTK_LABEL(label), &text);
#else
  const gchar * text;
  text = gtk_label_get_text(GTK_LABEL(label));
#endif
  // Search through all the connection objects until we find a host with
  // this name.

  if (!strcmp(text, "mudFTP"))
    return NULL;

  for (Connection * tmp = first_connection; tmp; tmp = tmp->getNext()) {
    if (!strcmp(tmp->getName(), text)) {
      return tmp;
    }
  }

  return (Connection *)NULL;
}

/**
 * Highlights a the tab red if new text appears and this is not the current
 * connection.
 */

void GTKTwoWindow::nextText(Connection * c) {

  // Figure out which page this connection belongs to...
  int page = gtk_notebook_page_num(GTK_NOTEBOOK(notebook), c->getVT()->getMainWindow());
  int curr_page = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
  
  if (page != curr_page) {
    
    GtkWidget * label = gtk_notebook_get_tab_label(GTK_NOTEBOOK(notebook), gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), page));

    GtkStyle * style = gtk_style_copy(label->style);
    GdkColor red = { 0, 0xffff, 0x0000, 0x0000 };
    gdk_color_alloc(gdk_colormap_get_system(), &red);
    style->fg[0] = red;
    
    gtk_widget_set_style(label, style);
#ifdef USE_GTK
    gtk_style_unref(style);
#else
    g_object_unref(G_OBJECT(style));
#endif
  }
}


/**
 * Called when the user changes notebook pages to set the focus.
 */

void GTKTwoWindow::pageSwitch(int p) {
  GtkWidget * label;

  printf ("In GTKTwoWindow::pageSwitch(guint p)\n");
  
  // Figure out which page we're on...
  
  label = gtk_notebook_get_tab_label(GTK_NOTEBOOK(notebook), gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), p));
#ifdef USE_GTK
  gtk_widget_set_rc_style(GTK_WIDGET(label));
#else
  gtk_widget_set_style(GTK_WIDGET(label), NULL);
#endif
  // Set focus.
  for (Connection * tmp = first_connection; tmp; tmp = tmp->getNext()) {
    // Find which page this connection's VT is situated on, and grab focus.
    guint page = gtk_notebook_page_num(GTK_NOTEBOOK(notebook), tmp->getVT()->getMainWindow());
    if (page == p) {
      tmp->getVT()->focus();
    }
  }

	phandler->pageSwitched();

}

int GTKTwoWindow::numPages() {
  int i = 0;

  while (1) {
    if (!gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), i)) {
      return i;
    }
    i++;
  }
  return i;
}

void GTKTwoWindow::switchPage(int page) {

  if (!gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), page))
    return;

#ifdef USE_GTK
  gtk_notebook_set_page(GTK_NOTEBOOK(notebook), page);
#else
  gtk_notebook_set_current_page(GTK_NOTEBOOK(notebook), page);

#endif
  for (Connection * tmp = first_connection; tmp; tmp = tmp->getNext()) {
    // Find which page this connection's VT is situated on, and delete it.
    gint p = gtk_notebook_page_num(GTK_NOTEBOOK(notebook), tmp->getVT()->getMainWindow());
    if (page == p) {
      tmp->getVT()->focus();
    }
  }

  printf ("Just exiting switchPage(int page).\n");
}

void menu_add_menu(GtkWidget *, GtkWidget **, char *, bool);

void GTKTwoWindow::nextPage() {
  gint page = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
  
  if (page < 0)
    return;

  page++;

  if (page >= numPages())
    page = 0;

#ifdef USE_GTK
  gtk_notebook_set_page(GTK_NOTEBOOK(notebook), page);
#else
  gtk_notebook_set_current_page(GTK_NOTEBOOK(notebook), page);
#endif

  for (Connection * tmp = first_connection; tmp; tmp = tmp->getNext()) {
    // Grab focus in the input box.
    gint p = gtk_notebook_page_num(GTK_NOTEBOOK(notebook), tmp->getVT()->getMainWindow());
    if (page == p)
      tmp->getVT()->focus();
  }
}

int get_x_id();

void GTKTwoWindow::swallow() {
#if 0

#ifndef WIN32
  int id = get_x_id();
  int parent_id = -1;
  if (id == -1)
    return;

  // Create a socket widget and make it visible.
  GtkSocket * socket = (GtkSocket *)gtk_socket_new();
  gtk_widget_show (GTK_WIDGET(socket));

  // Check if we're trying to steal papaya.
  socket->plug_window = gdk_window_lookup(id);
  /*
  if (socket->plug_window && socket->plug_window->user_data) {
    new Message(_("Error"), _("Papaya cannot swallow itself."), NULL);
    gtk_widget_hide(GTK_WIDGET(socket));
    gtk_widget_destroy(GTK_WIDGET(socket));
    return;
  }
  */

  GtkWidget * label = gtk_label_new(_("Swallowed Program"));
  gtk_widget_show(label);

  // Stick it in the notebook.
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), GTK_WIDGET(socket), label);

  // Figure out the ID of the child

  // This is the ID of the parent window
  parent_id = GDK_WINDOW_XWINDOW (GTK_WIDGET(socket)->window);

  // Note that this does NOT work for enlightenment, but will work for
  // fvwm and sawmill.  It seems to be an enlightenment bug.

#ifdef USE_GTK
  gtk_socket_steal(GTK_SOCKET(socket), id);
#else
  gtk_socket_add_id(GTK_SOCKET(socket), id);
#endif
#endif
#endif
}

void GTKTwoWindow::closeWindow() {

  // Figure which is the active notebook page, and what its title is.
  gint page = gtk_notebook_get_current_page(GTK_NOTEBOOK(notebook));
  GtkWidget * label = gtk_notebook_get_tab_label(GTK_NOTEBOOK(notebook), gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), page));
  

#ifdef USE_GTK
  gchar * text;
  gtk_label_get(GTK_LABEL(label), &text);
#else
  const gchar * text;
  text = gtk_label_get_text(GTK_LABEL(label));
#endif
  if (!strcmp(text, "Swallowed App") || !strcmp(text, "mudFTP")) {
    // Remove this page.
    gtk_notebook_remove_page(GTK_NOTEBOOK(notebook), page);
    return;
  }

  // Deleting this object should be enough to remove it neatly...

  Connection * conn = getCurrentConnection();
  if (!conn || conn->isClosed())
    return;
  delete getCurrentConnection();
}

GtkWidget * GTKTwoWindow::addPageForApp(char * name) {
#if 0
#ifndef WIN32
  GtkWidget *socket = gtk_socket_new();
  gtk_widget_show (socket);
  GtkWidget * label = gtk_label_new(name);
  gtk_widget_show(label);

  gtk_widget_hide(background_pixmap);
  gtk_widget_show(notebook);

  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), socket, label);
  /* The following call is only necessary if one of
   * the ancestors of the socket is not yet visible.
   */
  gtk_widget_realize (socket);
  return socket;
#else
  return NULL;
#endif
#endif
  return NULL;
}

void GTKTwoWindow::addPage(char * name, GtkWidget * frame) {

  GtkWidget * label = gtk_label_new(name);
  gtk_widget_show(label);

  gtk_widget_hide(background_pixmap);
  gtk_widget_show(notebook);

  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), frame, label);
}

void GTKTwoWindow::addCustomMenu(char * name, GtkItemFactoryCallback cb) {
//  mudmenu_add_menu(name, cb);
}

void GTKTwoWindow::getDimensions(int * x, int * y) {
  // These values don't seem to include the window decorations on WIN32
  *x = window->allocation.width;
  *y = window->allocation.height;
}

static int gtk_two_window_callback(int result, gpointer data) {
  GTKTwoWindow * gw = (GTKTwoWindow *)data;
  return gw->callback(result);
}

int GTKTwoWindow::callback(int result) {
  char buf[1024];

  snprintf(buf, 1024, _("%d users"), result);

  printf ("Users result: %s.\n", buf);
  
  // Set the text in the information area.
  gtk_statusbar_push(GTK_STATUSBAR(statusbar), gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), "normal"), buf);

  return 1;
}

int GTKTwoWindow::timeout() {

  if (globalPreferences->getSendStats())
    count_number_of_users(gtk_two_window_callback, (gpointer) this);
  return 1;
}

static int gtk_two_window_timeout(gpointer data) {
  GTKTwoWindow * gw = (GTKTwoWindow *)data;
  return gw->timeout();
}

int GTKTwoWindow::configureUsers() {
  int t = 1000 * 60 * 10; // 10 minutes.
  gtk_timeout_add(t, gtk_two_window_timeout, this);

  timeout();
  return 0;
}



