#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>

#ifndef WIN32
#include <unistd.h>
#else
#define strncasecmp strnicmp
#endif

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include "callbacks.h"
#include "support.h"
#include "Message.h"
#include "mudclient.h"
#include "MUDConnectorList.h"

/**
 * Reads the MUD Connector text MUD list.
 *
 * Saves offset into file and name of MUD.  When MUD is queried the data can be
 * read from the data file.
 */


MUDConnectorList::MUDConnectorList() {
  firstEntry = NULL;
  fp = NULL;
  readMudlist();
}

MUDConnectorList::~MUDConnectorList() {
  struct mce * tmp = firstEntry;
  struct mce * tmp_next;

  for (; tmp; tmp = tmp_next) {
    tmp_next = tmp->next;
    if (tmp->name)
      free(tmp->name);
    free(tmp);
  }

  if (fp)
    fclose(fp);
}

void MUDConnectorList::appendList(struct mce * entry) {

  struct mce * curr;

  if (!firstEntry) {
    firstEntry = entry;
    firstEntry->next = NULL;
    firstEntry->prev = NULL;
    return;
  }

  for (curr = firstEntry; curr; curr = curr->next) {
    if (!curr->next) {
      curr->next = entry;
      entry->next = NULL;
      entry->prev = curr;
      return;
    }
  }

}

void MUDConnectorList::readMudlist() {

  char line[1024];

  fp = openFile("mudlist.txt", NULL, "r");
  if (!fp) {
    new Message(_("Information"), "You do not have the MUD Connector MUD List installed.  To install please select the MUD List option from the Connection menu and follow the instructions there.", true);
    return;
  }

  // Attempt to read a line from the file.
  while (fgets(line, 1024, fp)) {
    // Is this a 'Mud        :' line?
    if (!strncmp(line, "Mud         :", 13)) {
      struct mce * entry = (struct mce *)malloc(sizeof(struct mce));
      memset(entry, 0, sizeof(struct mce));
      entry->offset = ftell(fp);
      entry->name = strdup(line+14);

      appendList(entry);
    }
  }
}

MUDConnectorMUD * MUDConnectorList::getEntry(char * name) {

  struct mce * tmp;

  for (tmp = firstEntry; tmp; tmp = tmp->next) {
    if (!strcmp(tmp->name, name)) {
      return readEntry(tmp);
    }
  }
  
  return NULL;
}

MUDConnectorMUD * MUDConnectorList::readEntry(struct mce * entry) {

  char line[1024];

  if (!fp)
    return NULL;

  // Set the pointer to offset bytes from the start of the file.
  if (fseek(fp, entry->offset, SEEK_SET) == -1) {
    perror("fseek");
    return NULL;
  }

  MUDConnectorMUD * mud = new MUDConnectorMUD();
  mud->setName(entry->name);

  while (fgets(line, 1024, fp)) {

    if (!strncmp(line, "Code Base", 9)) {
      mud->setCodeBase(line);
    } else if (!strncmp(line, "Telnet", 6)) {
      mud->setTelnet(line);
    } else if (!strncmp(line, "WWW", 3)) {
      mud->setWWW(line);
    } else if (!strncmp(line, "Description", 11)) {
      // Read until we hit 'Mud' or other garbage.

      while (1) {
	char * pc = fgets(line, 1024, fp);
	
	if (!pc)
	  return mud;

	if (!strncmp(line, "Mud         :", 13))
	  return mud;
	mud->appendDescription(line);
      }
      return mud;
    }
  }

  return mud;
}

MUDConnectorMUD::MUDConnectorMUD() {

  name = NULL;
  description = NULL;
  host = NULL;
  port = 0;
  www = NULL;
  codebase = NULL;

}

MUDConnectorMUD::~MUDConnectorMUD() {
  if (name)
    free(name);

  if (description)
    free(description);

  if (host)
    free(host);

  if (www)
    free(www);

  if (codebase)
    free(codebase);
}

void MUDConnectorMUD::setName(char * n) {
  if (name)
    free(name);
  name = strdup(n);

  while (strlen(name) >  0) {
    char * pc = &(name[strlen(name) - 1]);
    if (*pc == '\r')
      *pc = '\0';
    else if (*pc == '\n')
      *pc = '\0';
    else
      break;
  }
}

char * MUDConnectorMUD::getName() {
  return name;
}

void MUDConnectorMUD::setCodeBase(char * n) {
  if (codebase)
    free(codebase);
  codebase = strdup(n+13);
}

char * MUDConnectorMUD::getCodeBase() {
  return codebase;
}

void MUDConnectorMUD::setTelnet(char * h) {

  char ho[1024];
  int p;

  sscanf(h, "Telnet      : %s %d%*[^\n]", ho, &p);

  host = strdup(ho);
  port = p;
}

char * MUDConnectorMUD::getHost() {
  return host;
}

int MUDConnectorMUD::getPort() {
  return port;
}

void MUDConnectorMUD::appendDescription(char * d) {
  if (!description) {
    description = (char *)malloc(16384);
    description[0] = '\0';
  }

  strcat(description, d);
}

char * MUDConnectorMUD::getDescription() {
  return description;
}

void MUDConnectorMUD::setWWW(char * w) {
  if (www)
    free(www);
  www = strdup(w + 13);
}

char * MUDConnectorMUD::getWWW() {
  return www;
}

extern MUDConnectorList * mcl;
struct mce * mce_curr = NULL;
GtkWidget *mcl_mudname;
GtkWidget *mcl_host;
GtkWidget *mcl_codebase;
GtkWidget *mcl_homepage;

#ifdef USE_GTK_2
GtkTextBuffer * mcl_description;
#else
GtkWidget * mcl_description;
#endif

void redraw_mcl() {

  char buf[1024];

  if (!mce_curr)
    return;

  MUDConnectorMUD * mcl_current = mcl->getEntry(mce_curr->name);
  if (!mcl_current)
    return;

  gtk_entry_set_text(GTK_ENTRY(mcl_mudname), mcl_current->getName());

  sprintf(buf, "Host: %s port %d", mcl_current->getHost(), mcl_current->getPort());
  gtk_label_set_text(GTK_LABEL(mcl_host), buf);

  sprintf(buf, "Codebase: %s", mcl_current->getCodeBase());
  gtk_label_set_text(GTK_LABEL(mcl_codebase), buf);

  sprintf(buf, "Homepage: %s", mcl_current->getWWW());
  gtk_label_set_text(GTK_LABEL(mcl_homepage), buf);


#ifdef USE_GTK_2
  GtkTextIter start, end;
  gtk_text_buffer_get_bounds(GTK_TEXT_BUFFER(mcl_description), &start, &end);
  gtk_text_buffer_delete(GTK_TEXT_BUFFER(mcl_description), &start, &end);
  gtk_text_buffer_get_bounds(GTK_TEXT_BUFFER(mcl_description), &start, &end);
  gtk_text_buffer_insert(GTK_TEXT_BUFFER(mcl_description), &end, mcl_current->getDescription(), -1);
#else
  gtk_text_set_point(GTK_TEXT(mcl_description), 0);
  gtk_text_forward_delete(GTK_TEXT(mcl_description), gtk_text_get_length(GTK_TEXT(mcl_description)));

  gtk_text_insert(GTK_TEXT(mcl_description), NULL, NULL, NULL, mcl_current->getDescription(), -1);
#endif

  delete mcl_current;

}

#define UP_ARROW 65362
#define DOWN_ARROW 65364
#define BACKSPACE 65288
#define SHIFT 65505
#define ENTER 65293

void update_options() {
  if (!mce_curr)
    mce_curr = mcl->firstEntry;

  redraw_mcl();
}

void up_option(GtkWidget * widget) {
  if (mce_curr->prev)
    mce_curr = mce_curr->prev;

  redraw_mcl();
}

void down_option(GtkWidget * widget) {
  if (mce_curr->next)
    mce_curr = mce_curr->next;

  redraw_mcl();
}

gint on_mcl_mudname_activate(GtkWidget * widget, GdkEvent * event, gpointer data) {
  return 1;
}

gint on_mcl_mudname_keypress(GtkWidget * widget, GdkEventKey * event) {
  char text[1024];
  int pos = 0;

  switch (event->keyval) {
  case UP_ARROW:
    up_option(widget);
    return 1;
  case DOWN_ARROW:
    down_option(widget);
    return 1;
  case BACKSPACE:

    pos = gtk_editable_get_position(GTK_EDITABLE(mcl_mudname));
    sprintf(text, "%s", (gchar *)gtk_entry_get_text(GTK_ENTRY(mcl_mudname)));

    text[pos] = '\0';

    gtk_entry_set_text(GTK_ENTRY(mcl_mudname), text);
    return 0;
    // Remove the previous char

  case SHIFT:
    return 1;

  case ENTER:
    if (mce_curr)
      update_options();
    return 1;

  default:

    // Need to find out if this is a printable character.

    sprintf(text, "%s", (gchar *)gtk_entry_get_text(GTK_ENTRY(mcl_mudname)));

    // Find the first match for this string.

    for (struct mce * tmp = mcl->firstEntry; tmp; tmp = tmp->next) {
      if (!strncasecmp(tmp->name, text, strlen(text))) {
	gtk_entry_set_text(GTK_ENTRY(mcl_mudname), tmp->name);
	gtk_editable_select_region(GTK_EDITABLE(mcl_mudname), strlen(text), -1);
	gtk_editable_set_position(GTK_EDITABLE(mcl_mudname), strlen(text));

	mce_curr = tmp;
	return 0;
      }
    }

  }
  return 0;
}

GtkWidget*
create_mcl_window (void)
{
  GtkWidget *mcl_window;
  GtkWidget *hbox1;
  GtkWidget *vbox1;
  GtkWidget *hbox2;
  GtkWidget *label1;
   GtkWidget *vbox2;
  GtkWidget *scrolledwindow1;
  GtkWidget *vbuttonbox1;
  GtkWidget *mcl_connect;
  GtkWidget *mcl_connect_wizard;
  GtkWidget *mcl_close;

  mcl_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_object_set_data (GTK_OBJECT (mcl_window), "mcl_window", mcl_window);
  gtk_window_set_title (GTK_WINDOW (mcl_window), "MUD Connector MUD List (mudlist obtained with permission from The Mud Connector (http://www.mudconnect.com))");

  gtk_signal_connect(GTK_OBJECT(mcl_window), "destroy", GTK_SIGNAL_FUNC(on_mcl_destroyed), NULL);

  gtk_widget_set_usize(GTK_WIDGET(mcl_window), 500, 300);

  hbox1 = gtk_hbox_new (FALSE, 0);
  gtk_widget_ref (hbox1);
  gtk_object_set_data_full (GTK_OBJECT (mcl_window), "hbox1", hbox1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (hbox1);
  gtk_container_add (GTK_CONTAINER (mcl_window), hbox1);

  vbox1 = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (vbox1);
  gtk_object_set_data_full (GTK_OBJECT (mcl_window), "vbox1", vbox1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (vbox1);
  gtk_box_pack_start (GTK_BOX (hbox1), vbox1, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox1), 4);

  hbox2 = gtk_hbox_new (FALSE, 0);
  gtk_widget_ref (hbox2);
  gtk_object_set_data_full (GTK_OBJECT (mcl_window), "hbox2", hbox2,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (hbox2);
  gtk_box_pack_start (GTK_BOX (vbox1), hbox2, FALSE, FALSE, 0);

  label1 = gtk_label_new (_("MUD Name"));
  gtk_widget_ref (label1);
  gtk_object_set_data_full (GTK_OBJECT (mcl_window), "label1", label1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label1);
  gtk_box_pack_start (GTK_BOX (hbox2), label1, FALSE, FALSE, 0);


  mcl_mudname = gtk_entry_new();
  gtk_widget_ref (mcl_mudname);
  gtk_object_set_data_full (GTK_OBJECT (mcl_window), "mcl_mudname", mcl_mudname,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (mcl_mudname);
  gtk_box_pack_start (GTK_BOX (hbox2), mcl_mudname, TRUE, TRUE, 0);

  gtk_signal_connect_full (GTK_OBJECT (mcl_mudname),
                           "key_press_event",
                           GTK_SIGNAL_FUNC(on_mcl_mudname_keypress),
                           (GtkCallbackMarshal)NULL,
                           (gpointer)NULL, /*optional gpointer for sig func*/
                           (GtkDestroyNotify)NULL,
                           0,
                           GTK_RUN_LAST);

  vbox2 = gtk_vbox_new (FALSE, 4);
  gtk_widget_ref (vbox2);
  gtk_object_set_data_full (GTK_OBJECT (mcl_window), "vbox2", vbox2,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (vbox2);
  gtk_box_pack_start (GTK_BOX (vbox1), vbox2, TRUE, TRUE, 0);

  mcl_host = gtk_label_new (_("Host: site:port"));
  gtk_widget_ref (mcl_host);
  gtk_object_set_data_full (GTK_OBJECT (mcl_window), "mcl_host", mcl_host,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (mcl_host);
  gtk_box_pack_start (GTK_BOX (vbox2), mcl_host, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (mcl_host), 7.45058e-09, 0.5);

  mcl_codebase = gtk_label_new (_("Codebase:"));
  gtk_widget_ref (mcl_codebase);
  gtk_object_set_data_full (GTK_OBJECT (mcl_window), "mcl_codebase", mcl_codebase,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (mcl_codebase);
  gtk_box_pack_start (GTK_BOX (vbox2), mcl_codebase, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (mcl_codebase), 7.45058e-09, 0.5);

  mcl_homepage = gtk_label_new (_("Homepage:"));
  gtk_widget_ref (mcl_homepage);
  gtk_object_set_data_full (GTK_OBJECT (mcl_window), "mcl_homepage", mcl_homepage,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (mcl_homepage);
  gtk_box_pack_start (GTK_BOX (vbox2), mcl_homepage, FALSE, FALSE, 0);
  gtk_misc_set_alignment (GTK_MISC (mcl_homepage), 0, 0.5);

  scrolledwindow1 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_ref (scrolledwindow1);
  gtk_object_set_data_full (GTK_OBJECT (mcl_window), "scrolledwindow1", scrolledwindow1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (scrolledwindow1);
  gtk_box_pack_start (GTK_BOX (vbox2), scrolledwindow1, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow1), GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);

#ifdef USE_GTK_2
  mcl_description = gtk_text_buffer_new(NULL);
  GtkWidget * text_view = gtk_text_view_new_with_buffer(GTK_TEXT_BUFFER(mcl_description));
  gtk_widget_show(text_view);
  gtk_container_add(GTK_CONTAINER(scrolledwindow1), text_view);
#else
  mcl_description = gtk_text_new (NULL, NULL);
  gtk_widget_ref (mcl_description);
  gtk_object_set_data_full (GTK_OBJECT (mcl_window), "mcl_description", mcl_description,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (mcl_description);
  gtk_container_add (GTK_CONTAINER (scrolledwindow1), mcl_description);
#endif

  vbuttonbox1 = gtk_vbutton_box_new ();
  gtk_widget_ref (vbuttonbox1);
  gtk_object_set_data_full (GTK_OBJECT (mcl_window), "vbuttonbox1", vbuttonbox1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (vbuttonbox1);
  gtk_box_pack_start (GTK_BOX (hbox1), vbuttonbox1, FALSE, FALSE, 0);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (vbuttonbox1), GTK_BUTTONBOX_START);

  mcl_connect = gtk_button_new_with_label (_("Connect"));
  gtk_widget_ref (mcl_connect);
  gtk_object_set_data_full (GTK_OBJECT (mcl_window), "mcl_connect", mcl_connect,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (mcl_connect);
  gtk_container_add (GTK_CONTAINER (vbuttonbox1), mcl_connect);
  GTK_WIDGET_SET_FLAGS (mcl_connect, GTK_CAN_DEFAULT);

  mcl_connect_wizard = gtk_button_new_with_label (_("Connection\nWizard"));
  gtk_widget_ref (mcl_connect_wizard);
  gtk_object_set_data_full (GTK_OBJECT (mcl_window), "mcl_connect_wizard", mcl_connect_wizard,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (mcl_connect_wizard);
  gtk_container_add (GTK_CONTAINER (vbuttonbox1), mcl_connect_wizard);
  GTK_WIDGET_SET_FLAGS (mcl_connect_wizard, GTK_CAN_DEFAULT);

  mcl_close = gtk_button_new_with_label (_("Close"));
  gtk_widget_ref (mcl_close);
  gtk_object_set_data_full (GTK_OBJECT (mcl_window), "mcl_close", mcl_close,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (mcl_close);
  gtk_container_add (GTK_CONTAINER (vbuttonbox1), mcl_close);
  GTK_WIDGET_SET_FLAGS (mcl_close, GTK_CAN_DEFAULT);

  gtk_signal_connect (GTK_OBJECT (mcl_connect), "clicked",
                      GTK_SIGNAL_FUNC (on_mcl_connect_clicked),
                      NULL);
  gtk_signal_connect (GTK_OBJECT (mcl_connect_wizard), "clicked",
                      GTK_SIGNAL_FUNC (on_mcl_connect_wizard_clicked),
                      NULL);
  gtk_signal_connect (GTK_OBJECT (mcl_close), "clicked",
                      GTK_SIGNAL_FUNC (on_mcl_close_clicked),
                      NULL);

  update_options();

  return mcl_window;
}

