#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#endif

#include "callbacks.h"
#include "Prefs.h"
#include "Message.h"
#include "SystemColour.h"
#include "PluginHandler.h"
#include "ColouredLabel.h"
#include "PapayaVT.h"
#include "BaseWindow.h"

#include "CommandInterpreter.h"

#include "TeloptFilter.h"
#include "mudclient.h"

extern PluginHandler * phandler;
extern CommandInterpreter * commandInterpreter;
extern BaseWindow * mainWindow;

#define ESCAPE '\033'
#define PAGE_UP 65365
#define PAGE_DOWN 65366

gboolean prompt_expose(GtkWidget *, GdkEventExpose *, gpointer);

/*
  Then, you just need to know that the Editable widgets don't even to
  try to paste when not editable. So you need to connect to
  "button_press_event", and do a gtk_selection_convert() yourself.
  Then, you need to connect to "selection_received" to get the data when
  it arrives. Your "selection_received" handler then needs to
  do a gtk_signal_emit_stop_by_name(), so that the GtkEditable
  default selection_received handler doesn't get called.
*/

void papaya_selection_received(GtkWidget * widget, GtkSelectionData * selection_data, gpointer data) {

  GtkWidget * input = mainWindow->getCurrentConnection()->getVT()->getInput();

  if (selection_data->length < 0)
    return;

  char buf[16384];
  buf[0] = '\0';
  
  if (gtk_entry_get_text(GTK_ENTRY(input)))
    strcat(buf, gtk_entry_get_text(GTK_ENTRY(input))); 

  if (selection_data->type == gdk_atom_intern("STRING", FALSE)) {
    selection_data->data[selection_data->length] = 0;
    strcat(buf, (char *)selection_data->data);
  } else {
    if (selection_data->type == gdk_atom_intern("COMPOUND_TEXT", FALSE)
      || selection_data->type == gdk_atom_intern("TEXT", FALSE)) {
      gchar **list;
      gint count;
      gint i;
      
      count = gdk_text_property_to_text_list (selection_data->type,
					      selection_data->format, 
					      selection_data->data,
					      selection_data->length,
					      &list);
      for (i = 0; i < count; i++)
	if (list[i])
	  strcat(buf, list[i]);
    }
  }
 
  gtk_entry_set_text(GTK_ENTRY(input), buf);
  gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "selection_received");
  
}

void output_size_changed(GtkWidget * widget, GdkEvent * event, gpointer data) {

  Connection * c = (Connection *)data;

  // Tell this connection to redo NAWS.
  TeloptFilter * filter = static_cast<TeloptFilter*>(c->getSocket()->inputFilters.findFilter("TelnetOptions"));
  if (filter)
    filter->do_naws();

  // Cause this VT to scroll to bottom.

  VT * vt = c->getVT();
  vt->scroll();

}

void output_button_pressed(GtkWidget * widget, GdkEventButton * event, gpointer data) {

  GdkAtom ctext_atom = gdk_atom_intern("COMPOUND_TEXT", FALSE);

  switch (event->type) {
  case GDK_BUTTON_PRESS:
    switch (event->button) {
    case 2:
      gtk_signal_connect(GTK_OBJECT(widget), "selection_received", GTK_SIGNAL_FUNC(papaya_selection_received), data);
      
      gtk_selection_convert(widget, GDK_SELECTION_PRIMARY, ctext_atom, GDK_CURRENT_TIME);
      gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "button_press_event");
      
      return;
    }
    return;

  case GDK_2BUTTON_PRESS:
    switch (event->button) {
    case 1:
      // @@ Figure out which bit of text we're sitting over.

      return;
    }

  /* k0n pointed out that some combinations of GDK/C++ spew warnings if
   * the default option is not here.  (switch (enumeration) requires
   * a case statement for each item in the enumeration.)
   */
  default:
    return;
  }

}

PapayaVT::PapayaVT(Connection * c) : VT(c) {
  output = (GtkWidget *)NULL;
  input = (GtkWidget *)NULL;
  vscrollbar = (GtkWidget *)NULL;
  filtered_output = (GtkWidget *)NULL;
  filtered_hbox = (GtkWidget *)NULL;
  filtered_vscrollbar = (GtkWidget *)NULL;

  appended_to_filtered = false;
  appended_to_output = false;
  filtered_lines = 0;
  output_lines = 0;

  color_map = gdk_colormap_get_system();
  default_color.red = default_color.green = default_color.blue = 0xffff;
  color.red = default_color.red;
  color.green = default_color.green;
  color.blue = default_color.blue;

  if (!gdk_color_alloc(color_map, &color) || !gdk_color_alloc(color_map, &default_color)) {
    printf(_("Couldn't allocate colour."));
    printf(CRLF);
  }

  bold = FALSE;
  flashing = FALSE;
  fg = 0;
  bg = 7;

  foreground_color_default = TRUE;
  background_color_default = TRUE;
  papaya_colour = FALSE;
  flashing_text = false;
  should_scroll = TRUE;

  initTextArea();
}

PapayaVT::~PapayaVT() {
}

void PapayaVT::destroyFilteredWidget() {
  gtk_widget_hide(filtered_hbox);
  gtk_container_remove(GTK_CONTAINER(output_vbox), filtered_hbox);
  gtk_widget_destroy(filtered_hbox);
  filtered_output = (GtkWidget *)NULL;
  filtered_hbox = (GtkWidget *)NULL;
  filtered_vscrollbar = (GtkWidget *)NULL;
}

void PapayaVT::createFilteredWidget() {

  filtered_hbox = gtk_hbox_new(FALSE, 2);
  gtk_widget_show(filtered_hbox);

  filtered_output = gtk_text_new((GtkAdjustment *)NULL, (GtkAdjustment *)NULL);
  gtk_text_set_editable(GTK_TEXT(filtered_output), FALSE);
  gtk_text_set_word_wrap(GTK_TEXT(filtered_output), FALSE); // Need a horiz bar.
  gtk_widget_show(filtered_output);
  gtk_box_pack_start(GTK_BOX(filtered_hbox), filtered_output, TRUE, TRUE, 0);

  filtered_vscrollbar = gtk_vscrollbar_new (GTK_TEXT(filtered_output)->vadj);
  gtk_box_pack_start(GTK_BOX(filtered_hbox), filtered_vscrollbar, FALSE, FALSE, 0);
  gtk_widget_show(filtered_vscrollbar);

  gtk_box_pack_start(GTK_BOX(output_vbox), filtered_hbox, FALSE, FALSE, 0);

  gtk_signal_connect(GTK_OBJECT(filtered_output), "button_press_event",
		     GTK_SIGNAL_FUNC(output_button_pressed),
		     input);
}

void PapayaVT::initTextArea() {

  GdkColor white;

  white.red = 0xffff;
  white.green = 0xffff;
  white.blue = 0xffff;

  gdk_color_alloc(color_map, &white);

  vbox = gtk_vbox_new(FALSE, 2);
  hbox = gtk_hbox_new(FALSE, 2);

  output = gtk_text_new((GtkAdjustment *)NULL, (GtkAdjustment *)NULL);
  gtk_text_set_editable(GTK_TEXT(output), FALSE);
  gtk_text_set_word_wrap(GTK_TEXT(output), FALSE); // Need a horiz bar.

  //  gtk_signal_connect(GTK_OBJECT(output), "key_press_event", GTK_SIGNAL_FUNC(on_output_box_keypress), this);

  gtk_box_pack_start(GTK_BOX(hbox), output, TRUE, TRUE, 0);
  gtk_widget_show(output);

  vscrollbar = gtk_vscrollbar_new (GTK_TEXT(output)->vadj);
  gtk_box_pack_start(GTK_BOX(hbox), vscrollbar, FALSE, FALSE, 0);
  gtk_widget_show(vscrollbar);

  output_vbox = gtk_vbox_new(FALSE, 1);
  gtk_widget_show(output_vbox);

  gtk_box_pack_start(GTK_BOX(output_vbox), hbox, TRUE, TRUE, 0);
  if (conn->queryPreferences()->getFilteredOutput()) {
    createFilteredWidget();
  }

  //  gtk_text_set_background(GTK_TEXT(output), dark_gray);
  //  gtk_text_set_foreground(GTK_TEXT(output), white);

  // Load a font...

  prompt = new ColouredLabel(conn, "");
  prompt->createWidget();
  gtk_widget_show(prompt->getWidget());
  
  input = gtk_entry_new();
  gtk_widget_show (input);

  GtkStyle * style;
  gtk_widget_ensure_style(input);
  style = gtk_style_copy(input->style);

#ifdef USE_GTK_2
  GdkFont * f = gtk_style_get_font(style);
  conn->queryPreferences()->loadFont(conn->queryPreferences()->getInputFont(), &f);
#else
  conn->queryPreferences()->loadFont(conn->queryPreferences()->getInputFont(), &(style->font));

  gdk_font_ref(style->font);
  gtk_widget_set_style(input, style);
  gtk_style_unref(style);
#endif

  gtk_signal_connect (GTK_OBJECT (input), "activate",
                      GTK_SIGNAL_FUNC (papaya_input_entry_activate),
                      this);

  gtk_signal_connect_full (GTK_OBJECT (input),
                           "key_press_event",
                           GTK_SIGNAL_FUNC(papaya_keypress),
                           (GtkCallbackMarshal)NULL,
                           (gpointer)NULL, /*optional gpointer for sig func*/
                           (GtkDestroyNotify)NULL,
                           0,
                           GTK_RUN_LAST);

  gtk_signal_connect_full (GTK_OBJECT (output),
                           "key_press_event",
                           GTK_SIGNAL_FUNC(papaya_keypress),
                           (GtkCallbackMarshal)NULL,
                           (gpointer)NULL, /*optional gpointer for sig func*/
                           (GtkDestroyNotify)NULL,
                           0,
                           GTK_RUN_LAST);

  GTK_WIDGET_UNSET_FLAGS( GTK_WIDGET(output), GTK_CAN_FOCUS );

  /*
    gtk_signal_connect (GTK_OBJECT (input), "key_press_event",
    GTK_SIGNAL_FUNC (on_input_entry_keypress),
    NULL);
  */
  gtk_box_pack_start(GTK_BOX(vbox), output_vbox, TRUE, TRUE, 0);

  input_vbox = gtk_vbox_new(FALSE, 2);
  gtk_widget_show(input_vbox);

  if (conn->queryPreferences()->getPromptOnInputLine()) {
    GtkWidget * pbox = gtk_hbox_new(FALSE, 2);
    gtk_box_pack_start(GTK_BOX(pbox), prompt->getWidget(), FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(pbox), input, TRUE, TRUE, 0);
    gtk_widget_show(pbox);
    
    gtk_box_pack_start(GTK_BOX(input_vbox), pbox, FALSE, FALSE, 0);
  } else {
    gtk_box_pack_start(GTK_BOX(input_vbox), prompt->getWidget(), FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(input_vbox), input, FALSE, FALSE, 0);
  }

  gtk_box_pack_start(GTK_BOX(vbox), input_vbox, FALSE, FALSE, 0);

  // Why does this center the widget in the box?

  gtk_widget_show(hbox);

  gtk_container_add(GTK_CONTAINER(box), vbox);
  gtk_widget_show(vbox);

  gtk_signal_connect(GTK_OBJECT(output), "button_press_event",
		     GTK_SIGNAL_FUNC(output_button_pressed),
		     input);

  // Need to connect up a signal for when the size of this window is changed.
  gtk_signal_connect(GTK_OBJECT(output), "size_allocate",
		     GTK_SIGNAL_FUNC(output_size_changed), conn);

}

void PapayaVT::hideInput() {
  gtk_widget_hide(input_vbox);
}

void PapayaVT::showInput() {
  gtk_widget_show(input_vbox);
}

char * PapayaVT::zapChars(char * text) {

  char * pc;

  if (*text == '\r')
    text++;

  pc = text;

  while (*pc != '\0') {

    if (*pc == '\r') {
      memmove(pc, pc + 1, strlen(pc + 1) + 1);
      // Don't increment pc, as we just moved the next char to there
      continue;
    }

    pc++;
  }

  return text;
}

/**
 * Any text to be considered for the display will be sent to this method.
 * This includes prompts, which the VT must remove if it wishes to display
 * them in a separate area.
 */

void PapayaVT::append(char * text) {

  char * pc;
  char * last_marker;
  
  gtk_text_freeze(GTK_TEXT(output));

  if (filtered_output)
    gtk_text_freeze(GTK_TEXT(filtered_output));

  snarfPrompt(text); // note: modifies text in place & sets prompt
  
  pc = zapChars(text);
  last_marker = pc;

  if (strlen(pc) > 0)
    mainWindow->nextText(conn);

  while ((pc = strchr(pc, ESCAPE))) {
    // NUL terminate the last_marker string.
    *pc = '\0';

    // Write all stuff between last_marker and pc to the VT.
    printVT(last_marker);
    pc = parseVTCode(pc);
    last_marker = pc;
  }

  printVT(last_marker);

  gtk_text_thaw(GTK_TEXT(output));
  if (filtered_output)
    gtk_text_thaw(GTK_TEXT(filtered_output));


  if (appended_to_filtered)
    filtered_lines++;
  if (appended_to_output)
    output_lines++;

  appended_to_filtered = false;
  appended_to_output = false;
}

char * PapayaVT::getColourCodes(char * s) {

  char * pc;
  static char buf[16384];
  int count = 0;

  while ((pc = strchr(s, ESCAPE))) {

    buf[count++] = *pc++; // ESCAPE
    buf[count++] = *pc++; // [
    while (*pc && !isalpha(*pc))
      buf[count++] = *pc++;

    if (*pc)
	buf[count++] = *pc++; // alpha terminating character

    s = pc;
  }

  buf[count] = '\0';

  return buf;
}

void PapayaVT::snarfPrompt(char * s)
{
    // Find last prompt-end marker
    char *marker = strrchr(s, MAGIC_PROMPT);
    if (!marker) // no eol to be seen, no changes needed.
	return;

    // Find last line-end before marker
    char *pc;
    for (pc = marker - 1; pc >= s; --pc)
	if (*pc == '\n')
	    break;

    ++pc; // point to start of line (or buffer if no line-end found)

    // Take a copy of the prompt.
    int prompt_len = marker - pc;
    char *prompt = new char[prompt_len + 1];
    memcpy(prompt, pc, prompt_len);
    prompt[prompt_len] = 0;

    setPrompt(prompt);

    // Now meddle with the output buffer.
    if (conn->queryPreferences()->getDoublePrompt())
    {
	delete[] prompt;
	return; // no changes needed.
    }

    // Replace prompt with only the color codes from it.
    strcpy(pc, getColourCodes(prompt)); // guaranteed to be smaller
    strcpy(pc+strlen(pc), marker+1);
    delete[] prompt;
}

/**
 * Sets the prompt to 'text' after stripping colour codes.
 */

void PapayaVT::setPrompt(char * text) {
  
  phandler->promptFilters(conn, text);
  if (strlen(text) == 0)
    return;

  char * pr = strdup(text); // Assume it isn't safe to edit text directly.

  prompt->setText(pr);

  /*

  char * ptr = pr;
  char * last;

  // Strip the colours from the string before setting the prompt
  
  while (*ptr != '\0') {

    // If we have hit ESCAPE, look for the next alpha character.
    if (*ptr == '\033') {
      last = ptr;

      while (!isalpha(*ptr))
	ptr++;

      ptr++;

      memmove(last, ptr, strlen(ptr) + 1);
      ptr = last;
      
      continue;
    }

    ptr++;
  }

  gtk_label_set_text(GTK_LABEL(prompt), pr);
  // Free the duplicated string.
  */
  free(pr);
}

/**
 * 'code' is a pointer to the ESCAPE char at the start of a VT100 code.
 * The returned pointer points at the first character after the complete
 * VT100 code.
 */

char * PapayaVT::parseVTCode(char * code) {

  char *pc = code+1;

  switch (*pc) {
  case '[':
    pc = parseBracketVTCode(pc);
    break;

  default:
    pc = parsePlainVTCode(pc);
    break;
  }

  pc++;
  return pc;
}

char * PapayaVT::parseBracketVTCode(char * pc) {

  char * orig = pc;

  while (!isalpha(*pc))
    pc++;

  switch (*pc) {

  case 'm': // Colour code
    setColour(orig);
    break;

  case 'p': // Internal Papaya colour code.
    setPapayaColour(orig);
    break;

  case 'q': // Reset Papaya colour.
    resetPapayaColour();

    // Cursor Control

  case 'H': // Set cursor to x;y
  case 'f': // Set cursor to x;y
    setPosition(orig);
    break;

  case 'A': // Move the cursor up x rows
    break;

  case 'B': // Move cursor down x rows
    break;

  case 'C': // Move cursor forward x columns
    break;

  case 's': // Save current current position
  case '7':
    break;

  case 'u': // Restore cursor position to last saved.
  case '8':
    break;

    // Device Status

  case 'c': // Requests a "Report Device Code" response
    break;

  case 'n': // Various device response stuff.
    break;

    // Terminal Setup
   
  case 'h': // Line wrap enabled (7h)
    break;

  case 'l': // Line wrap disabled (7l)
    break;

    // Scrolling

  case 'r': // Enable scrolling "x,y" = rows x to y.  No args = whole display
    break;

    // Tab Control

  case 'g': // Clears tab at the current position (3g = all tabs)
    break;

    // Erasing Text

  case 'K': // Various line based erases
    break;

  case 'J': // Various row based erases
    break;

    // Printing

  case 'i': // Printing related commands
    break;

  }

  // Now at the character that marks the start of normal text or the ESCAPE
  // for a new code.
  return pc;
}

char * PapayaVT::parsePlainVTCode(char * pc) {

  while (!isalpha(*pc))
    pc++;

  switch (*pc) {

    // Scrolling
    
  case 'D': // Scroll down one line
    break;
    
  case 'M': // Scroll up one line
    break;

    // Tab Control
    
  case 'H': // Sets a tab at the current position
    break;
    
    // Terminal Setup

  case 'c': // Reset all terminal settings to default
    break;

    // Fonts

  case ')': // Set default font
    break;

  case '(': // Use alternate font
    break;

  }


  pc++;
  return pc;
}

void PapayaVT::setPapayaColour(char * buf) {

  int red;
  int green;
  int blue;

  // allocate color

  char * pc;

  pc = strchr(buf, 'p');
  if (!pc)
    return;

  if (pc - buf != 12) { // [rrr;ggg;bbb
    printf(_("Dud Papaya Colour code."));
    printf(CRLF);
    return;
  }

  *pc = '\0';
  if (sscanf(buf, "[%d;%d;%d", &red, &green, &blue) != 3) {
    printf(_("PapayaColour: unable to sscanf"));
    printf(CRLF);
    return;
  }

  papaya_colour = TRUE;

  color.red = red * 0xffff / 255;
  color.green = green * 0xffff / 255;
  color.blue = blue * 0xffff / 255;

  gdk_color_alloc(color_map, &color);
}

void PapayaVT::resetPapayaColour() {
  papaya_colour = FALSE;

  if (bold) {
    struct colour_table * table = conn->queryPreferences()->getColourTable();

    color.red = table[fg+8].red;
    color.green = table[fg+8].green;
    color.blue = table[fg+8].blue;

    background_color.red = table[bg + 8].red;
    background_color.green = table[bg + 8].green;
    background_color.blue = table[bg + 8].blue;

    gdk_color_alloc(color_map, &color);
    gdk_color_alloc(color_map, &background_color);
  } else {
    struct colour_table * table = conn->queryPreferences()->getColourTable();

    color.red = table[fg].red;
    color.green = table[fg].green;
    color.blue = table[fg].blue;

    background_color.red = table[bg].red;
    background_color.green = table[bg].green;
    background_color.blue = table[bg].blue;

    gdk_color_alloc(color_map, &color);
    gdk_color_alloc(color_map, &background_color);
  }
}

void PapayaVT::setColour(char * buf) {
  char * pc = buf;
  char buf2[10];
  int pos = 0;

  buf2[0] = '\0';
  
  papaya_colour = FALSE;

  while (1) {

    // If we haven't got a [ or ; indicating more codes, leave the loop.
    if (!(*pc == ';' || *pc == '['))
      break;

    // Rip out the bit until the next non-alphanumeric character.
    pos = 0;

    buf2[pos] = *pc++;

    while (1) {
      if (isalpha(*pc) || *pc == ';')
        break;

      buf2[pos++] = *pc++;
    }
    
    buf2[pos] = '\0';
    setActualColour(buf2);
  }
}

void PapayaVT::setActualColour(char * buf) {
  if (!conn->queryPreferences()->getColour())
    return;

  if (buf[0] == '0' || strlen(buf) == 0) {

    // Set everything to defaults.
    background_color_default = TRUE;
    foreground_color_default = TRUE;

    bold = FALSE;
    flashing = FALSE;
    return;
  }

  if (buf[0] == '1') {
    bold = TRUE;

    struct colour_table * table = conn->queryPreferences()->getColourTable();

    color.red = table[fg+8].red;
    color.green = table[fg+8].green;
    color.blue = table[fg+8].blue;

    background_color.red = table[bg + 8].red;
    background_color.green = table[bg + 8].green;
    background_color.blue = table[bg + 8].blue;

    gdk_color_alloc(color_map, &color);
    gdk_color_alloc(color_map, &background_color);

    return;
  }

  if (buf[0] == '5') {
    flashing = TRUE;
    return;
  }

  /* Look up which colour we're to use. */

  /* buf[2] should contain the colour. */

  if (strlen(buf) >= 2) {
    for (int i = 0; i < 8; i++) {

      struct colour_table * table = conn->queryPreferences()->getColourTable();

      if (table[i].colour_character == buf[1]) {

        if (buf[0] == '3') {
	  foreground_color_default = FALSE;

          if (bold) {
            color.red = table[i+8].red;
            color.green = table[i+8].green;
            color.blue = table[i+8].blue;
          } else {
            color.red = table[i].red;
            color.green = table[i].green;
            color.blue = table[i].blue;
          }
          gdk_color_alloc(color_map, &color);
          fg = i;
          return;
        }
        
        if (buf[0] == '4') {
          background_color_default = FALSE;

          if (bold) {
            background_color.red = table[i+8].red;
            background_color.green = table[i+8].green;
            background_color.blue = table[i+8].blue;
          } else {
            background_color.red = table[i].red;
            background_color.green = table[i].green;
            background_color.blue = table[i].blue;
          }
          gdk_color_alloc(color_map, &background_color);
          bg = i;
          return;
        }
        
        return;

      }
    }
  }
}

void PapayaVT::beep(char * buf) {
  char * b;
  while ((b = strchr(buf, '\007'))) {
    gdk_beep();
    buf = b + 1;
  }
}

void PapayaVT::printVT(char * text) {
  char * start = text;

  struct colour_table * table = conn->queryPreferences()->getColourTable();

  if (conn->queryPreferences()->getBeep())
    beep(start);

  if (flashing_text && (start[0] == '\n' || start[0] == '\r')) {
    start++;
    flashing_text = false;
  }
    
  // Write this text to the VT at the current cursor position.
  gtk_text_forward_delete(GTK_TEXT(output), strlen(start));

  // We've already parsed the colours.  Now if background_colour_default
  // or foreground_colour_default are toggled on then we need to adjust
  // background_color and foreground_color accordingly.

  GdkColor * fg_tmp = &color;
  GdkColor * bg_tmp = &background_color;

  if (background_color_default) {
    if (conn->queryPreferences()->getDefaultBackgroundColour() == -1) {
      bg_tmp = (GdkColor *)NULL;
    } else {
      // Alloc the colour
      int i = conn->queryPreferences()->getDefaultBackgroundColour();
      if (bold) {
	background_color.red = table[i+8].red;
	background_color.green = table[i+8].green;
	background_color.blue = table[i+8].blue;
      } else {
	background_color.red = table[i].red;
	background_color.green = table[i].green;
	background_color.blue = table[i].blue;
      }
      gdk_color_alloc(color_map, &background_color);
    }
  }

  if (!papaya_colour && foreground_color_default) {
    if (conn->queryPreferences()->getDefaultForegroundColour() == -1) {
      fg_tmp = (GdkColor *)NULL;
    } else {
      int i = conn->queryPreferences()->getDefaultForegroundColour();
      if (bold && !conn->queryPreferences()->getBold()) {
	color.red = table[i+8].red;
	color.green = table[i+8].green;
	color.blue = table[i+8].blue;
      } else {
	color.red = table[i].red;
	color.green = table[i].green;
	color.blue = table[i].blue;
      }
      gdk_color_alloc(color_map, &color);
    }
  }

  insertText(GTK_TEXT(output), conn->queryPreferences()->getOutput(), fg_tmp, bg_tmp, start, -1);
}

void PapayaVT::insertText(GtkText * output, GdkFont * outputFont, GdkColor * fg_tmp, GdkColor * bg_tmp, char * text, int len){
  if (flashing && conn->queryPreferences()->getFilteredOutput()) {
    if (!filtered_output) {
      createFilteredWidget();
      scroll();
    }
    
    gtk_text_insert(GTK_TEXT(filtered_output),
      (!bold) ? conn->queryPreferences()->getOutput()
              : conn->queryPreferences()->getBold(),
      fg_tmp, bg_tmp, text, len);
    
    if (strlen(text) > 0 && text[strlen(text)-1] != '\n' && text[strlen(text)-1] != '\r') {
      flashing_text = true;
      gtk_text_insert(GTK_TEXT(filtered_output),
        (!bold) ? conn->queryPreferences()->getOutput()
                : conn->queryPreferences()->getBold(),
        fg_tmp, bg_tmp, "\n", -1);

      appended_to_filtered = true;
    }
    
  }
  else {
    gtk_text_insert(GTK_TEXT(output),
      (!bold) ? conn->queryPreferences()->getOutput()
              : conn->queryPreferences()->getBold(),
      fg_tmp, bg_tmp, text, len);
    appended_to_output = true;
  }

}

void PapayaVT::scroll() {

  // Trim the amount of text to the maximum number of characters.
  if (conn->queryPreferences()->getUseMaxLines()
      && conn->queryPreferences()->getMaxLines() > 0) {

    // Figure out how many lines to remove.
    int remove_output = output_lines - conn->queryPreferences()->getMaxLines();

    // Are there any lines to remove.
    if (remove_output > 0) {
      gtk_text_freeze(GTK_TEXT(output));
      
      while (remove_output > 0) {
	// Remove the first line from the box.
	char * text = gtk_editable_get_chars(GTK_EDITABLE(output), 0, -1);
	// gtk_editable_get_chars does a strdup ...
	if (text) {
	  char * pc = strchr(text, '\n');
	  if (!pc)
	    break;
	  gtk_text_set_point(GTK_TEXT(output), 0);
	  gtk_text_forward_delete(GTK_TEXT(output), pc - text + 1);
	  output_lines--;
	  remove_output--;
	  free(text);
	}
      }
      gtk_text_set_point(GTK_TEXT(output), gtk_text_get_length(GTK_TEXT(output)));
      gtk_text_thaw(GTK_TEXT(output));
    }

    if (filtered_output) {
      int remove_filtered = filtered_lines - conn->queryPreferences()->getMaxLines();
      // Are there any lines to remove.
      if (remove_filtered > 0) {
	gtk_text_freeze(GTK_TEXT(filtered_output));
	
	while (remove_filtered > 0) {
	  
	  // Remove the first line from the box.
	  char * text = gtk_editable_get_chars(GTK_EDITABLE(output), 0, -1);
	  if (text) {
	    char * pc = strchr(text, '\n');
	    gtk_text_set_point(GTK_TEXT(filtered_output), 0);
	    gtk_text_forward_delete(GTK_TEXT(filtered_output), pc - text + 1);
	    filtered_lines--;
	    remove_filtered--;
	  }
	}
	gtk_text_set_point(GTK_TEXT(filtered_output), gtk_text_get_length(GTK_TEXT(filtered_output)));
	gtk_text_thaw(GTK_TEXT(filtered_output));
      }
    }
  }

  bool scr = doScroll();
  if (scr) {
    // Scroll to the bottom if scroll on new input is enabled.
    gtk_text_freeze(GTK_TEXT(output));
    GtkAdjustment * adj = GTK_RANGE(vscrollbar)->adjustment;
    gtk_adjustment_clamp_page(adj, adj->upper, adj->upper);
    gtk_text_thaw(GTK_TEXT(output));

    if (filtered_output) {
      gtk_text_freeze(GTK_TEXT(filtered_output));
      GtkAdjustment * adj = GTK_RANGE(filtered_vscrollbar)->adjustment;
      gtk_adjustment_clamp_page(adj, adj->upper, adj->upper);
      gtk_text_thaw(GTK_TEXT(filtered_output));
    }
  }

}

bool PapayaVT::doScroll() {
  if (conn->queryPreferences()->getScrollOnInput())
    return TRUE;

  return should_scroll;
}


void PapayaVT::preInsert() {
  GtkAdjustment * adj = GTK_RANGE(vscrollbar)->adjustment;
  
  if (adj->value + adj->page_size == adj->upper)
    should_scroll = TRUE;
  else
    should_scroll = FALSE;
}

void PapayaVT::use() {
  GdkColor background_colour;

  background_colour.red = conn->queryPreferences()->getScreenBackgroundRed();
  background_colour.green = conn->queryPreferences()->getScreenBackgroundGreen();
  background_colour.blue = conn->queryPreferences()->getScreenBackgroundBlue();

  if (!gdk_color_alloc(color_map, &background_colour)) {
    printf (_("Failed to allocate colour."));
    printf(CRLF);
  }
  gdk_window_set_background(GTK_TEXT(output)->text_area, &background_colour);

  if (filtered_output)
    gdk_window_set_background(GTK_TEXT(filtered_output)->text_area, &background_colour);

}

GtkWidget * PapayaVT::getInput() {
  return input;
}

GtkWidget * PapayaVT::getOutput() {
  return output;
}

GtkWidget * PapayaVT::getBox() {
  return vbox;
}

GtkWidget * PapayaVT::getPrompt() {
  return prompt->getWidget();
}

int PapayaVT::width() {
  int width, height;
  getScreenSize(&width, &height);
  return width;
}

int PapayaVT::height() {
  int width, height;
  getScreenSize(&width, &height);
  return height;
}

void PapayaVT::getScreenSize(int * width, int * height) {
  int h, w;
  int ww, wh;

  h = conn->queryPreferences()->getOutput()->ascent + conn->queryPreferences()->getOutput()->descent;
  w = gdk_char_width(conn->queryPreferences()->getOutput(), 'W');

  ww = output->allocation.width;
  wh = output->allocation.height;

  *width = (ww/w);
  *height = (wh/h);
}

void PapayaVT::setPosition(char * text) {
  
  // Returning immediately since this support is nowhere near complete.
  return;

  int row, column;
  char first[10];
  char second[10];
  int fc = 0, sc = 0;

  // Get past the [
  text++;

  while (isdigit(*text)) {
    first[fc++] = *text++;
  }
  first[fc] = '\0';

  // Get past the ;
  text++;

  while (isdigit(*text)) {
    second[sc++] = *text++;
  }
  second[sc] = '\0';

  row = atoi(first);
  column = atoi(second);

  setPosition(row, column);
}

void PapayaVT::setPosition(int r, int c) {
  int rows, columns;
  getScreenSize(&columns, &rows);

  if (r > rows || c > columns) {
    return;
  }

  // Retrieve the text from the text box.
  char * text = gtk_editable_get_chars(GTK_EDITABLE(output), 0, -1);

  char * ptr = text + strlen(text);

  // Need to figure out which line is the right one.
  for (int i = rows; i >= 0; i--) {

    // Move to the char after '\n';
    ptr = findStartOfLine(ptr, text);

    if (rows - i == r) {

      char * eol = ptr;
      while (*eol != '\n')
	eol++;
      eol--; // Back off the \n

      while (1) {
	if (c <= eol - ptr) {
	  gtk_text_set_point(GTK_TEXT(output), (ptr - text) + c);
	  return;
	}

	gtk_text_set_point(GTK_TEXT(output), (int)(eol - text));
	insertText(GTK_TEXT(output), conn->queryPreferences()->getOutput(), (GdkColor *)NULL, (GdkColor *)NULL, " ", -1);
	eol++;
      }
      
      // We might be off by one at this point.
      return;
    }
  }
}

char * PapayaVT::findStartOfLine(char * ptr, char * limit) {

  ptr--;
  ptr--;

  while (*ptr != '\n' && ptr >= limit)
    ptr--;

  ptr++; // Go past the \n

  if (ptr < limit)
    ptr = limit;

  return ptr;
}

void PapayaVT::moveLeft() {
}

void PapayaVT::moveRight() {
}

void PapayaVT::setInput(char * s) {
  gtk_entry_set_text(GTK_ENTRY(input), s);
}

/**
 * The preferences have been updated, so let's check the settings.
 */

void PapayaVT::prefsUpdated() {


  GtkWidget * widget;
  GtkStyle * style;

  // prompt fonts
  /*
  GtkWidget * widget = getPrompt();
  GtkStyle * style = gtk_style_copy(widget->style);
  conn->queryPreferences()->loadFont(conn->queryPreferences()->getPromptFont(), &(style->font));
  gdk_font_ref(style->font);
  gtk_widget_set_style(widget, style);
  gtk_style_unref(style);
  */

  // input fonts
  
  widget = getInput();
  style = gtk_style_copy(widget->style);

#ifdef USE_GTK_2
  GdkFont * f = gtk_style_get_font(style);
  conn->queryPreferences()->loadFont(conn->queryPreferences()->getInputFont(), &f);
#else
  conn->queryPreferences()->loadFont(conn->queryPreferences()->getInputFont(), &(style->font));

  gdk_font_ref(style->font);
  gtk_widget_set_style(input, style);
  gtk_style_unref(style);
#endif

  // This will set the background colour.
  use();

  // Change where flashing text goes.
  if (conn->queryPreferences()->getFilteredOutput() && !filtered_output) {
    createFilteredWidget();
    scroll();
    return;
  }

  if (!conn->queryPreferences()->getFilteredOutput() && filtered_output) {
    destroyFilteredWidget();
    scroll();
    return;
  }
}

void PapayaVT::keypressEvent(GtkWidget * widget, GdkEventKey * event) {
  papaya_keypress(widget, event);

  // We need to detect which widget currently has focus.  If it's not the
  // input, then we need to switch focus and append the character.

  GtkArg arg;
  arg.name = "has_focus";
  
#ifdef USE_GTK_2
  if (GTK_WIDGET_HAS_FOCUS(widget))
    return;
#else
  gtk_widget_get(GTK_WIDGET(getInput()), &arg);
  if (GTK_VALUE_BOOL(arg))
    return;
#endif

  focus();

  // append the character.
  gchar * current_input = (gchar *)gtk_entry_get_text(GTK_ENTRY(getInput()));
  gchar * new_input = (gchar *)g_malloc(strlen(current_input) + 2);
  snprintf(new_input, strlen(current_input) + 2, "%s%c", current_input, event->keyval);
  setInput(new_input);
  g_free(new_input);

}

int papaya_keypress(GtkWidget * widget, GdkEventKey * event) {

  Connection * c = mainWindow->getCurrentConnection();
  if (!c)
    return 0;

  /* If this is line mode, check for command history. */
  if (c->getVT()->getMode() == VTLineMode) {
    switch(event->keyval) {

    case P:
      if (!(event->state & GDK_CONTROL_MASK))
	break;
    case UP_ARROW:
      c->getVT()->setInput(c->getHistory()->getPrev());
      return 1;

    case N:
      if (!(event->state & GDK_CONTROL_MASK))
	break;
    case DOWN_ARROW:
      c->getVT()->setInput(c->getHistory()->getNext());
      return 1;

    /* Check for scroll up/down. */
  case PAGE_UP:
    c->getVT()->scrollUp();
    break;
  case PAGE_DOWN:
    c->getVT()->scrollDown();
    }

    return 1;
  }

  if (event->keyval == ENTER || event->keyval == KP_ENTER) {
    c->getSocket()->write("\r\0", 2);
    return 1;
  }

  if (event->string && strlen(event->string) > 0) {
    c->getSocket()->write(event->string, strlen(event->string));
    return 1;
  }

  return 0;
}

int papaya_input_entry_activate(GtkEditable * editable, gpointer data) {

  Connection * conn = mainWindow->getCurrentConnection();
  if (!conn)
    return 0;

  if (conn->getVT()->getMode() == VTCharacterMode) {
    conn->getVT()->setInput("");
    return 1;
  }

  PapayaVT * pvt = (PapayaVT *)conn->getVT();

  gchar * text = (gchar *)gtk_entry_get_text(GTK_ENTRY(pvt->getInput()));

  if (!commandInterpreter)
    commandInterpreter = new CommandInterpreter();

  commandInterpreter->interpret(conn, text);
  return 1;
}

void PapayaVT::setMode(VTMode m) {

  VT::setMode(m);

  if (getMode() == VTCharacterMode)
    hideInput();
  
  if (getMode() == VTLineMode)
    showInput();

  focus();
}

void PapayaVT::focus() {
  switch (getMode()) {
  case VTCharacterMode:
    gtk_widget_grab_focus(output);
    break;
  case VTLineMode:
    gtk_widget_grab_focus(input);
    break;
  }
}

void PapayaVT::hideCommands() {
  gtk_entry_set_visibility(GTK_ENTRY(input), false);
}

void PapayaVT::showCommands() {
  gtk_entry_set_visibility(GTK_ENTRY(input), true);
}

bool PapayaVT::showingCommands() {
#ifdef USE_GTK_2
  return GTK_WIDGET_VISIBLE(GTK_ENTRY(input)->editable);
#else
  return GTK_ENTRY(input)->editable.visible;
#endif
}

void PapayaVT::scrollUp() {
  gtk_text_freeze(GTK_TEXT(output));
  GtkAdjustment * adj = GTK_RANGE(vscrollbar)->adjustment;
  
  float position = adj->value - adj->page_size;
  if (position < 0)
    position = 0;
  
  gtk_adjustment_clamp_page(adj, position, position + adj->page_size);
  gtk_text_thaw(GTK_TEXT(output));
}

void PapayaVT::scrollDown() {
 gtk_text_freeze(GTK_TEXT(output));
 GtkAdjustment * adj = GTK_RANGE(vscrollbar)->adjustment;
 
 float position = adj->value + adj->page_size;
 if (position > adj->upper)
   position = 0;
 
 gtk_adjustment_clamp_page(adj, position, position + adj->page_size);
 gtk_text_thaw(GTK_TEXT(output));
}
