#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef WIN32
#include <dlfcn.h>
#include <unistd.h>
#endif

#include <dirent.h>

#include <stdlib.h>
#include <string.h>

#include "system.h"
#include "mudclient.h"

#include "PapayaList.h"
#include "PluginData.h"
#include "PluginHandler.h"
#include "callbacks.h"

#include "Win32PluginAPITable.h"
#include "Win32PluginAPIImplementation.h"

#ifdef USE_GTK_2
void plugin_selection(GtkTreeView * view, void * arg1, GtkTreeViewColumn * column, gpointer data);
#endif

extern PluginHandler * phandler;

void on_plugin_ok_clicked(GtkButton * button, gpointer data);
void on_plugin_destroyed(GtkWidget * widget, gpointer data);

PluginHandler::PluginHandler() {
  first_handler = NULL;
  availablePlugins = new PapayaList();
  pluginList = new PapayaList();
  module_message[0] = '\0';
}

PluginHandler::~PluginHandler() {
  saveConfigFile();
  unloadPlugins();

  PapayaListElement * tmp_next;
  for (PapayaListElement * tmp = availablePlugins->firstElement; tmp; tmp = tmp_next) {
    tmp_next = tmp->getNext();
    PluginData * pd = (PluginData *)tmp->getData();
    delete pd;
    tmp->setData(NULL);
    availablePlugins->deleteEntry(tmp);
  }
    
  delete pluginList;
  delete availablePlugins;
}

void PluginHandler::loadPlugins() {
  // Loads the data from the lib/papaya/plugins/ directory.
  loadPluginData();
  // Loads the user's plugin configuration file and actually loads the plugins.
  loadConfigFile();
  // See if there are any unknown plugins for the user to enable.
  checkPlugins();

  // Display any error messages?
  if (strlen(module_message) > 0) {
    new Message(_("Error"), module_message, true);
    module_message[0] = '\0';
  }
}

void PluginHandler::unloadPlugins() {
  struct handler_entry * tmp, * tmp_next;

  // Unreference the plugins first, just so that they can't be called when
  // the shared object has been closed.

  for (tmp = first_handler; tmp; tmp = tmp_next) {
    tmp_next = tmp->next;
    // Call the plugin_cleanup function
    dl_call_func(tmp->handler, "plugin_cleanup");

    // Close the plugin - does this call global dtors?
#ifndef MALLOC
    g_module_close(tmp->handler);
#endif
    free(tmp);
  }

  // This should delete all the data structures in each Plugin item.
  //  pluginList->deleteAll();

}

/**
 * Goes through the new plugins in turn, asking the user if they
 * should be enabled.
 */

void PluginHandler::checkPlugins() {

  for (PapayaListElement * tmp = availablePlugins->firstElement; tmp; tmp = tmp->getNext()) {
    PluginData * pd = (PluginData *)tmp->getData();

    if (!pd->seen) {
      checkToLoad(pd, false);
	  continue;
    }
  }
}

#ifdef USE_GTK
static void on_load_button_clicked(GtkButton * button, gpointer data) {
  PluginData * pd = (PluginData *)data;

  pd->enable(TRUE);
  pd->seen = 1;

  gtk_widget_hide(pd->dialog);
  gtk_widget_destroy(pd->dialog);
  pd->dialog = NULL;
}

static void on_tp_load_button_clicked(GtkButton * button, gpointer data) {
  on_load_button_clicked(button, data);

  // Loads the data from the lib/papaya/plugins/ directory.
  phandler->loadPluginData();
  // Loads the user's plugin configuration file and actually loads the plugins.
  phandler->loadConfigFile();
  // See if there are any unknown plugins for the user to enable.
  phandler->checkPlugins();
}

static void on_dont_load_button_clicked(GtkButton * button, gpointer data) {
  PluginData * pd = (PluginData *)data;
  pd->enable(FALSE);
  pd->seen = 1;

  gtk_widget_hide(pd->dialog);
  gtk_widget_destroy(pd->dialog);
  pd->dialog = NULL;
}

static void on_tp_dont_load_button_clicked(GtkButton * button, gpointer data) {
  on_dont_load_button_clicked(button, data);

  // Loads the data from the lib/papaya/plugins/ directory.
  phandler->loadPluginData();
  // Loads the user's plugin configuration file and actually loads the plugins.
  phandler->loadConfigFile();
  // See if there are any unknown plugins for the user to enable.
  phandler->checkPlugins();
}
#else
static void on_gtk2_turf_plugin_response(GtkDialog * dialog, gint response, gpointer data) {
  PluginData * pd = (PluginData *)data;

  gtk_widget_hide(GTK_WIDGET(dialog));
  gtk_widget_destroy(GTK_WIDGET(dialog));

  if (response == GTK_RESPONSE_YES)
    pd->enable(true);
  else
    pd->enable(false);
 
  pd->seen = true;

  // Loads the data from the lib/papaya/plugins/ directory.
  phandler->loadPluginData();
  // Loads the user's plugin configuration file and actually loads the plugins.
  phandler->loadConfigFile();
  // See if there are any unknown plugins for the user to enable.
  phandler->checkPlugins();
}

static void on_gtk2_plugin_response(GtkDialog * dialog, gint response, gpointer data) {
  PluginData * pd = (PluginData *)data;
  
  if (response == GTK_RESPONSE_YES)
    pd->enable(true);
  else
    pd->enable(false);

  pd->seen = true;

  //  gtk_widget_hide(GTK_WIDGET(dialog));
  //  gtk_widget_destroy(GTK_WIDGET(dialog));

}
#endif

void PluginHandler::checkToLoad(PluginData * pd, bool tp) {

  GtkWidget *new_plugin;
#ifdef USE_GTK
  GtkWidget *vbox1;
  GtkWidget *new_plugin_text;
  GtkWidget *hbuttonbox1;
  GtkWidget *load_button;
  GtkWidget *dont_load_button;
#endif

  char buf[1024];

  sprintf(buf, _("A new plugin %s (version %d.%d) is available.  Do you wish to load it?"), pd->getName(), pd->getMajor(), pd->getMinor());

#ifdef USE_GTK_2
  new_plugin = gtk_message_dialog_new(NULL,
				      GTK_DIALOG_DESTROY_WITH_PARENT,
				      (GtkMessageType)GTK_MESSAGE_QUESTION,
				      GTK_BUTTONS_YES_NO,
				      buf);

  if (tp)
    g_signal_connect_data(G_OBJECT(new_plugin), "response",
			  G_CALLBACK(on_gtk2_turf_plugin_response), 
			  (gpointer) pd,
			  NULL,
			  (GConnectFlags) 0);
  else
    g_signal_connect_data(G_OBJECT(new_plugin), "response",
			  G_CALLBACK(on_gtk2_plugin_response), 
			  (gpointer) pd,
			  NULL,
			  (GConnectFlags) 0);

  gtk_dialog_run(GTK_DIALOG(new_plugin));
  if (!tp)
    gtk_widget_destroy(new_plugin);
#else
  new_plugin = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_object_set_data (GTK_OBJECT (new_plugin), "new_plugin", new_plugin);
  gtk_window_set_title (GTK_WINDOW (new_plugin), "New Plugin Available");
  gtk_window_set_position (GTK_WINDOW (new_plugin), GTK_WIN_POS_CENTER);
  gtk_window_set_modal (GTK_WINDOW (new_plugin), TRUE);

  vbox1 = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (vbox1);
  gtk_object_set_data_full (GTK_OBJECT (new_plugin), "vbox1", vbox1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (vbox1);
  gtk_container_add (GTK_CONTAINER (new_plugin), vbox1);


  new_plugin_text = gtk_label_new(buf);
  gtk_widget_ref (new_plugin_text);
  gtk_object_set_data_full (GTK_OBJECT (new_plugin), "new_plugin_text", new_plugin_text,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (new_plugin_text);
  gtk_box_pack_start (GTK_BOX (vbox1), new_plugin_text, TRUE, TRUE, 0);
  gtk_label_set_justify (GTK_LABEL (new_plugin_text), GTK_JUSTIFY_RIGHT);
  gtk_label_set_line_wrap (GTK_LABEL (new_plugin_text), TRUE);
  gtk_misc_set_alignment (GTK_MISC (new_plugin_text), 0, 7.45058e-09);
  gtk_misc_set_padding (GTK_MISC (new_plugin_text), 4, 4);

  hbuttonbox1 = gtk_hbutton_box_new ();
  gtk_widget_ref (hbuttonbox1);
  gtk_object_set_data_full (GTK_OBJECT (new_plugin), "hbuttonbox1", hbuttonbox1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (hbuttonbox1);
  gtk_box_pack_start (GTK_BOX (vbox1), hbuttonbox1, FALSE, FALSE, 0);
  gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox1), GTK_BUTTONBOX_SPREAD);

  load_button = gtk_button_new_with_label (_("Load"));
  gtk_widget_ref (load_button);
  gtk_object_set_data_full (GTK_OBJECT (new_plugin), "load_button", load_button,
			    (GtkDestroyNotify) gtk_widget_unref);

  gtk_widget_show (load_button);
  gtk_container_add (GTK_CONTAINER (hbuttonbox1), load_button);
  GTK_WIDGET_SET_FLAGS (load_button, GTK_CAN_DEFAULT);

  dont_load_button = gtk_button_new_with_label (_("Don't Load"));
  gtk_widget_ref (dont_load_button);
  gtk_object_set_data_full (GTK_OBJECT (new_plugin), "dont_load_button", dont_load_button,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (dont_load_button);
  gtk_container_add (GTK_CONTAINER (hbuttonbox1), dont_load_button);
  GTK_WIDGET_SET_FLAGS (dont_load_button, GTK_CAN_DEFAULT);

  if (!tp) {
#ifdef USE_GTK_2
    g_signal_connect_data(load_button, "clicked",
			  GTK_SIGNAL_FUNC(on_load_button_clicked),
			  pd,
			  NULL,
			  (GConnectFlags)0);

    g_signal_connect_data(dont_load_button, "clicked",
			  GTK_SIGNAL_FUNC(on_dont_load_button_clicked),
			  pd,
			  NULL,
			  (GConnectFlags)0);
#else
    gtk_signal_connect (GTK_OBJECT (load_button), "clicked",
			GTK_SIGNAL_FUNC (on_load_button_clicked),
			pd);
    gtk_signal_connect (GTK_OBJECT (dont_load_button), "clicked",
			GTK_SIGNAL_FUNC (on_dont_load_button_clicked),
			pd);
#endif
  } else {
#ifdef USE_GTK_2
    g_signal_connect_data(load_button, "clicked",
			  GTK_SIGNAL_FUNC(on_tp_load_button_clicked),
			  pd,
			  NULL,
			  (GConnectFlags)0);

    g_signal_connect_data(dont_load_button, "clicked",
			  GTK_SIGNAL_FUNC(on_tp_dont_load_button_clicked),
			  pd,
			  NULL,
			  (GConnectFlags)0);
#else
    gtk_signal_connect (GTK_OBJECT (load_button), "clicked",
			GTK_SIGNAL_FUNC (on_tp_load_button_clicked),
			pd);
    gtk_signal_connect (GTK_OBJECT (dont_load_button), "clicked",
			GTK_SIGNAL_FUNC (on_tp_dont_load_button_clicked),
			pd);
#endif
  }

  pd->dialog = new_plugin;
  gtk_widget_show(new_plugin);
#endif
}

void PluginHandler::unloadFile(PluginData * pd) {

  // Look for this plugin's handler entry.
  GModule * handler = findPlugin(pd->getName());

  // Call the plugin's cleanup function.
  dl_call_func(handler, "plugin_cleanup");

  // The Plugin MUST remove itself from the plugin list, or we go bang.
  
  // Remove the handler entry.
  if (first_handler->handler == handler)
    first_handler = first_handler->next;
  else {
    struct handler_entry * tmp;
    struct handler_entry * tmp_next;
    for (tmp = first_handler; tmp; tmp = tmp_next) {
      tmp_next = tmp->next;
      
      if (handler == tmp_next->handler) {
	tmp->next = tmp_next->next;
	break;
      }
    }
  }

  // Finally close the link to the shared object.
#ifndef MALLOC
  g_module_close(handler);
#endif
}

#ifdef WIN32
static plugin_address_table_t pat;

int socket_write(Socket *, char *, int);
void vt_append(VT *, char *);

void init_pat() {
	pat.p_register_plugin = register_plugin;
	pat.p_unregister_plugin = unregister_plugin;
	
	pat.p_connection_get_socket = connection_get_socket;
	pat.p_connection_get_vt = connection_get_vt;
    pat.p_connection_get_name = connection_get_name;
	pat.p_connection_get_mud = connection_get_mud;
	pat.p_connection_get_next = connection_get_next;
	pat.p_connection_query_preferences = connection_query_preferences;

    pat.p_socket_write = socket_write;
	
	pat.p_vt_append = vt_append;
	pat.p_vt_add_to_tray = vt_add_to_tray;
	pat.p_vt_remove_from_tray = vt_remove_from_tray;
    pat.p_vt_scroll = vt_scroll;

	pat.p_event_get_type = event_get_type;
	
	pat.p_preferences_set_preference = preferences_set_preference;
	pat.p_preferences_get_preference = preferences_get_preference;

	pat.p_get_prefix = get_prefix;
	pat.p_get_global_preferences = get_global_preferences;
    pat.p_get_command_interpreter = get_command_interpreter;
	pat.p_get_first_connection = get_first_connection;
	pat.p_get_main_window = get_main_window;
	pat.p_get_entity_handler = get_entity_handler;
	pat.p_get_plugin_handler = get_plugin_handler;
	pat.p_get_connection_by_name = get_connection_by_name;

	pat.p_main_window_get_current_connection = main_window_get_current_connection;
	pat.p_main_window_get_item_factory = main_window_get_item_factory;

	pat.p_entity_handler_add_entity = entity_handler_add_entity;
	pat.p_entity_handler_remove_entity = entity_handler_remove_entity;

	pat.p_plugin_handler_client_message = plugin_handler_client_message;

	pat.p_command_interpreter_interpret = command_interpreter_interpret;

    pat.p_mud_get_preferences = mud_get_preferences;

	pat.p_turf_protocol_add_command = turf_protocol_add_command;
	pat.p_turf_protocol_is_supported = turf_protocol_is_supported;
}

#endif

void PluginHandler::loadFile(PluginData * pd) {
  char * name = pd->getFilename();
// @@ move this elsewhere.
#ifdef WIN32
  init_pat();
#endif

  splash_update("Loading Plugin: %s", pd->getName());

  GModule * handler = g_module_open(name, (GModuleFlags)0);
  if (!handler) {
    const gchar * msg = g_module_error();
    if (msg) {
      printf (_("Failed to open %s: %s."), name, msg);
      printf(CRLF);
      strcat(module_message, msg);
      strcat(module_message, "\n");
    }
    return;
  }

  void (*init_function)(void *);

  init_function = dl_get_init_function(handler);
#ifdef WIN32
  init_function(&pat);
#else
  init_function(NULL);
#endif
  addHandlerEntry(handler);
}

void PluginHandler::addHandlerEntry(GModule * handler) {

  struct handler_entry * entry = (struct handler_entry *)malloc(sizeof(struct handler_entry));

  entry->next = first_handler;
  entry->handler = handler;

  first_handler = entry;
}

void PluginHandler::unregisterPlugin(Plugin * pl) {
  // Remove this plugin from the plugin list.

  PapayaListElement * ele = pluginList->findEntry(pl);
  if (ele) {
    ele->setData(NULL);
    pluginList->deleteEntry(ele);
  }
  else {
    printf(_("CRITICAL - Could not remove plugin from plugin list.  Expect crash."));
    printf(CRLF);
  }
}

void PluginHandler::registerPlugin(Plugin * pl, char * version) {

  // Check the 'mudclient' VERSION embedded in the Plugin to be installed.

  if (strcmp(VERSION, version)) {
    char buf[16384];
    sprintf(buf, _("WARNING: Plugin version (%s) doesn't match client version (%s)."), version, VERSION);
    new Message(_("Information"), buf, true);
  }

  pluginList->newEntry(pl);

  // Generate a Connected event for every MUD window open at the moment.
  Event * e = new Event(EvConnect, NULL);
  for (Connection * tmp = first_connection; tmp; tmp = tmp->getNext()) {
    pl->onEvent(e, tmp);
  }
  delete e;

}

void PluginHandler::outputFilters(Connection * c, char * in) {

  for (PapayaListElement * tmp = pluginList->firstElement; tmp; tmp = tmp->getNext()) {
    Plugin * p = (Plugin *)tmp->getData();
    if (!p)
      continue;
	    
    p->output(c, in);

    if (strlen(in) == 0)
      return;
  }
}

void PluginHandler::promptFilters(Connection * c, char * in) {
  for (PapayaListElement * tmp = pluginList->firstElement; tmp; tmp = tmp->getNext()) {
    Plugin * p = (Plugin *)tmp->getData();
    if (!p)
      continue;
	    
    p->prompt(c, in);

    if (strlen(in) == 0)
      return;
  }
}

void PluginHandler::inputFilters(Connection * c, char * in) {
  for (PapayaListElement * tmp = pluginList->firstElement; tmp; tmp = tmp->getNext()) {
    Plugin * p = (Plugin *)tmp->getData();
    if (!p)
      continue;
	    
    p->input(c, in);

    if (strlen(in) == 0)
      return;

  }
}

void PluginHandler::clientMessage(Connection * c, char * in) {
  for (PapayaListElement * tmp = pluginList->firstElement; tmp; tmp = tmp->getNext()) {
    Plugin * p = (Plugin *)tmp->getData();
    if (!p)
      continue;
	    
    p->clientMessage(c, in);

    if (strlen(in) == 0)
      return;

  }
}

void PluginHandler::fireEvent(Event * e, Connection * c) {

  for (PapayaListElement * tmp = pluginList->firstElement; tmp; tmp = tmp->getNext()) {
    Plugin * p = (Plugin *)tmp->getData();
    if (!p)
      continue;

    p->onEvent(e, c);
  }
}

void PluginHandler::pageSwitched() {
	printf ("PluginHandler::pageSwitched.\n");
	
	for (PapayaListElement * tmp = pluginList->firstElement; tmp; tmp = tmp->getNext()) {
		Plugin * p = (Plugin *)tmp->getData();
		if (p)
			p->pageSwitched();
	}
}

/**
 * This function loads the PluginData config file from the user's home
 * directory, and enables/disables the appropriate plugins.
 */

void PluginHandler::loadConfigFile() {
  char plugin[1024];
  int status;
  int major, minor;
  int personal = 0;

  FILE * fp = openFile("plugins.conf", NULL, "r");
  if (!fp)
    return;

  while (!feof(fp)) {
    int res = fscanf(fp, "%s %d.%d %d %d\n", plugin, &major, &minor, &personal, &status);
    if (res != 5) {
      res = fscanf(fp, "%s %d.%d %d\n", plugin, &major, &minor, &status);
      if (res != 4)
	continue; // Possibly blank config file or malformed config file.
    }

    for (PapayaListElement * ele = availablePlugins->firstElement; ele; ele = ele->getNext()) {
      PluginData * pd = (PluginData *)ele->getData();

	// Right name?  Right location?
	if (!strcmp(pd->getName(), plugin)
	    && ((personal && !strstr(pd->getFilename(), getPrefix()))
		|| (!personal && strstr(pd->getFilename(), getPrefix()))
		)
	    ) {
	  
	  if (pd->getMajor() > major || pd->getMajor() == major && pd->getMinor() > minor) {
	    char buf[1024];
	    snprintf(buf, 1024, _("Plugin %s is now at version %d.%d"), plugin, pd->getMajor(), pd->getMinor());
	    /* The user has never used this version before. */
	    new Message(_("Information"), buf, false);
	  }
	  
	  pd->enable((bool)status);
	  pd->seen = TRUE;
	  break;
	}

	  }
  }

  fclose(fp);
}

/**
 * Saves the PluginData config file to the user's home directory.
 */

void PluginHandler::saveConfigFile() {

  FILE * fp = openFile("plugins.conf", NULL, "w");
  if (!fp) {
    new Message(_("Error"), _("Papaya was unable to save your personal plugin configuration file."), true);
    return;
  }

  for (PapayaListElement * tmp = availablePlugins->firstElement; tmp; tmp = tmp->getNext()) {
    PluginData * pd = (PluginData *)tmp->getData();
    int personal = 0;
    if (!strstr(pd->getFilename(), getPrefix())) // Is this a personal plugin?
      personal = 1;

    fprintf(fp, "%s %d.%d %d %d\n", pd->getName(), pd->getMajor(), pd->getMinor(), personal, pd->enabled());
  }

  fclose(fp);
}

#ifdef USE_GTK_2

static GtkListStore * plugin_list_store;

GtkWidget * PluginHandler::create_plugin_list() {

  GtkTreeViewColumn *column;
  GtkWidget * plugin_view;
  GtkCellRenderer * cell_renderer;

  GtkWidget * window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(window), "Plugins");
  gtk_window_set_resizable(GTK_WINDOW(window), true);
  gtk_widget_show(window);

  plugin_list_store = gtk_list_store_new(5, 
					 G_TYPE_STRING,
					 G_TYPE_STRING,
					 G_TYPE_STRING,
					 G_TYPE_STRING,
					 G_TYPE_STRING);
  populate_plugin_list();

  plugin_view = gtk_tree_view_new_with_model(GTK_TREE_MODEL(plugin_list_store));

  g_signal_connect_data(plugin_view, "row-activated",
			GTK_SIGNAL_FUNC(plugin_selection), NULL,
			NULL, (GConnectFlags)0);


  //  g_object_unref(plugin_list_store);

  cell_renderer = gtk_cell_renderer_text_new();

  // Now how to put the data in the tree_view
  column = gtk_tree_view_column_new_with_attributes(_("Name"),
						    cell_renderer,
						    "text", 0,
						    NULL);


  gtk_tree_view_append_column(GTK_TREE_VIEW(plugin_view), column);

  column = gtk_tree_view_column_new_with_attributes(_("Version"),
						    cell_renderer,
						    "text", 1,
						    NULL);


  gtk_tree_view_append_column(GTK_TREE_VIEW(plugin_view), column);

  column = gtk_tree_view_column_new_with_attributes(_("Location"),
						    cell_renderer,
						    "text", 2,
						    NULL);


  gtk_tree_view_append_column(GTK_TREE_VIEW(plugin_view), column);

  column = gtk_tree_view_column_new_with_attributes(_("Description"),
						    cell_renderer,
						    "text", 3,
						    NULL);


  gtk_tree_view_append_column(GTK_TREE_VIEW(plugin_view), column);

  column = gtk_tree_view_column_new_with_attributes(_("Enabled?"),
						    cell_renderer,
						    "text", 4,
						    NULL);


  gtk_tree_view_append_column(GTK_TREE_VIEW(plugin_view), column);



  gtk_widget_show(plugin_view);
  gtk_container_add(GTK_CONTAINER(window), plugin_view);
  return window;
}

void PluginHandler::populate_plugin_list() {

  GtkTreeIter iter;

  // Nuke the contents of the plugin list.
  gtk_list_store_clear(plugin_list_store);

  for (PapayaListElement * tmp = availablePlugins->firstElement; tmp; tmp = tmp->getNext()) {
    PluginData * p = (PluginData *)tmp->getData();

    
    gtk_list_store_append(plugin_list_store,
			  &iter);
    
    char version_buf[1024];
    snprintf(version_buf, 1024, "%d.%d", p->getMajor(), p->getMinor());
    
    gtk_list_store_set(plugin_list_store, &iter,
		       0, p->getName(),
		       1, version_buf,
		       2, strstr(p->getFilename(), getPrefix()) ? "global" : "Personal",
		       3, p->getDescription(),
		       4, p->enabled() ? "Enabled" : "Disabled",
		       -1);
  
  }
}

#endif

static GtkWidget * plugin_list = NULL;

void PluginHandler::displayAvailablePlugins() {

  if (plugin_list) {
    gtk_widget_show(plugin_list);
    return;
  }

  plugin_list = create_plugin_list();
  gtk_widget_show(plugin_list);
}

void on_plugin_destroyed(GtkWidget * widget, gpointer data) {
  phandler->saveConfigFile();

  plugin_list = NULL;
}

void on_plugin_ok_clicked(GtkButton * button, gpointer data) {
  gtk_widget_hide(plugin_list);
  gtk_widget_destroy(plugin_list);
}


#ifndef USE_GTK_2

void plugin_toggled(GtkWidget * checkbutton, gpointer data) {

  PluginData * pd = (PluginData *)data;

  if (GTK_TOGGLE_BUTTON(checkbutton)->active)
    pd->enable(TRUE);
  else
    pd->enable(FALSE);

}

static char * plugin_clist_title[] = {
  {"Plugin"}, 
  {"Version"}, 
  {"Source"},
  {"Description"},
  {"Status"}
};

void PluginHandler::populate_plugin_clist(GtkWidget * plugin_clist) {

  char * data[5];
  char version_buf[1024];

  gtk_clist_clear(GTK_CLIST(plugin_clist));

  for (PapayaListElement * tmp = availablePlugins->firstElement; tmp; tmp = tmp->getNext()) {
    PluginData * p = (PluginData *)tmp->getData();

    data[0] = strdup(p->getName());
    
    snprintf(version_buf, 1024, "%d.%d", p->getMajor(), p->getMinor());
    data[1] = version_buf;

    if (strstr(p->getFilename(), getPrefix()))
      data[2] = strdup(_("Global"));
    else
      data[2] = strdup(_("Personal"));

    data[3] = strdup(p->getDescription());

    if (p->enabled())
      data[4] = strdup(_("Enabled"));
    else
      data[4] = strdup(_("Disabled"));

    gtk_clist_append(GTK_CLIST(plugin_clist), data);

    free(data[0]);
    free(data[2]);
    free(data[3]);
    free(data[4]);

  }
}

#endif

#ifdef USE_GTK_2

void plugin_selection(GtkTreeView * view, void * arg1, GtkTreeViewColumn * column, gpointer data) {

  GtkTreeSelection * selection = gtk_tree_view_get_selection(view);
  GtkTreeIter iter;
  GtkTreeModel * model;

  if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
    gchar * name;
    gtk_tree_model_get(model, &iter, 0, &name, -1);

    printf ("Double clicked plugin %s.\n", name);
    phandler->toggleEnabled((char *)name);
  }
}

void PluginHandler::toggleEnabled(char * name) {
  for (PapayaListElement * tmp = availablePlugins->firstElement; tmp; tmp = tmp->getNext()) {
    PluginData * p = (PluginData *)tmp->getData();
    if (!strcmp(p->getName(), name)) {
      p->enable(!p->enabled());
      phandler->populate_plugin_list();
      return;
    }
  }
}

#else

void plugin_selection(GtkWidget * plugin_clist, gint row, gint column, GdkEventButton * event, gpointer data) {

  if (!event)
    return;

  phandler->selection(plugin_clist, row, column, event);
}
#endif

#ifdef USE_GTK
void PluginHandler::selection(GtkWidget * plugin_clist, gint row, gint column, GdkEventButton * event) {

  if (event->type == GDK_2BUTTON_PRESS) {
    // Enable/Disable this plugin.

    int count = 0;
    for (PapayaListElement * tmp = availablePlugins->firstElement; tmp; tmp = tmp->getNext()) {
      PluginData * p = (PluginData *)tmp->getData();

      if (count == row) {
	  p->enable(!p->enabled());
	  phandler->populate_plugin_clist(plugin_clist);
	  return;
      }

      count++;
    }
  }
}

GtkWidget * PluginHandler::create_plugin_list() {

  GtkWidget * plugin_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(plugin_window), "Papaya: available plugins");

  GtkWidget * box = gtk_vbox_new(FALSE, 2);
  gtk_widget_show(box);
  gtk_container_add(GTK_CONTAINER(plugin_window), box);
  
  // Create the list widget.
  GtkWidget * plugin_clist = gtk_clist_new(5);
  
  // Add and display the column labels.
  int i;
  for (i = 0; i < 5; i++) {
    GtkWidget * label = gtk_label_new(plugin_clist_title[i]);
    gtk_clist_set_column_widget(GTK_CLIST(plugin_clist), i, label);
    gtk_widget_show(label);
  }
  gtk_clist_column_titles_show(GTK_CLIST(plugin_clist));

  // Populate the plugin clist
  populate_plugin_clist(plugin_clist);

  // Set the optimal column widths now that the clist is populated.
  for (i = 0; i < 5; i++)
    gtk_clist_set_column_width(GTK_CLIST(plugin_clist), i, gtk_clist_optimal_column_width(GTK_CLIST(plugin_clist), i));

  // Pack the clist widget into the box.
  gtk_box_pack_start(GTK_BOX(box), plugin_clist, TRUE, TRUE, 0);
  gtk_widget_show(plugin_clist);

  // Create the close button thing.
  GtkWidget * plugin_ok = gtk_button_new_with_label(_("Ok"));
  gtk_widget_show(plugin_ok);

  gtk_signal_connect(GTK_OBJECT(plugin_clist), "select_row", 
		     GTK_SIGNAL_FUNC(plugin_selection), NULL);
  gtk_signal_connect (GTK_OBJECT (plugin_ok), "clicked",
                      GTK_SIGNAL_FUNC (on_plugin_ok_clicked),
                      NULL);
  gtk_signal_connect(GTK_OBJECT(plugin_window), "destroy",
		     GTK_SIGNAL_FUNC(on_plugin_destroyed), NULL);

  // Create a button box to stick the close button in.
  GtkWidget * buttonbox = gtk_hbutton_box_new();
  gtk_widget_show(buttonbox);
  gtk_box_pack_start(GTK_BOX(box), buttonbox, TRUE, TRUE, 0);

  gtk_container_add(GTK_CONTAINER(buttonbox), plugin_ok);

  return plugin_window;
}

#endif

/**
 * Yet more new plugin loading code.
 */

void PluginHandler::loadPluginData() {
  char dirname[1024];

#ifdef WIN32
  // @@ is getPrefix()/plugins the same as plugins?
  
//  snprintf(dirname, 1024, "plugins");
//  checkDirForPlugins(dirname);
#else
  // Check all plugins in user's dir
  char * home = getenv("HOME");
  if (home) {
    // Don't just use sprintf, user might have $HOME set funny
    snprintf(dirname, 1024, "%s/.papaya/plugins", home);
    dirname[1023] = '\0';
    checkDirForPlugins(dirname);
  }
#endif

  // Check all plugins in systemwide dir
#ifdef WIN32
  snprintf(dirname, 1024, "%s/plugins", getPrefix());
#else
  snprintf(dirname, 1024, "%s/lib/papaya/plugins", getPrefix());
#endif
  checkDirForPlugins(dirname);

  return;
}

void PluginHandler::checkDirForPlugins(char * dirname) {
	DIR * dir;
  struct dirent * entry;

  printf ("Checking %s for plugins.\n", dirname);

  dir = opendir(dirname);
  if (!dir)
    return;

  while ((entry = readdir(dir))) {
    if (!strcmp(entry->d_name, ".") || !strcmp(entry->d_name, ".."))
      continue;
    
    PluginData * pd = queryPlugin(entry->d_name, dirname);
    if (pd)
      availablePlugins->newEntry(pd);
  }

  closedir(dir);
}

PluginData * PluginHandler::queryPlugin(char * n, char * path) {
  char name[1024];
  char description[1024];
  char minor[1024];
  char major[1024];
  char fname[1024];

  struct stat file_stat;

  if (path) {
    snprintf(fname, 1024, "%s/%s", path, n);
  } else {
    snprintf(fname, 1024, "%s/lib/papaya/plugins/%s", getPrefix(), n);
  }

  /* Ensure that the requested file exists. */
  if (stat(fname, &file_stat) == -1)
    return NULL;

  GModule * handle = g_module_open(fname, (GModuleFlags)0);
  if (!handle) {
    const gchar * msg = g_module_error();
    if (msg) {
      printf (_("Failed to open %s: %s"), fname, msg);
      printf(CRLF);
      strcat(module_message, msg);
      strcat(module_message, "\n");
    }
    return NULL;
  }

  dl_call_func_ret(handle, "plugin_query_name", name);
  dl_call_func_ret(handle, "plugin_query_description", description);
  dl_call_func_ret(handle, "plugin_query_major", major);
  dl_call_func_ret(handle, "plugin_query_minor", minor);

#ifndef MALLOC  
  g_module_close(handle);
#endif

  PluginData * pd = new PluginData(name);
  pd->setDescription(description);
  pd->setMajor(major);
  pd->setMinor(minor);
  pd->setFilename(fname);
  
  return pd;
}

void PluginHandler::getPrefsWidgets(GtkNotebook * prefs_notebook, MUD * mud) {
  for (PapayaListElement * tmp = pluginList->firstElement; tmp; tmp = tmp->getNext()) {
    Plugin * p = (Plugin *)tmp->getData();
  
    if (!p)
      continue;

    GtkWidget * widget = p->getPrefsWidget(mud);
    if (!widget)
      continue;
    
    GtkWidget * label = gtk_label_new(p->getName());
    gtk_widget_show(label);
    gtk_widget_show(widget);
    gtk_notebook_append_page(prefs_notebook, widget, label);
  }
}

void PluginHandler::onPrefsApply(MUD * mud) {
  for (PapayaListElement * tmp = pluginList->firstElement; tmp; tmp = tmp->getNext()) {
    Plugin * p = (Plugin *)tmp->getData();
  
    if (!p)
      continue;

    p->onPrefsApply(mud);
  }
}

void PluginHandler::onPrefsOk(MUD * mud) {
  for (PapayaListElement * tmp = pluginList->firstElement; tmp; tmp = tmp->getNext()) {
    Plugin * p = (Plugin *)tmp->getData();
  
    if (!p)
      continue;

    p->onPrefsOk(mud);
  }
}

void PluginHandler::savePrefs(FILE * fp, MUD * mud) {
  for (PapayaListElement * tmp = pluginList->firstElement; tmp; tmp = tmp->getNext()) {
    Plugin * p = (Plugin *)tmp->getData();
  
    if (!p)
      continue;

    p->savePrefs(fp, mud);
  }
}

void PluginHandler::onPrefsCancel(MUD * mud) {
  for (PapayaListElement * tmp = pluginList->firstElement; tmp; tmp = tmp->getNext()) {
    Plugin * p = (Plugin *)tmp->getData();
  
    if (!p)
      continue;

    p->onPrefsCancel(mud);
  }
}

GModule * PluginHandler::findPlugin(char * name) {
  char plugin_name[1024];
  struct handler_entry * tmp;

  for (tmp = first_handler; tmp; tmp = tmp->next) {
    dl_call_func_ret(tmp->handler, "plugin_query_name", plugin_name);
    if (!strcmp(name, plugin_name))
      return tmp->handler;
  }
  return NULL;
}

extern "C" GModule * plugin_handler_find_plugin(char * name) {
	return phandler->findPlugin(name);
}

void * PluginHandler::findPlugin(char * name, char * func) {
  char plugin_name[1024];
  struct handler_entry * tmp;

  for (tmp = first_handler; tmp; tmp = tmp->next) {
    dl_call_func_ret(tmp->handler, "plugin_query_name", plugin_name);
    if (!strcmp(name, plugin_name))
      return findFunction(tmp->handler, func);
  }
  return NULL;
}

