#include <glib.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#ifdef WIN32
//#include "DBCSSTR.H"
#define strcasecmp stricmp
#define strncasecmp strnicmp
#endif

#include "mudclient.h"
#include "SystemColour.h"
#include "MUD.h"

#include "Prefs.h"

#define FREE(x) if (x) free (x);
#define DELETE(x) if (x) delete x;

Prefs::Prefs(MUD * mud) {
  pref_list = NULL;

  echo = false;
  beep = true;
  promptOnInputLine = true;
  doublePrompt = false;
  semicolon = false;
  papayavt = true;
  history_ignore_repeats = 0;
  history_honour_null = 0;
  history_use_limit = false;
  history_limit = 500;
  history_circular = true;

  compression = true;
  compression_off_command = strdup("config -mccp");
  mudftp = true;
  socks = false;
  socks4 = false;
  socks4_pass = strdup("");
  socks4_user = strdup("guest");
  socks4_port = strdup("1080");
  socks4_host = strdup("socks");
  mcw_start = true;

  browser = strdup("mozilla");

  negotiate_charmode = true;

  http_proxy = false;
  http_proxy_port = strdup("8080");
  http_proxy_host = strdup("wwwcache.lancs.ac.uk");

  http_connect = false;
  
  turf_httpd = false;

  triggers = true;
  system_triggers = true;

  scroll = true;
  useMaxLines = false;
  maxLines = 500;

  filtered = false;

  colour = true;
  fg_colour = 7; // White default foreground
  bg_colour = -1; // Default scheme bg
  colours_loaded = 0;

  for (int i = 0; i < 16; i++) {
    colours[i][0] = 0;
    colours[i][1] = 0;
    colours[i][2] = 0;
  }

  bg_r = 58664;
  bg_g = 58664;
  bg_b = 58664;

  usePromptTerminator = false;
  promptTerminator = strdup(">>>");

#ifdef USE_GTK
#  ifndef WIN32
  input_font = strdup("-*-courier-medium-r-normal-*-*-140-*-*-*-*-*-*");
  output_font = strdup("-*-courier-medium-r-normal-*-*-140-*-*-*-*-*-*");
  prompt_font = strdup("-*-courier-medium-r-normal-*-*-140-*-*-*-*-*-*");
  bold_font = strdup("-*-courier-medium-r-normal-*-*-140-*-*-*-*-*-*");
  use_bold_font = FALSE;
#  else
  input_font = strdup("-*-Courier New-normal-r-normal-*-*-140-*-*-m-*-*-*");
  output_font = strdup("-*-Courier New-normal-r-normal-*-*-140-*-*-m-*-*-*");
  prompt_font = strdup("-*-Courier New-normal-r-normal-*-*-140-*-*-m-*-*-*");
  bold_font = strdup("-*-Courier New-normal-r-normal-*-*-140-*-*-m-*-*-*");
  use_bold_font = FALSE;
#  endif
#else
  output_font = strdup("monospace medium 13");
  input_font = strdup("monospace medium 13");
  bold_font = strdup("monospace bold 13");
  prompt_font = strdup("monospace medium 13");
  use_bold_font = TRUE;
#endif

  toolbar = true; // By default enable the toolbar

  inputFont = (GdkFont *)NULL;
  outputFont = (GdkFont *)NULL;
  promptFont = (GdkFont *)NULL;
  boldFont = (GdkFont *)NULL;

#ifdef USE_GTK
  loadFont(input_font, &inputFont);
  loadFont(output_font, &outputFont);
  loadFont(prompt_font, &promptFont);
  loadFont(bold_font, &boldFont);
#endif

  multilineprompt = 0;
  autoreconnect = false;
  window_x = 640;
  window_y = 480;

  send_keepalive = false;
  keepalive_timer = 10;

  send_version = true;
  send_stats = true;

  // put NOTHING after these

  load(mud);
  colour_table = new SystemColour(this);

  setPreference("charset", "ISO-8859-1");
}

Prefs::~Prefs() {

  FREE(socks4_user);
  FREE(socks4_pass);
  FREE(socks4_host);
  FREE(socks4_port);

  FREE(http_proxy_port);
  FREE(http_proxy_host);

  FREE(input_font);
  FREE(output_font);
  FREE(prompt_font);
  FREE(bold_font);

  FREE(compression_off_command);
  FREE(browser);

  FREE(promptTerminator);

  if (inputFont)
    gdk_font_unref(inputFont);
  if (outputFont)
    gdk_font_unref(outputFont);
  if (promptFont)
    gdk_font_unref(promptFont);
  if (boldFont)
    gdk_font_unref(boldFont);

  DELETE(colour_table);

  // Nuke all the custom preferences
  struct pref_data * tmp_next;
  for (struct pref_data * tmp = pref_list; tmp; tmp = tmp_next) {
    tmp_next = tmp->next;
    
    FREE(tmp->name);
    FREE(tmp->data);
    FREE(tmp);
  }

}

void Prefs::loadFont(char * string, GdkFont ** ptr) {

  if (!string)
    return;

  if (*ptr)
    gdk_font_unref(*ptr);

  *ptr = gdk_font_load(string);
  if (!*ptr) {
    *ptr = gdk_font_load(DEFAULT_FONT);
    if (!*ptr)
      new Message(_("Error"), _("Papaya was not able to load the font specified in the preferences and the default 'failsafe' font was unavailable.  Please edit this connection's preferences and choose another font."), true);
  }
}

void Prefs::setMultiLinePrompt(int b) {
  multilineprompt = b;
}

int Prefs::getMultiLinePrompt() {
  return multilineprompt;
}

void Prefs::setInputFont(char * c) {
  if (input_font)
    free(input_font);

  input_font = strdup(c);

#ifdef USE_GTK
  loadFont(input_font, &inputFont);
#endif
}

void Prefs::setOutputFont(char * c) {
  if (output_font)
    free(output_font);

  output_font = strdup(c);

#ifdef USE_GTK
  loadFont(output_font, &outputFont);
#endif
}

void Prefs::setBoldFont(char * c) {
  if (bold_font)
    free(bold_font);

  bold_font = strdup(c);

#ifdef USE_GTK
  loadFont(bold_font, &boldFont);
#endif
}

void Prefs::setPromptFont(char * c) {
  if (prompt_font)
    free(prompt_font);

  prompt_font = strdup(c);

#ifdef USE_GTK
  loadFont(prompt_font, &promptFont);
#endif
}


GdkFont * Prefs::getOutput() {
	return outputFont;
}

GdkFont * Prefs::getBold() {
    if (use_bold_font)
	return boldFont;
    else
        return outputFont;
}

GdkFont * Prefs::getInput() {
	return inputFont;
}

GdkFont * Prefs::getPrompt() {
	return promptFont;
}

char * Prefs::getInputFont() {
  return input_font;
}

char * Prefs::getOutputFont() {
  return output_font;
}

char * Prefs::getPromptFont() {
  return prompt_font;
}

char * Prefs::getBoldFont() {
  return bold_font;
}

bool Prefs::getUseBoldFont() {
  return use_bold_font;
}

void Prefs::setUseBoldFont(bool b) {
  use_bold_font = b;
}

void Prefs::setEcho(bool b){
  echo = b;
}

bool Prefs::getEcho(){
  return echo;
}

void Prefs::setBeep(bool b){
  beep = b;
}

bool Prefs::getBeep() {
  return beep;
}

void Prefs::setNegotiateCharmode(bool b) {
  negotiate_charmode = b;
}

bool Prefs::getNegotiateCharmode() {
  return negotiate_charmode;
}

void Prefs::setPromptOnInputLine(bool b) {
  promptOnInputLine = b;
}

bool Prefs::getPromptOnInputLine() {
  return promptOnInputLine;
}

void Prefs::setMudftp(bool b) {
  mudftp = b;
}

bool Prefs::getMudftp() {
  return mudftp;
}

void Prefs::setCompression(bool b) {
  compression = b;
}

bool Prefs::getCompression() {
  return compression;
}

void Prefs::setMCW(bool b) {
  mcw_start = b;
}

bool Prefs::getMCW() {
  return mcw_start;
}


void Prefs::setCompressionOffCommand(char * c) {
  if (compression_off_command)
    free(compression_off_command);
  compression_off_command = strdup(c);
}

char * Prefs::getCompressionOffCommand() {
  return compression_off_command;
}

void Prefs::setSocks4(bool b) {
  socks4 = b;
}

bool Prefs::getSocks4() {
  return socks4;
}

void Prefs::setSocks(bool b) {
  socks = b;
}

bool Prefs::getSocks() {
  return socks;
}

void Prefs::setSocksUser(char * n) {
  if (socks4_user)
    free(socks4_user);

  socks4_user = strdup(n);
}

void Prefs::setSocksPass(char * n) {
  if (socks4_pass)
    free(socks4_pass);

  socks4_pass = strdup(n);
}

void Prefs::setSocksHost(char * n) {
  if (socks4_host)
    free(socks4_host);

  socks4_host = strdup(n);
}

void Prefs::setSocksPort(char * n) {
  if (socks4_port)
    free(socks4_port);

  socks4_port = strdup(n);
}

char * Prefs::getSocksUser() {
  return socks4_user;
}

char * Prefs::getSocksPass() {
  return socks4_pass;
}

char * Prefs::getSocksHost() {
  return socks4_host;
}

char * Prefs::getSocksPort() {
  return socks4_port;
}

void Prefs::setTriggers(bool b) {
  triggers = b;
}

bool Prefs::getTriggers() {
  return triggers;
}

bool Prefs::getSystemTriggers() {
  return system_triggers;
}

void Prefs::setSystemTriggers(bool b) {
  system_triggers = b;
}

void Prefs::setUsePromptTerminator(bool b) {
  usePromptTerminator = b;
}

bool Prefs::getUsePromptTerminator() {
  return usePromptTerminator;
}

void Prefs::setPromptTerminator(char * term) {
  if (promptTerminator)
    free(promptTerminator);

  promptTerminator = strdup(term);
}

// This may return NULL.

char * Prefs::getPromptTerminator() {
  return promptTerminator;
}

void Prefs::setColour(bool b) {
  colour = b;
}

bool Prefs::getColour() {
  return colour;
}

bool Prefs::getLoadedColours() {
  return colours_loaded;
}

void Prefs::setColours(int i, int r, int g, int b) {
  colours[i][0] = r;
  colours[i][1] = g;
  colours[i][2] = b;
}

void Prefs::setColours(char * line) {

  char * pc = line;
  char * pc2 = (char *)NULL;

  int r, g, b;

  for (int i = 0; i < 16; i++) {
    pc2 = strchr(pc, ' ');
    if (!pc2) {
      debug(_("Couldn't find space character in line."));
      debug(CRLF);
      abort();
    }

    *pc2 = '\0';

    // Stuff between pc2 and pc is what we want.
    if (sscanf(pc, "%d,%d,%d", &r, &g, &b) != 3) {
      debug(_("Failed to read colours from config file."));
      debug(CRLF);
      abort();
    }

    colours[i][0] = r;
    colours[i][1] = g;
    colours[i][2] = b;
    pc = pc2 + 1;
  }

  colours_loaded = 1;
}

void Prefs::getColours(int index, int * r, int * g, int * b) {
  if (index > 15 || index < 0) {
    debug(_("Invalid index in getColours(int, int *, int *, int *)"));
    abort();
  }

  *r = colours[index][0];
  *g = colours[index][1];
  *b = colours[index][2];
}

char * Prefs::getColours() {
  static char buf[16384];
  char buf2[16384];

  buf[0] = '\0';

  for (int i=0; i<16; i++) {
    snprintf(buf2, 16384, "%d,%d,%d ", colours[i][0], colours[i][1], colours[i][2]);
    strcat(buf, buf2);
  }

  return buf;
}

struct colour_table * Prefs::getColourTable() {
  return colour_table->table;
}

SystemColour * Prefs::getSysColour() {
  return colour_table;
}

void Prefs::setDefaultForegroundColour(int c) {
  fg_colour = c;
}

int Prefs::getDefaultForegroundColour() {
  return fg_colour;
}

void Prefs::setDefaultBackgroundColour(int c) {
  bg_colour = c;
}

int Prefs::getDefaultBackgroundColour() {
  return bg_colour;
}

void Prefs::setScreenBackground(char * str) {

  int r;
  int g;
  int b;

  if ((sscanf(str, "%d %d %d", &r, &g, &b)) != 3) {
    debug(_("Bad background colours."));
    debug(CRLF);
    return;
  }

  setScreenBackground(r, g, b);
}

void Prefs::setScreenBackground(int r, int g, int b) {
  bg_r = r;
  bg_g = g;
  bg_b = b;
}

int Prefs::getScreenBackgroundRed() {
  return bg_r;
}

int Prefs::getScreenBackgroundGreen() {
  return bg_g;
}

int Prefs::getScreenBackgroundBlue() {
  return bg_b;
}

void Prefs::setScrollOnInput(bool b) {
  scroll = b;
}

bool Prefs::getScrollOnInput() {
  return scroll;
}

void Prefs::setSemiColon(bool b) {
  semicolon = b;
}

bool Prefs::getSemiColon() {
  return semicolon;
}

void Prefs::setUsePapayaVT(bool b) {
  papayavt = b;
}

bool Prefs::getUsePapayaVT() {
  return papayavt;
}

void Prefs::setHistoryUseLimit(bool b) {
  history_use_limit = b;
}

bool Prefs::getHistoryUseLimit() {
  return history_use_limit;
}

void Prefs::setHistoryLimit(int i) {
  history_limit = i;
}

int Prefs::getHistoryLimit() {
  return history_limit;
}

void Prefs::setHistoryHonourNull(bool b) {
  history_honour_null = b;
}

bool Prefs::getHistoryHonourNull() {
  return history_honour_null;
}

void Prefs::setHistoryIgnoreRepeats(bool b) {
  history_ignore_repeats = b;
}

bool Prefs::getHistoryIgnoreRepeats() {
  return history_ignore_repeats;
}

bool Prefs::getHistoryIsCircular() {
  return history_circular;
}

void Prefs::setHistoryIsCircular(bool b) {
  history_circular = b;
}

void Prefs::setUseMaxLines(bool b) {
  useMaxLines = b;
}

bool Prefs::getUseMaxLines() {
  return useMaxLines;
}

void Prefs::setMaxLines(int i) {
  maxLines = i;
}

int Prefs::getMaxLines() {
  return maxLines;
}

void Prefs::setAutoReconnect(bool t) {
  autoreconnect = t;
}

bool Prefs::getAutoReconnect() {
  return autoreconnect;
}

void Prefs::setTurfHttpd(bool b) {
  turf_httpd = b;
}

bool Prefs::getTurfHttpd() {
  return turf_httpd;
}

void Prefs::setHttpConnect(bool b) {
  http_connect = b;
}

bool Prefs::getHttpConnect() {
  return http_connect;
}

void Prefs::setHttpProxy(bool b) {
  http_proxy = b;
}

void Prefs::setHttpProxyHost(char * h) {
  if (http_proxy_host)
    free(http_proxy_host);
  http_proxy_host = strdup(h);
}

void Prefs::setHttpProxyPort(char * p) {
  if (http_proxy_port)
    free(http_proxy_port);
  http_proxy_port = strdup(p);
}

bool Prefs::getHttpProxy() {
  return http_proxy;
}

char * Prefs::getHttpProxyHost() {
  return http_proxy_host;
}

char * Prefs::getHttpProxyPort() {
  return http_proxy_port;
}

void Prefs::setFilteredOutput(bool b) {
  filtered = b;
}

bool Prefs::getFilteredOutput() {
  return filtered;
}

void Prefs::setWindowX(int x) {
  window_x = x;
}

void Prefs::setWindowY(int y) {
  window_y = y;
}

int Prefs::getWindowX() {
  return window_x;
}

int Prefs::getWindowY() {
  return window_y;
}

void Prefs::setDoublePrompt(bool t) {
  doublePrompt = t;
}

bool Prefs::getDoublePrompt() {
  return doublePrompt;
}

bool Prefs::getSendKeepalive() {
  return send_keepalive;
}

void Prefs::setSendKeepalive(bool b) {
  send_keepalive = b;
}

int Prefs::getKeepaliveTimer() {
  return keepalive_timer;
}

void Prefs::setKeepaliveTimer(int t) {
  keepalive_timer = t;
}

bool Prefs::getSendStats() {
  return send_stats;
}

void Prefs::setSendStats(bool t) {
  send_stats = t;
}

bool Prefs::getSendVersion() {
  return send_version;
}

void Prefs::setSendVersion(bool t) {
  send_version = t;
}

void Prefs::setUseToolbar(bool t) {
  toolbar = t;
}

bool Prefs::getUseToolbar() {
  return toolbar;
}

void Prefs::setBrowser(char * b) {
  FREE(browser);

  if (!strstr(b, "%s")) {
    char buf[16384];
    snprintf(buf, 16384, "%s %%s", b);
    browser = strdup(buf);
  } else
    browser = strdup(b);
}

char * Prefs::getBrowser() {
  return browser;
}
    
struct pref_data * Prefs::findPreference(char * prefName) {
  for (struct pref_data * tmp = pref_list; tmp; tmp = tmp->next) {
    if (!strcmp(prefName, tmp->name))
      return tmp;
  }
  return NULL;
}

void Prefs::setPreference(char * prefName, char * prefData) {
  struct pref_data * data = findPreference(prefName);

  if (!data) {
    data = (struct pref_data *)malloc(sizeof(struct pref_data));
    memset(data, 0, sizeof(struct pref_data));
    data->name = strdup(prefName);
    data->next = pref_list;
    pref_list = data;
  }

  if (data->data)
    free(data->data);
  data->data = strdup(prefData);
}

char * Prefs::getPreference(char * prefName) {
  struct pref_data * data = findPreference(prefName);
  if (data)
    return data->data;

  return NULL;
}

void Prefs::saveCustomPrefs(FILE * fp) {
  for (struct pref_data * tmp = pref_list; tmp; tmp = tmp->next) {
    fprintf(fp, "Custom%s %s\n", tmp->name, (char *)tmp->data);
  }
}

void Prefs::set(char * key, char * value) {

  switch (toupper(key[0])) {
  case 'A':
    if (!strcasecmp("autoreconnect", key))
      setAutoReconnect((bool)atoi(value));
    break;

  case 'B':
    if (!strcasecmp("beep", key))
      setBeep((bool)atoi(value));
    if (!strcasecmp("bgcolour", key))
      setDefaultBackgroundColour(atoi(value));
    if (!strcasecmp("bgscreencolour", key))
      setScreenBackground(value);
    if (!strcasecmp("boldfont", key))
      setBoldFont(value);
    if (!strcasecmp("browser", key))
      setBrowser(value);
    break;

  case 'C':
    if (!strncasecmp("custom", key, 6))
      setPreference(key+6, value);
    if (!strcasecmp("compression", key))
      setCompression((bool)atoi(value));
    if (!strcasecmp("compressionoffcommand", key))
      setCompressionOffCommand(value);
    if (!strcasecmp("colour", key))
      setColour((bool)atoi(value));
    if (!strcasecmp("colours", key))
      setColours(value);
    break;

  case 'D':
    if (!strcasecmp("doubleprompt", key))
      setDoublePrompt((bool)atoi(value));
    break;

  case 'E':
    if (!strcasecmp("echo", key))
      setEcho((bool)atoi(value));
    break;

  case 'F':
    if (!strcasecmp("fgcolour", key))
      setDefaultForegroundColour(atoi(value));
    if (!strcasecmp("filtered", key))
      setFilteredOutput(atoi(value));
    break;

  case 'H':
    if (!strcasecmp("httpconnect", key))
      setHttpConnect((bool)atoi(value));
    if (!strcasecmp("httpproxy", key))
      setHttpProxy((bool)atoi(value));
    if (!strcasecmp("httpproxyport", key))
      setHttpProxyPort(value);
    if (!strcasecmp("httpproxyhost", key))
      setHttpProxyHost(value);
    if (!strcasecmp("historyuselimit", key))
      setHistoryUseLimit((bool)atoi(value));
    if (!strcasecmp("historylimit", key))
      setHistoryLimit(atoi(value));
    if (!strcasecmp("historyignorerepeats", key))
      setHistoryIgnoreRepeats((bool)atoi(value));
    if (!strcasecmp("historyhonournull", key))
      setHistoryHonourNull((bool)atoi(value));
    if (!strcasecmp("historycircular", key))
      setHistoryIsCircular((bool)atoi(value));
    break;

  case 'I':
    if (!strcasecmp("inputfont", key))
      setInputFont(value);
    break;

  case 'K':
    if (!strcasecmp("keepalivetimer", key))
      setKeepaliveTimer(atoi(value));
    break;

  case 'M':
    if (!strcasecmp("mudftp", key))
      setMudftp((bool)atoi(value));
    if (!strcasecmp("maxlines", key))
      setMaxLines(atoi(value));
    if (!strcasecmp("multilineprompt", key))
      setMultiLinePrompt(atoi(value));
    if (!strcasecmp("mcw", key))
      setMCW((bool)atoi(value));
    break;

  case 'N':
    if (!strcasecmp("negotiatecharmode", key))
      setNegotiateCharmode((bool)atoi(value));
    break;

  case 'O':
    if (!strcasecmp("outputfont", key))
      setOutputFont(value);
    break;

  case 'P':
    if (!strcasecmp("prompt", key))
      setPromptOnInputLine((bool)atoi(value));
    if (!strcasecmp("promptfont", key))
      setPromptFont(value);
    if (!strcasecmp("papayavt", key))
      setUsePapayaVT((bool)atoi(value));
#ifdef USE_GTK_2
    if (!strcasecmp("pangoinputfont", key))
      setInputFont(value);
    if (!strcasecmp("pangooutputfont", key))
      setOutputFont(value);
    if (!strcasecmp("pangoboldoutputfont", key))
      setBoldFont(value);
    if (!strcasecmp("pangopromptfont", key))
      setPromptFont(value);
#endif
    if (!strcasecmp("promptterminator", key))
      setPromptTerminator(value);

    break;

  case 'S':
    if (!strcasecmp("scroll", key))
      setScrollOnInput((bool)atoi(value));
    if (!strcasecmp("semicolon", key))
      setSemiColon((bool)atoi(value));
    if (!strcasecmp("sendkeepalive", key))
      setSendKeepalive((bool)atoi(value));
    if (!strcasecmp("socks", key))
      setSocks((bool)atoi(value));
    if (!strcasecmp("socks4", key))
      setSocks4((bool)atoi(value));
    if (!strcasecmp("socks4user", key))
      setSocksUser(value);
    if (!strcasecmp("socks4pass", key))
      setSocksPass(value);
    if (!strcasecmp("socks4host", key))
      setSocksHost(value);
    if (!strcasecmp("socks4port", key))
      setSocksPort(value);
    if (!strcasecmp("sendversion", key))
      setSendVersion((bool)atoi(value));
    if (!strcasecmp("sendstats", key))
      setSendStats((bool)atoi(value));
    if (!strcasecmp("systemtriggers", key))
      setSystemTriggers((bool)atoi(value));
    break;

  case 'T':
    if (!strcasecmp("triggers", key))
      setTriggers((bool)atoi(value));
    if (!strcasecmp("turfhttpd", key))
      setTurfHttpd((bool)atoi(value));
    if (!strcasecmp("toolbar", key))
      setUseToolbar((bool)atoi(value));
    break;

  case 'U':
    if (!strcasecmp("useboldfont", key))
      setUseBoldFont((bool)atoi(value));
    if (!strcasecmp("usemaxlines", key))
      setUseMaxLines((bool)atoi(value));
    if (!strcasecmp("usepromptterminator", key))
      setUsePromptTerminator((bool)atoi(value));
    break;

  case 'W':
    if (!strcasecmp("windowx", key))
      setWindowX(atoi(value));
    if (!strcasecmp("windowy", key))
      setWindowY(atoi(value));
    break;

  default:
    debug(_("WARNING: Unrecognised key %s in config file."), key);
    debug(CRLF);
    break;
  }

}

bool Prefs::parseLine(FILE * fp) {

  char buffer[4096];
  char key[1024];
  char value[1024];

  while (fgets(buffer, 4096, fp)) {
    if (sscanf(buffer, "%1023s %1023[^\n]\n", key, value) != 2) {
      // No longer return false, we accept non-matching strings as having
      // null values.  Hopefully only string-based preferences will not match.
      set(key, "");
      continue;
    }

    set(key, value);
  }

  return FALSE;
}

bool Prefs::load(MUD * mud) {

  FILE * fp = (FILE *)NULL;

  char filename[1024];

  if (!mud || !mud->getOwnPrefs())
    snprintf(filename, 1024, "papayarc");
  else
    snprintf(filename, 1024, "prefs/%s", mud->getName());

  // Open papayarc from user's home dir, or system install dir
  fp = openFile(filename, (char *)NULL, "r");
  if (!fp) {
    debug(_("WARNING: couldn't load preferences."));
    debug(CRLF);
    return FALSE;
  }

  while (parseLine(fp));
  fclose(fp);
  return TRUE;
}


bool Prefs::save(MUD * mud) {
  
  char filename[1024];
  
  if (mud && mud->getPreferences())
    snprintf(filename, 1024, "prefs/%s", mud->getName());
  else
    snprintf(filename, 1024, "papayarc");

  return save(filename);
}

bool Prefs::save(char * filename) {
  FILE * fp = (FILE *)NULL;

  // Open file in user's ~/.papaya directory to write to
  fp = openFile(filename, (char *)NULL, "w");
  if (!fp) {
    perror("fopen");
    debug(_("WARNING: couldn't save preferences"));
    debug(CRLF);
    return FALSE;
  }

  fprintf(fp, "Autoreconnect %d\n", autoreconnect);
  fprintf(fp, "Echo %d\n", echo);
  fprintf(fp, "Beep %d\n", beep);
  fprintf(fp, "BGScreenColour %d %d %d\n", bg_r, bg_g, bg_b);
  fprintf(fp, "Prompt %d\n", promptOnInputLine);
  fprintf(fp, "PapayaVT %d\n", papayavt);
  fprintf(fp, "NegotiateCharmode %d\n", negotiate_charmode);
  fprintf(fp, "Compression %d\n", compression);
  fprintf(fp, "CompressionOffCommand %s\n", compression_off_command);
  fprintf(fp, "Filtered %d\n", filtered);
  fprintf(fp, "Socks %d\n", socks);
  fprintf(fp, "Socks4 %d\n", socks4);
  fprintf(fp, "Socks4Pass %s\n", socks4_pass);
  fprintf(fp, "Socks4User %s\n", socks4_user);
  fprintf(fp, "Socks4Host %s\n", socks4_host);
  fprintf(fp, "Socks4Port %s\n", socks4_port);

  fprintf(fp, "HTTPConnect %d\n", http_connect);
  fprintf(fp, "HTTPProxy %d\n", http_proxy);
  fprintf(fp, "HTTPProxyHost %s\n", http_proxy_host);
  fprintf(fp, "HTTPProxyPort %s\n", http_proxy_port);

  fprintf(fp, "TurfHTTPd %d\n", turf_httpd);
  fprintf(fp, "SendVersion %d\n", send_version);
  fprintf(fp, "SendStats %d\n", send_stats);

  fprintf(fp, "MCW %d\n", mcw_start);
  fprintf(fp, "Triggers %d\n", triggers);
  fprintf(fp, "SystemTriggers %d\n", system_triggers);
  fprintf(fp, "Colour %d\n", colour);
  fprintf(fp, "FGColour %d\n", fg_colour);
  fprintf(fp, "BGColour %d\n", bg_colour);
  fprintf(fp, "Colours %s\n", getColours());
  fprintf(fp, "Mudftp %d\n", mudftp);
  fprintf(fp, "Scroll %d\n", scroll);
  fprintf(fp, "UseMaxLines %d\n", useMaxLines);
  fprintf(fp, "MaxLines %d\n", maxLines);
  fprintf(fp, "Semicolon %d\n", semicolon);

  fprintf(fp, "InputFont %s\n", input_font);
  fprintf(fp, "OutputFont %s\n", output_font);
  fprintf(fp, "BoldFont %s\n", bold_font);
  fprintf(fp, "PromptFont %s\n", prompt_font);

  fprintf(fp, "UseBoldFont %d\n", use_bold_font);

  fprintf(fp, "MultiLinePrompt %d\n", multilineprompt);

  fprintf(fp, "WindowX %d\n", getWindowX());
  fprintf(fp, "WindowY %d\n", getWindowY());

  fprintf(fp, "HistoryUseLimit %d\n", getHistoryUseLimit());
  fprintf(fp, "HistoryLimit %d\n", getHistoryLimit());
  fprintf(fp, "HistoryIgnoreRepeats %d\n", getHistoryIgnoreRepeats());
  fprintf(fp, "HistoryHonourNull %d\n", getHistoryHonourNull());
  fprintf(fp, "HistoryCircular %d\n", getHistoryIsCircular());
  fprintf(fp, "SendKeepalive %d\n", getSendKeepalive());
  fprintf(fp, "KeepaliveTimer %d\n", getKeepaliveTimer());

  fprintf(fp, "Browser %s\n", getBrowser());

  fprintf(fp, "UsePromptTerminator %d\n", getUsePromptTerminator());
  fprintf(fp, "PromptTerminator %s\n", getPromptTerminator());

  fprintf(fp, "Toolbar %d\n", getUseToolbar());

  saveCustomPrefs(fp);

  fclose(fp);
  return TRUE;
}
