#ifdef WIN32
#include <windows.h>
#include <io.h>
#endif

#include <sys/stat.h>
#include <sys/types.h>
#include <signal.h>

#include "system.h"
#include "mudclient.h"

#include "interface.h"
#include "support.h"

#include "Socket.h"

#include "Connection.h"
#include "Prefs.h"
#include "Message.h"

#include "callbacks.h"
#include "PluginHandler.h"
#include "EntityHandler.h"

#include "MUD.h"
#include "MUDSelector.h"
#include "MUDConnectorList.h"

#include "GTKTwoWindow.h"

#include "libusers.h"

#define PACKAGE "papaya"

MUDSelector * mud_selector;
MUDConnectorList * mcl;
BaseWindow * mainWindow;
EntityHandler * entities;
PluginHandler * phandler;
Prefs * globalPreferences;

#ifdef USE_GTK_2
GtkWidget * debug_view = NULL;
GtkTextBuffer * debug_buffer = NULL;
#else
 GtkWidget * debug_text = NULL;
#endif
GtkWidget * debug_window = NULL;

void debug_init() {
  debug_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(debug_window), "Papaya Debug");
  gtk_widget_show(debug_window);

#ifdef USE_GTK_2
  debug_buffer = gtk_text_buffer_new(NULL);
  debug_view = gtk_text_view_new_with_buffer(debug_buffer);
  gtk_widget_show(debug_view);
  
  gtk_widget_set_name(debug_view, "debug_view");

  gtk_container_add(GTK_CONTAINER(debug_window), debug_view);
#else
  debug_text = gtk_text_new(NULL, NULL);
  gtk_widget_show(debug_text);
  gtk_container_add(GTK_CONTAINER(debug_window), debug_text);
#endif

  while (gtk_events_pending())
    gtk_main_iteration();
}

void debug(char * str, ...) {
  /*
  if (!globalPreferences->getDisplayDebugMessages())
    return;
  */
  char buf[16384];
  va_list ap;

  va_start(ap, str);
#ifdef WIN32
#define vsnprintf _vsnprintf
#endif
  vsnprintf(buf, 16384, str, ap);
  va_end(ap);

#ifdef USE_GTK_2
  GtkTextIter start, end;
  gtk_text_buffer_get_bounds(debug_buffer, &start, &end);
  gtk_text_buffer_insert(debug_buffer, &end, buf, -1);
#else
  gtk_text_insert(GTK_TEXT(debug_text), NULL, NULL, NULL, buf, -1);
#endif

  while (gtk_events_pending())
    gtk_main_iteration();
}

static GtkWidget * splash_window = NULL;
static GtkWidget * splash_label = NULL;
static GtkWidget * splash_progress = NULL;

#define SPLASH_PIXMAP "splash.xpm"

void splash_create() {
  /*
  if (!globalPreferences->getUseSplashScreen())
    return;
  */

  GtkWidget * splash_pixmap;

  splash_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  //  gtk_window_set_decorated(GTK_WINDOW(splash_window), false);

#ifdef USE_GTK
  gtk_object_set_data(GTK_OBJECT(splash_window), "splash_window", splash_window);
#else
  g_object_set_data(G_OBJECT(splash_window), "splash_window", splash_window);
#endif
  gtk_widget_show(splash_window);

  GtkWidget * vbox = gtk_vbox_new(false, 2);
  gtk_widget_show(vbox);
  gtk_container_add(GTK_CONTAINER(splash_window), vbox);

  splash_pixmap = create_pixmap(splash_window, SPLASH_PIXMAP);
  gtk_widget_show(splash_pixmap);
  gtk_box_pack_start(GTK_BOX(vbox), splash_pixmap, true, true, 0);

  splash_label = gtk_label_new("");
  gtk_widget_show(splash_label);
  gtk_box_pack_start(GTK_BOX(vbox), splash_label, false, false, 0);

  splash_progress = gtk_progress_bar_new();
  //  gtk_widget_show(splash_progress);
  gtk_box_pack_start(GTK_BOX(vbox), splash_progress, false, false, 0);

  while (gtk_events_pending())
    gtk_main_iteration();

}

void splash_destroy() {
  if (splash_window) {
    gtk_widget_hide(splash_window);
    gtk_widget_destroy(splash_window);
    splash_window = NULL;
  }
}

void splash_update(gchar * text, ...) {
  /*
  if (!globalPreferences->getUseSplashScreen())
    return;
  */

  if (!splash_window)
    return;

  char buf[16384];
  va_list ap;

  va_start(ap, text);
  //#ifdef WIN32
  //#define vsnprintf _vsnprintf
  //#endif
  vsnprintf(buf, 16384, text, ap);
  va_end(ap);

  gtk_label_set_text(GTK_LABEL(splash_label), buf);

  debug("%s\n", buf);

  while (gtk_events_pending())
    gtk_main_iteration();
}

void splash_increment() {

  /*
  if (!globalPreferences->getUseSplashScreen())
    return;
  */

  if (!splash_window)
    return;
  // @@ Update the statusbar by one unit.

}

int papaya_startup(int argc, char * argv[]) {

  bool copyover = false;

  char buf[1024];

#ifdef WIN32
  snprintf(buf, 1024, "%s", getPrefix());
#else
  snprintf(buf, 1024, "%s/share/papaya/bitmaps", getPrefix());
#endif
  add_pixmap_directory("bitmaps");
  add_pixmap_directory(buf);

#ifndef WIN32
  // Ensure that we have a $HOME/.papaya/ directory if possible.
  char * home = getenv("HOME");
  if (home) {
    char filename[1024];
    snprintf(filename, 1024, "%s/.papaya", home);
    mkdir(filename, 0700);
  }
#endif

  setlocale(LC_ALL, "");
  bindtextdomain(PACKAGE, GNOMELOCALEDIR);
  textdomain(PACKAGE);

  gtk_init(&argc, &argv);

#if defined(sun)
  sigignore(SIGPIPE);
#endif

#ifndef WIN32
  signal(SIGCHLD, SIG_IGN);
  signal(SIGPIPE, SIG_IGN);
#endif

  debug_init();

  // Load in the global preferences so we know if the user wants a splash
  // screen or not.
  globalPreferences = new Prefs((MUD *)NULL);

  // Create the splash screen
  splash_create();

#ifdef WIN32
  splash_update(_("Initialising network"));
  WSADATA wsadata;
  if (WSAStartup(MAKEWORD(2,2), &wsadata) != 0) {
    new Message("Error", "Failed to initialise Windows Sockets.", true);
    exit(1);
  }
  splash_increment();
#endif
  
#ifdef USE_GTK_2
  struct stat stat_buf;
  
  //  if (globalPreferences->getLoadGTKRC()) {
  splash_update(_("Loading theme"));
#ifdef WIN32
  snprintf(buf, 1024, "gtkrc");
  if (stat(buf, &stat_buf) != 0) {
    snprintf(buf, 1024, "%s\\gtkrc", getPrefix());
  }
#else
  home = getenv("HOME");
  snprintf(buf, 1024, "%s/.papaya/gtkrc", home);
  if (stat(buf, &stat_buf) != 0) {
    snprintf(buf, 1024, "%s/share/papaya/gtkrc", getPrefix());
  }
#endif
  printf ("Loading theme from %s.\n", buf);
  gtk_rc_parse(buf);
  splash_increment();
  //  }
#endif
    
  splash_update(_("Loading MUD Connection Wizard data"));
  mud_selector = new MUDSelector();
  splash_increment();

  splash_update(_("Loading MUD Connector MUD list"));
  mcl = new MUDConnectorList();
  splash_increment();
  
  splash_update(_("Creating main window"));
  mainWindow = new GTKTwoWindow();
  splash_increment();

  splash_update(_("Loading Aliases and Triggers"));
  entities = new EntityHandler();
  entities->load();
  splash_increment();

  splash_update(_("Loading Plugins"));
  phandler = new PluginHandler();
  phandler->loadPlugins();
  splash_increment();

  splash_destroy();

  mainWindow->display();
  for (int i = 0; i < argc; i++) {
    if (!strcmp(argv[i], "copyover")) {
      copyover_recover();
      copyover = TRUE;
      break;
    }
  }

  if (!copyover && globalPreferences->getMCW())
    on_connection_wizard_button_clicked((GtkButton *)NULL, (gpointer)NULL, 0);

  gtk_main();
  return 1;
}

int papaya_shutdown() {

  if (globalPreferences->getSendStats()) {
    int count_socket = create_socket();
    disconnect_from_server(count_socket);
    close(count_socket);
  }

#if !defined(sun) || !defined(WIN32)
  signal(SIGSEGV, NULL);
#endif

  delete mainWindow;
  delete phandler;
  delete globalPreferences;
  delete mud_selector;
  delete mcl;

  entities->save();
  delete entities;

#ifdef WIN32
  WSACleanup();
#endif

  return 0;
}
