#include <gdk/gdk.h>
#include <string.h>
#include "SystemColour.h"

#ifdef WIN32
#include <memory.h>
#else
#include <unistd.h>
#endif

#include "Prefs.h"
#include "mudclient.h"

// This class MUST be instantiated AFTER preferences

// This is the default colour table used by SystemColour class

struct colour_table ctable[] = {
  { color_black, 0, 0, 0, "black" },
  { color_red, 0xcccc, 0x0000, 0x0000, "red" },
  { color_green, 0, 0xcccc, 0, "green" },
  { color_yellow, 0xcccc, 0xcccc, 0x0000, "yellow" },
  { color_blue, 0, 0, 0xcccc, "blue" },
  { color_magenta, 0xcccc, 0, 0xcccc, "magenta" },
  { color_cyan, 0, 0xcccc, 0xcccc, "cyan" },
  { color_white, 0xcccc, 0xcccc, 0xcccc, "white" },

  { color_bold_black, 0, 0, 0, "bold black" },
  { color_bold_red, 0xffff, 0x0000, 0x0000, "bold red" },
  { color_bold_green, 0, 0xffff, 0, "bold green" },
  { color_bold_yellow, 0xffff, 0xffff, 0x0000, "bold yellow" },
  { color_bold_blue, 0, 0, 0xffff, "bold blue" },
  { color_bold_magenta, 0xffff, 0, 0xffff, "bold magenta" },
  { color_bold_cyan, 0, 0xffff, 0xffff, "bold cyan" },
  { color_bold_white, 0xffff, 0xffff, 0xffff, "bold white" },
  { 0, 0, 0, 0, "" }
};

SystemColour::SystemColour(Prefs * p) {
  int r, g, b;

  memcpy(&table, &ctable, sizeof(colour_table) * 17);

  default_fg = 0;
  default_bg = -1; // Transparent.

  // Get the defaults from the preferences
  if (p->getLoadedColours()) {
    for (int i = 0; i < 16; i++) {
      p->getColours(i, &r, &g, &b);
      setColor(i, r, g, b);
    }
  } else {
    for (int i = 0; i < 16; i++)
      p->setColours(i, table[i].red, table[i].green, table[i].blue);
  }
}

SystemColour::~SystemColour() {

}

void SystemColour::setColor(char color, int r, int g, int b) {

  for (int i = 0; i < 8; i++) {
      if (table[i].colour_character == color) {
        // Set this colour as the new one.
        table[i].red = r;
        table[i].green = g;
        table[i].blue = b;
        return;
      }
  }
}

void SystemColour::setColor(int color, int r, int g, int b) {
  table[color].red = r;
  table[color].green = g;
  table[color].blue = b;
}
