#include "mudclient.h"
#include "PluginHandler.h"
#include "Prefs.h"
#include "Connection.h"
#include "CommandInterpreter.h"
#include "EntityHandler.h"
#include "BaseWindow.h"

extern Prefs * globalPreferences;
extern PluginHandler * phandler;
extern BaseWindow * mainWindow;
extern EntityHandler * entities;
extern Connection * first_connection;
extern CommandInterpreter * commandInterpreter;

void register_plugin(Plugin * p, char * ver) {
	phandler->registerPlugin(p, ver);
}

void unregister_plugin(Plugin * p) {
	phandler->unregisterPlugin(p);
}

Socket * connection_get_socket(Connection * c) {
	return c->getSocket();
}

VT * connection_get_vt(Connection * c) {
	return c->getVT();
}

char * connection_get_name(Connection * c) {
	return c->getName();
}

EvType event_get_type(Event * e) {
	return e->getType();
}

int socket_write(Socket * s, char * t, int len) {
	return s->write(t, len);
}

void vt_append(VT * vt, char * t) {
	vt->append(t);
}

void vt_add_to_tray(VT * vt, GtkWidget * w1, GtkWidget ** w2) {
	vt->addToTray(w1, w2);
}

void vt_remove_from_tray(VT * vt, GtkWidget * w1, GtkWidget * w2) {
	vt->removeFromTray(w1, w2);
}

char * get_prefix() {
	return getPrefix();
}

MUD * connection_get_mud(Connection * c) {
	return c->getMUD();
}

char * preferences_get_preference(Prefs * p, char * name) {
	return p->getPreference(name);
}

void preferences_set_preference(Prefs * p, char * name, char * value) {
	p->setPreference(name, value);
}

Prefs * get_global_preferences() {
	return globalPreferences;
}

Prefs * mud_get_preferences(MUD * m) {
	return m->getPreferences();
}

Connection * connection_get_next(Connection * c) {
	return c->getNext();
}

Connection * get_first_connection() {
	return first_connection;
}

CommandInterpreter * get_command_interpreter() {
	return commandInterpreter;
}

void command_interpreter_interpret(CommandInterpreter * ci, Connection * c, char * cmd) {
	ci->interpret(c, cmd);
}

BaseWindow * get_main_window() {
	return mainWindow;
}

EntityHandler * get_entity_handler() {
	return entities;
}

PluginHandler * get_plugin_handler() {
	return phandler;
}

Connection * main_window_get_current_connection(BaseWindow * w) {
	return w->getCurrentConnection();
}

GtkItemFactory * main_window_get_item_factory(BaseWindow * w) {
	return w->getItemFactory();
}

void entity_handler_add_entity(EntityHandler * eh, char * group, Entity * e) {
	eh->addEntity(group, e);
}

void entity_handler_remove_entity(EntityHandler * eh, Entity * e) {
	eh->removeEntity(e);
}

void plugin_handler_client_message(PluginHandler * ph, Connection * c, char * msg) {
	ph->clientMessage(c, msg);
}

void vt_scroll(VT * vt) {
	vt->scroll();
}

Connection * get_connection_by_name(char * c) {
	return findConnection(c);
}

Prefs * connection_query_preferences(Connection * c) {
	return c->queryPreferences();
}

#ifndef WIN32
void plugin_address_table_init(void *) {
}
#endif
