#include <gtk/gtk.h>

#include "Prefs.h"
#include "mudclient.h"
#include "SystemColour.h"
#include "ColourEditor.h"
#include "FontChooser.h"
#include "Connection.h"
#include "History.h"
#include "BaseWindow.h"

extern BaseWindow * mainWindow;

extern GtkWidget *prefs_terminal_echo;
extern GtkWidget *prefs_terminal_beep;
extern GtkWidget *prefs_terminal_prompt;
extern GtkWidget *prefs_terminal_double_prompt;
extern GtkWidget *prefs_terminal_scroll;
extern GtkWidget *prefs_terminal_semicolon;
extern GtkWidget *prefs_terminal_use_maxlines;
extern GtkWidget *prefs_terminal_maxlines;
extern GtkWidget *prefs_terminal_charset;
extern GtkWidget *prefs_terminal_filtered_output;
extern GtkWidget *prefs_browser;
extern GtkWidget *prefs_terminal_limit_history;
extern GtkWidget *prefs_terminal_max_history;
extern GtkWidget *prefs_terminal_history_repeat;
extern GtkWidget *prefs_terminal_history_tabnull;
extern GtkWidget *prefs_terminal_history_circular;
extern GtkWidget *prefs_terminal_negotiate_charmode;
extern GtkWidget *prefs_terminal_prompt_terminator;
extern GtkWidget *prefs_terminal_prompt_terminator_string;
extern GtkWidget *prefs_terminal_clear_input;
extern GtkWidget *prefs_connection_compression;
extern GtkWidget *prefs_connection_compression_off_command;
extern GtkWidget *prefs_connection_mudftp;
extern GtkWidget *prefs_connection_mcw;
extern GtkWidget *prefs_connection_send_keepalive;
extern GtkWidget *prefs_connection_keepalive;
extern GtkWidget *prefs_connection_autoreconnect;
extern GtkWidget *prefs_proxy_direct_radio;
extern GtkWidget *prefs_proxy_socks_radio;
extern GtkWidget *prefs_proxy_socks4_radio;
extern GtkWidget *prefs_proxy_socks5_radio;
extern GtkWidget *prefs_proxy_socks_pass;
extern GtkWidget *prefs_proxy_socks_user;
extern GtkWidget *prefs_proxy_socks_host;
extern GtkWidget *prefs_proxy_socks_port;
extern GtkWidget *prefs_proxy_httpd_radio;
extern GtkWidget *prefs_proxy_httpd_connect_radio;
extern GtkWidget *prefs_proxy_http;
extern GtkWidget *prefs_proxy_http_host;
extern GtkWidget *prefs_proxy_http_port;
extern GtkWidget *prefs_triggers_enable;
extern GtkWidget *prefs_colour_enable;
extern GtkWidget *prefs_colour_defbg_menu;
extern GtkWidget *prefs_colour_deffg_menu;

extern GtkWidget * prefs_stats_enable;
extern GtkWidget * prefs_stats_enable_version;

extern GtkWidget * prefs_gui_tooltips;
extern GtkWidget * prefs_gui_toolbar;

extern ColourEditor *white;
extern ColourEditor *red;
extern ColourEditor *green;
extern ColourEditor *blue;
extern ColourEditor *yellow;
extern ColourEditor *magenta;
extern ColourEditor *cyan;
extern ColourEditor *black;
extern ColourEditor *white_bold;
extern ColourEditor *red_bold;
extern ColourEditor *green_bold;
extern ColourEditor *blue_bold;
extern ColourEditor *yellow_bold;
extern ColourEditor *magenta_bold;
extern ColourEditor *cyan_bold;
extern ColourEditor *black_bold;
extern ColourEditor *background;

extern FontChooser * output_font_chooser;
extern FontChooser * bold_font_chooser;
extern FontChooser * input_font_chooser;
extern FontChooser * prompt_font_chooser;
extern GtkWidget * prefs_use_bold;

extern Prefs * globalPreferences;
extern Connection * first_connection;

void get_terminal_prefs(Prefs * preferences) {

  if (GTK_TOGGLE_BUTTON(prefs_terminal_echo)->active)
    preferences->setEcho(1);
  else
    preferences->setEcho(0);

  if (GTK_TOGGLE_BUTTON(prefs_terminal_beep)->active)
    preferences->setBeep(0);
  else
    preferences->setBeep(1);

  if (GTK_TOGGLE_BUTTON(prefs_terminal_prompt)->active)
    preferences->setPromptOnInputLine(1);
  else
    preferences->setPromptOnInputLine(0);

  if (GTK_TOGGLE_BUTTON(prefs_terminal_double_prompt)->active)
    preferences->setDoublePrompt(1);
  else
    preferences->setDoublePrompt(0);

  if (GTK_TOGGLE_BUTTON(prefs_terminal_semicolon)->active)
    preferences->setSemiColon(1);
  else
    preferences->setSemiColon(0);
  
  if (GTK_TOGGLE_BUTTON(prefs_terminal_scroll)->active)
    preferences->setScrollOnInput(1);
  else
    preferences->setScrollOnInput(0);

  if (GTK_TOGGLE_BUTTON(prefs_terminal_negotiate_charmode)->active)
    preferences->setNegotiateCharmode(true);
  else
    preferences->setNegotiateCharmode(false);

  if (GTK_TOGGLE_BUTTON(prefs_terminal_filtered_output)->active)
    preferences->setFilteredOutput(1);
  else
    preferences->setFilteredOutput(0);

  if (GTK_TOGGLE_BUTTON(prefs_terminal_limit_history)->active)
    preferences->setHistoryUseLimit(1);
  else
    preferences->setHistoryUseLimit(0);

  if (GTK_TOGGLE_BUTTON(prefs_terminal_clear_input)->active)
    preferences->setPreference("clearinput", "1");
  else
    preferences->setPreference("clearinput", "0");

  if (GTK_TOGGLE_BUTTON(prefs_terminal_history_circular)->active)
    preferences->setHistoryIsCircular(true);
  else
    preferences->setHistoryIsCircular(false);

  preferences->setHistoryLimit(gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(prefs_terminal_max_history)));

  for (Connection * conn = first_connection; conn; conn = conn->getNext()) {
    if (conn->queryPreferences() == preferences) {
      conn->getHistory()->setMaxSize(conn->queryPreferences()->getHistoryUseLimit() ? conn->queryPreferences()->getHistoryLimit() : 0);
      conn->getHistory()->setCircular(preferences->getHistoryIsCircular());
    }
  }

  if (GTK_TOGGLE_BUTTON(prefs_terminal_history_tabnull)->active)
    preferences->setHistoryHonourNull(1);
  else
    preferences->setHistoryHonourNull(0);

  if (GTK_TOGGLE_BUTTON(prefs_terminal_history_repeat)->active)
    preferences->setHistoryIgnoreRepeats(1);
  else
    preferences->setHistoryIgnoreRepeats(0);

  if (GTK_TOGGLE_BUTTON(prefs_terminal_use_maxlines)->active)
    preferences->setUseMaxLines(TRUE);
  else
    preferences->setUseMaxLines(FALSE);
 preferences->setMaxLines(gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(prefs_terminal_maxlines)));

 if (GTK_TOGGLE_BUTTON(prefs_terminal_prompt_terminator)->active)
   preferences->setUsePromptTerminator(TRUE);
 else
   preferences->setUsePromptTerminator(FALSE);

 const gchar * text = gtk_entry_get_text(GTK_ENTRY(prefs_terminal_prompt_terminator_string));
 preferences->setPromptTerminator((char *)text);

 text = gtk_entry_get_text(GTK_ENTRY(prefs_terminal_charset));
 preferences->setPreference("charset", (gchar *)text);

}

void get_connection_prefs(Prefs * preferences) {

  if (GTK_TOGGLE_BUTTON(prefs_connection_autoreconnect)->active)
    preferences->setAutoReconnect(1);
  else
    preferences->setAutoReconnect(0);

  if (GTK_TOGGLE_BUTTON(prefs_connection_compression)->active)
    preferences->setCompression(1);
  else
    preferences->setCompression(0);


  preferences->setSendStats(GTK_TOGGLE_BUTTON(prefs_stats_enable)->active);
  preferences->setSendVersion(GTK_TOGGLE_BUTTON(prefs_stats_enable_version)->active);

  preferences->setCompressionOffCommand((gchar *)gtk_entry_get_text(GTK_ENTRY(prefs_connection_compression_off_command)));

  if (GTK_TOGGLE_BUTTON(prefs_connection_mudftp)->active)
    preferences->setMudftp(1);
  else
    preferences->setMudftp(0);

  if (GTK_TOGGLE_BUTTON(prefs_connection_mcw)->active)
    preferences->setMCW(1);
  else
    preferences->setMCW(0);

  if (GTK_TOGGLE_BUTTON(prefs_connection_send_keepalive)->active)
    preferences->setSendKeepalive(TRUE);
  else
    preferences->setSendKeepalive(FALSE);

  preferences->setKeepaliveTimer(
                      gtk_spin_button_get_value_as_int(
                          GTK_SPIN_BUTTON(prefs_connection_keepalive)));

}

void get_gui_prefs(Prefs * preferences) {

  if (GTK_TOGGLE_BUTTON(prefs_gui_toolbar)->active)
    preferences->setUseToolbar(true);
  else
    preferences->setUseToolbar(false);

#if 0
  if (GTK_TOGGLE_BUTTON(prefs_gui_tooltips)->active)
    preferences->setUseTooltips(true);
  else
    preferences->setUseTooltips(false);
#endif

}

void get_proxy_prefs(Prefs * preferences) {

  // handle the toplevel radio buttons
  if (GTK_TOGGLE_BUTTON(prefs_proxy_direct_radio)->active) {
    preferences->setSocks(false);
    preferences->setTurfHttpd(false);
    preferences->setHttpConnect(false);
  } else if (GTK_TOGGLE_BUTTON(prefs_proxy_socks_radio)->active) {
    preferences->setSocks(true);
    preferences->setTurfHttpd(false);
    preferences->setHttpConnect(false);
  } else if (GTK_TOGGLE_BUTTON(prefs_proxy_httpd_radio)->active) {
    preferences->setSocks(false);
    preferences->setTurfHttpd(true);
    preferences->setHttpConnect(false);
  } else if (GTK_TOGGLE_BUTTON(prefs_proxy_httpd_connect_radio)->active) {
    preferences->setSocks(false);
    preferences->setTurfHttpd(false);
    preferences->setHttpConnect(true);
  } else {
    // Shouldn't need this, but one never knows ...
    preferences->setSocks(false);
    preferences->setTurfHttpd(false);
    preferences->setHttpConnect(false);
  }

  // handle socks 4/5 radios
  if (GTK_TOGGLE_BUTTON(prefs_proxy_socks4_radio)->active)
    preferences->setSocks4(1);
  else if (GTK_TOGGLE_BUTTON(prefs_proxy_socks5_radio)->active)
    preferences->setSocks4(0);
  else {
    // Shouldn't need this, but one never knows ...
    preferences->setSocks4(1);
  }

  // Read all the socks entry boxes
  preferences->setSocksHost(
                (gchar *)gtk_entry_get_text(GTK_ENTRY(prefs_proxy_socks_host)));
  preferences->setSocksPort(
                (gchar *)gtk_entry_get_text(GTK_ENTRY(prefs_proxy_socks_port)));
  preferences->setSocksUser(
                (gchar *)gtk_entry_get_text(GTK_ENTRY(prefs_proxy_socks_user)));
  preferences->setSocksPass(
                (gchar *)gtk_entry_get_text(GTK_ENTRY(prefs_proxy_socks_pass)));

  // Handle the HTTP proxy checkbox
  if (GTK_TOGGLE_BUTTON(prefs_proxy_http)->active)
    preferences->setHttpProxy(true);
  else
    preferences->setHttpProxy(false);
  
  // Read in from the HTTP entry boxes
  preferences->setHttpProxyPort(
                (gchar *)gtk_entry_get_text(GTK_ENTRY(prefs_proxy_http_port)));
  preferences->setHttpProxyHost(
                (gchar *)gtk_entry_get_text(GTK_ENTRY(prefs_proxy_http_host)));
}

void get_trigger_prefs(Prefs * preferences) {

  if (GTK_TOGGLE_BUTTON(prefs_triggers_enable)->active)
    preferences->setTriggers(1);
  else
    preferences->setTriggers(0);


#ifndef WIN32
  const gchar * text = gtk_entry_get_text(GTK_ENTRY(prefs_browser));
  preferences->setBrowser((gchar *)text);
#endif
}

void get_colour_prefs(Prefs * preferences) {
  // Is colour enabled
  if (GTK_TOGGLE_BUTTON(prefs_colour_enable)->active)
    preferences->setColour(1);
  else
    preferences->setColour(0);

  // Foreground colour
  GtkWidget * active_fg = gtk_menu_get_active(GTK_MENU(prefs_colour_deffg_menu));
  int fg_colour = g_list_index(GTK_MENU_SHELL(prefs_colour_deffg_menu)->children, active_fg);
  if (fg_colour == 8)
    fg_colour = -1;
  preferences->setDefaultForegroundColour(fg_colour);

  // Background colour
  GtkWidget * active_bg = gtk_menu_get_active(GTK_MENU(prefs_colour_defbg_menu));
  int bg_colour = g_list_index(GTK_MENU_SHELL(prefs_colour_defbg_menu)->children, active_bg);
  if (bg_colour == 8)
    bg_colour = -1;
  preferences->setDefaultBackgroundColour(bg_colour);

  // White (colour 7)
  preferences->setColours(7, white->getRed(), white->getGreen(),
                          white->getBlue());
  preferences->getSysColour()->setColor(7, white->getRed(), white->getGreen(),
                                        white->getBlue());

  // Red (colour 1)
  preferences->setColours(1, red->getRed(), red->getGreen(), red->getBlue());
  preferences->getSysColour()->setColor(1, red->getRed(), red->getGreen(),
                                        red->getBlue());
  
  // Green (colour 2)
  preferences->setColours(2, green->getRed(), green->getGreen(),
                          green->getBlue());
  preferences->getSysColour()->setColor(2, green->getRed(), green->getGreen(),
                                        green->getBlue());

  // Blue (colour 4)
  preferences->setColours(4, blue->getRed(), blue->getGreen(), blue->getBlue());
  preferences->getSysColour()->setColor(4, blue->getRed(), blue->getGreen(),
                                        blue->getBlue());

  // Yellow (colour 3)
  preferences->setColours(3, yellow->getRed(), yellow->getGreen(),
                          yellow->getBlue());
  preferences->getSysColour()->setColor(3, yellow->getRed(), yellow->getGreen(),
                                        yellow->getBlue());

  // Magenta (colour 5)
  preferences->setColours(5, magenta->getRed(), magenta->getGreen(),
                          magenta->getBlue());
  preferences->getSysColour()->setColor(5, magenta->getRed(),
                                        magenta->getGreen(),
                                        magenta->getBlue());
  // Cyan (colour 6)
  preferences->setColours(6, cyan->getRed(), cyan->getGreen(), cyan->getBlue());
  preferences->getSysColour()->setColor(6, cyan->getRed(), cyan->getGreen(),
                                        cyan->getBlue());

  // Black (colour 0)
  preferences->setColours(0, black->getRed(), black->getGreen(),
                          black->getBlue());
  preferences->getSysColour()->setColor(0, black->getRed(), black->getGreen(),
                                        black->getBlue());

  // Bold White (colour 15)
  preferences->setColours(15, white_bold->getRed(), white_bold->getGreen(),
                          white_bold->getBlue());
  preferences->getSysColour()->setColor(15, white_bold->getRed(),
                                        white_bold->getGreen(),
                                        white_bold->getBlue());


  // Bold Red (colour 9)
  preferences->setColours(9, red_bold->getRed(), red_bold->getGreen(),
                          red_bold->getBlue());
  preferences->getSysColour()->setColor(9, red_bold->getRed(),
                                        red_bold->getGreen(),
                                        red_bold->getBlue());

  // Bold Green (colour 10)
  preferences->setColours(10, green_bold->getRed(), green_bold->getGreen(),
                          green_bold->getBlue());
  preferences->getSysColour()->setColor(10, green_bold->getRed(),
                                        green_bold->getGreen(),
                                        green_bold->getBlue());

  // Bold Blue (colour 12)
  preferences->setColours(12, blue_bold->getRed(), blue_bold->getGreen(),
                          blue_bold->getBlue());
  preferences->getSysColour()->setColor(12, blue_bold->getRed(),
                                        blue_bold->getGreen(),
                                        blue_bold->getBlue());

  // Bold Yellow (colour 11)
  preferences->setColours(11, yellow_bold->getRed(), yellow_bold->getGreen(),
                          yellow_bold->getBlue());
  preferences->getSysColour()->setColor(11, yellow_bold->getRed(),
                                        yellow_bold->getGreen(),
                                        yellow_bold->getBlue());
  // Bold Magenta (colour 13)
  preferences->setColours(13, magenta_bold->getRed(), magenta_bold->getGreen(),
                          magenta_bold->getBlue());
  preferences->getSysColour()->setColor(13, magenta_bold->getRed(),
                                        magenta_bold->getGreen(),
                                        magenta_bold->getBlue());

  // Bold Cyan (colour 14)
  preferences->setColours(14, cyan_bold->getRed(), cyan_bold->getGreen(),
                          cyan_bold->getBlue());
  preferences->getSysColour()->setColor(14, cyan_bold->getRed(),
                                        cyan_bold->getGreen(),
                                        cyan_bold->getBlue());

  // Bold Black (colour 8)
  preferences->setColours(8, black_bold->getRed(), black_bold->getGreen(),
                          black_bold->getBlue());
  preferences->getSysColour()->setColor(8, black_bold->getRed(),
                                        black_bold->getGreen(),
                                        black_bold->getBlue());

  // Background
  preferences->setScreenBackground(background->getRed(),
                                   background->getGreen(), 
                                   background->getBlue());
}

void get_font_prefs(Prefs * preferences) {
  gchar * font = output_font_chooser->getFont();
  preferences->setOutputFont(font);
  free(font);

  font = bold_font_chooser->getFont();
  preferences->setBoldFont(font);
  free(font);

  font = input_font_chooser->getFont();
  preferences->setInputFont(font);
  free(font);
  
  font = prompt_font_chooser->getFont();
  preferences->setPromptFont(font);
  free(font);

  preferences->setUseBoldFont(GTK_TOGGLE_BUTTON(prefs_use_bold)->active);
}

void get_prefs(Prefs * preferences) {
  get_terminal_prefs(preferences);
  get_connection_prefs(preferences);
  get_proxy_prefs(preferences);
  get_trigger_prefs(preferences);
  get_colour_prefs(preferences);
  get_font_prefs(preferences);
  get_gui_prefs(preferences);

  int width, height;

  mainWindow->getDimensions(&width, &height);
  preferences->setWindowX(width);
  preferences->setWindowY(height);

}
