#include <time.h>
#include <string.h>
#include <stdio.h>

#include "mudclient.h"
#include "PluginHandler.h"
#include "ConnectionTimer.h"
#include "version.h"

#include "Win32PluginAPI.cpp"

extern PluginHandler * phandler;

static ConnectionTimer * ctimer;

#define MAJOR "1"
#define MINOR "1"

extern "C" G_MODULE_EXPORT char * plugin_query_name() {
  return "ConnectionTimer";
}

extern "C" G_MODULE_EXPORT char * plugin_query_description() {
  return "Displays how long this connection has been active in the plugin tray.";
}

extern "C" G_MODULE_EXPORT char * plugin_query_major() {
  return MAJOR;
}

extern "C" G_MODULE_EXPORT char * plugin_query_minor() {
  return MINOR;
}

extern "C" G_MODULE_EXPORT void plugin_init(plugin_address_table_t * pat) {
    plugin_address_table_init(pat);
	ctimer = new ConnectionTimer();
}

extern "C" G_MODULE_EXPORT void plugin_cleanup(void) {
  delete ctimer;
}

ConnectionTimer::ConnectionTimer() {
  cList = new List();
  version = 1.0;
  name = strdup("Connection Timer");

  register_plugin(this, VERSION);
}

ConnectionTimer::~ConnectionTimer() {
  free(name);
  name = NULL;

  // Go through each entry, removing the timer widget from the tray.
  PapayaListElement * ele_next;
  for (PapayaListElement * ele = cList->firstElement; ele; ele = ele_next) {
    ele_next = ele->getNext();
    struct ctdata * d = (struct ctdata *)ele->getData();

	vt_remove_from_tray(connection_get_vt(ele->getConnection()), d->event_box, d->frame);
//    ele->getConnection()->getVT()->removeFromTray(d->event_box, d->frame);
    gtk_timeout_remove(d->timeout);

    cList->deleteEntry(ele);
  }

  unregister_plugin(this);

  delete cList;
  cList = NULL;
}

gint conn_timer_timeout(gpointer data) {
  char buf[1024];
  ListElement * element = (ListElement *)data;
  struct ctdata * ct = (struct ctdata *)element->getData();

  time_t currtime;
  time(&currtime);

  int seconds = currtime - ct->start;
  int minutes = seconds / 60;
  int hours = minutes / 60;
  minutes %= 60;
  seconds %= 60;

  sprintf(buf, "%.2d:%.2d:%.2ds", hours, minutes, seconds);
  gtk_label_set_text(GTK_LABEL(ct->widget), buf);
  return 1;
}

void ConnectionTimer::onEvent(Event * e, Connection * c) {

  if (event_get_type(e) == EvConnect) {
    ListElement * tmp = cList->findEntry(c);
    if (!tmp)
      tmp = cList->newEntry(c, NULL);

    struct ctdata * data = (struct ctdata *)malloc(sizeof(struct ctdata));
    
    // Configure the connection time.
    time(&data->start);
    // Setup the timeout
    data->timeout = gtk_timeout_add(1000, conn_timer_timeout, (gpointer)tmp);

    data->event_box = gtk_event_box_new();

    // Configure the output Widget.
    data->widget = gtk_label_new("NOT CONNECTED");
    gtk_widget_show(data->widget);
    gtk_container_add(GTK_CONTAINER(data->event_box), data->widget);

    data->tooltips = gtk_tooltips_new();
    gtk_tooltips_set_tip(data->tooltips, data->event_box, "ConnectionTimer: displays the length of time connected to a MUD.", NULL);


	vt_add_to_tray(connection_get_vt(c), data->event_box, &data->frame);
//    c->getVT()->addToTray(data->event_box, &data->frame);

    tmp->setData((void *)data);
    return;
  }

  if (event_get_type(e) == EvDisconnect) {
    ListElement * tmp = cList->findEntry(c);
    if (!tmp) {
      printf("ConnectionTimer:  Can't find connection.\n");
      return;
    }

    // @@ Make removeFromTray delete the widget in future
    struct ctdata * d = (struct ctdata *)tmp->getData();

    vt_remove_from_tray(connection_get_vt(c), d->event_box, d->frame);
//    c->getVT()->removeFromTray(d->event_box, d->frame);
    gtk_timeout_remove(d->timeout);
    g_free(d->tooltips);
    // This takes care of deleting 'data'
    cList->deleteEntry(c);
    return;
  }

}

char * ConnectionTimer::getDescription() {
  return "Displays the length of time a connection has been active.";
};
