#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#ifdef WIN32
#include <direct.h>
#else
#include <unistd.h>
#endif

#include "mudclient.h"
#include "PluginHandler.h"
#include "PapayaList.h"
#include "version.h"

#include "MudLog.h"

#include "Win32PluginAPI.cpp"

static MudLog * ml;

#define MAJOR "1"
#define MINOR "1"

extern "C" G_MODULE_EXPORT char * plugin_query_name() {
  return "MudLog";
}

extern "C" G_MODULE_EXPORT char * plugin_query_description() {
  return "Logs data from connections to separate files.";
}

extern "C" G_MODULE_EXPORT char * plugin_query_major() {
  return MAJOR;
}

extern "C" G_MODULE_EXPORT char * plugin_query_minor() {
  return MINOR;
}

extern "C" G_MODULE_EXPORT void plugin_init(plugin_address_table_t * pat) {
  plugin_address_table_init(pat);
  ml = new MudLog();
}

extern "C" G_MODULE_EXPORT void plugin_cleanup(void) {
  delete ml;
}

MudLog::MudLog() {
  version = 1.0;
  name = strdup("MudLog");
  list = new List();

  char path[1024];

#ifdef WIN32
  snprintf(path, 1024, "%s/logs", get_prefix());
#else
  char * home = getenv("HOME");
  if (!home)
    snprintf(path, 1024, "logs");
  else
    snprintf(path, 1024, "%s/.papaya/logs", home);
#endif

#ifdef WIN32
  if (mkdir(path) == - 1) {
#else
  if (mkdir(path, 0700) == -1) {
#endif
	if (errno != EEXIST)
    perror("mkdir");
  }

  register_plugin(this, VERSION);
}

MudLog::~MudLog() {

  for (ListElement * ele = list->firstElement; ele; ele = ele->getNext()) {
    fclose((FILE *)ele->getData());
    ele->setData(NULL);
  }

  delete list;

  unregister_plugin(this);
}

void MudLog::input(Connection * c, char * in) {
  write(c, in, 1);
}


void MudLog::output(Connection * c, char * in) {
  write(c, in, 0);
}

void MudLog::write(Connection * c, char * in, bool myInput) {

  ListElement * tmp = list->findEntry(c);
  if (!tmp) {
    tmp = createLogfile(c);
  }

  FILE * fp = (FILE *)tmp->getData();
  if (!fp) {
    return;
  }

  if (myInput)
    fprintf(fp, "INPUT: %s\n", in);
  else
    fprintf(fp, "OUTPUT: %s\n", in);

  fflush(fp);

  return;
}

ListElement * MudLog::createLogfile(Connection * c) {

  char buf[2048], time_buf[1024];
  char * home = getenv("HOME");
  time_t t;
  struct tm * tm_t;

  // Create the folder for the connection.  We always try to do this.
#ifdef WIN32
  snprintf(buf, 2048, "%s/logs/%s", get_prefix(), connection_get_name(c));
#else
  snprintf(buf, 2048, "%s/.papaya/logs/%s", home, connection_get_name(c));
#endif

#ifdef WIN32
  if (mkdir(buf) == - 1) {
#else
  if (mkdir(buf, 0700) == -1) {
#endif
	if (errno != EEXIST)
      perror("mkdir");
  }


  // Figure out the timestamp for the log file.
  time(&t);
  tm_t = gmtime(&t);
  strftime(time_buf, 1024, "%d-%b-%Y-%H.%M.%S", tm_t);

#ifdef WIN32
  snprintf(buf, 2048, "%s\\logs\\%s\\%s", get_prefix(), connection_get_name(c), time_buf);
#else
  snprintf(buf, 2048, "%s/.papaya/logs/%s/%s", home, connection_get_name(c), time_buf);
#endif

  // Open the file in append mode.
  FILE * fp = fopen(buf, "a");
  if (!fp) {
    perror("fopen");
    return NULL;
  }
  
  ListElement * ele = list->newEntry(c, NULL);
  ele->setData(fp);
  return ele;
}

void MudLog::closeLogfile(ListElement * e) {
  FILE * fp = (FILE *)e->getData();
  fclose(fp);
  
  e->setData(NULL);

  list->deleteEntry(e);
}

void MudLog::onEvent(Event * e, Connection * c) {
  
  if (event_get_type(e) == EvConnect) {
    createLogfile(c);
    return;
  }

  if (event_get_type(e) == EvDisconnect) {
    ListElement * e = list->findEntry(c);
    if (!e)
      return;

    closeLogfile(e);
    return;
  }

}
