//#include <sys/time.h>
#include "mudclient.h"

#include <stdlib.h>
#include <regex.h>
#include <stdio.h>
#include <string.h>

#include "version.h"
#include "Message.h"
#include "Win32PluginAPI.cpp"
#include "Spells.h"

static Spells * spells = NULL;
static regex_t spell_regex;
static List * sList = NULL;

#define MAJOR "0"
#define MINOR "3"

#define SPELL_REGEX "spell_string"
#define SPELL_FILE "spell_file"
#define SPELL_TEXT "spell_text"
#define SPELL_ACTIVE "spell_active"

#define DEFAULT_SPELL_FILE "spells/default"
#define DEFAULT_SPELL_REGEX "Spell: '(.*)' .*"

extern "C" G_MODULE_EXPORT char * plugin_query_name()
{
  return "Spells";
}

extern "C" G_MODULE_EXPORT char * plugin_query_description()
{
  return "Informs the user which spells their character is missing.";
}

extern "C" G_MODULE_EXPORT char * plugin_query_major()
{
  return MAJOR;
}

extern "C" G_MODULE_EXPORT char * plugin_query_minor()
{
  return MINOR;
}

extern "C" G_MODULE_EXPORT void plugin_init(plugin_address_table_t * pat)
{
	plugin_address_table_init(pat);
	spells = new Spells();
}

extern "C" G_MODULE_EXPORT void plugin_cleanup(void)
{
  if (spells)
    delete spells;
}

Spells::Spells()
{
  name = strdup("Spells");
  sList = new List();

  register_plugin(this, VERSION);
}

Spells::~Spells()
{
  unregister_plugin(this);
  delete sList;
  sList = NULL;
}

void spells_callback(Connection * c, char * buf, void * d)
{
  int nmatch = 2;
  regmatch_t match[2];
  char spell[128];
  int start, end;
  
  ListElement * tmp = sList->findEntry(c);
  if (!tmp || !tmp->getData())
    return;

  struct spell_data * data = (struct spell_data *)tmp->getData();

  if (!buf) {
    /* Print a list of the missing spells for this connection. */
    for (int i = 0; i < MAX_SPELLS; i++) {

      if (data->spell[i] && data->seen[i] == 0) {
		char buf[1024];
		sprintf(buf, "\033[1;34mMissing: %s\033[0m\n", data->spell[i]);
		vt_append(connection_get_vt(c), buf);
      }

      // Free the data.
      if (data->spell[i]) {
		free(data->spell[i]);
		data->spell[i] = NULL;
      }

    }

	vt_scroll(connection_get_vt(c));
	if (d)
		free(d);
    
	return;
  }


  char * spell_string = preferences_get_preference(connection_query_preferences(c), SPELL_REGEX);
  if (spell_string)
    regcomp(&spell_regex, spell_string, REG_ICASE|REG_EXTENDED);
  else
    regcomp(&spell_regex, DEFAULT_SPELL_REGEX, REG_ICASE|REG_EXTENDED);

  if (regexec(&spell_regex, buf, nmatch, match, 0) == REG_NOMATCH) {
    regfree(&spell_regex);
    return;
  }

  start = match[1].rm_so;
  end = match[1].rm_eo;

  if (end-start > 127) {
    regfree(&spell_regex);
    return;
  }

  memcpy(spell, buf + start, end - start);
  spell[end-start] = '\0';

  /* spell contains the string name of the spell we found */

  for (int i = 0; i < MAX_SPELLS; i++) {
    if (data->spell[i] && !strncasecmp(data->spell[i], spell, strlen(spell))) {
      data->seen[i] = 1;
    }
  }

  regfree(&spell_regex);

}

void Spells::input(Connection * c, char * buf)
{
  char fname[1024];
  FILE * fp;

  if (strncasecmp(buf, "missing", 7))
    return;

  if (!turf_protocol_is_supported(c))
	  return;

  // Do we have a spell_data for this connection?

  ListElement * tmp = sList->findEntry(c);
  if (!tmp)
    tmp = create_new_entry(c);

  struct spell_data * data = (struct spell_data *)tmp->getData();

  gchar * text = preferences_get_preference(connection_query_preferences(c), SPELL_ACTIVE);

  if (text && !strcmp(text, "1")) {
    // We're using the text box's data.
    gchar * tmp = strdup(preferences_get_preference(connection_query_preferences(c), SPELL_TEXT));
    getSpellData(tmp, data);
    free(tmp);
  } else {
    // We're loading from a file.

    char * filename = preferences_get_preference(connection_query_preferences(c), SPELL_FILE);
    if (filename)
      sprintf(fname, "%s", filename);
    else
      sprintf(fname, "%s", DEFAULT_SPELL_FILE);
    
    fp = openFile(fname, NULL, "r");
    if (!fp) {
      char buf[1024];
      sprintf(buf, "Could not open missing spells file: %s\r\n", fname);
	  vt_append(connection_get_vt(c), buf);
      return;
    }
    
    getSpellData(fp, data);
    
  }

  buf[0] = '\0';

  turf_protocol_add_command(c, "affects", spells_callback, NULL);
  return;
}

static GtkWidget * spells_widget = NULL;
static GtkWidget * spells_regex = NULL;
#ifdef USE_GTK_2
static GtkTextBuffer * spells_buffer = NULL;
static GtkWidget * spells_view = NULL;
#else
static GtkWidget * spells_text = NULL;
#endif
static GtkWidget * spells_radio_text = NULL;
static GtkWidget * spells_radio_file = NULL;

GtkWidget * Spells::getPrefsWidget(MUD * mud) {
  GtkWidget *vbox1;
  GtkWidget *frame4;
  GtkWidget *hbox1;
  GtkWidget *label1;
  GtkWidget *frame5;
  GtkWidget *vbox4;
  GtkWidget *vbox5;
  GSList *SpellsRadio_group = NULL;
  GtkWidget *scrolledwindow2;
  GtkWidget *vbox6;

  spells_widget = gtk_frame_new ("Missing Spells Configuration");
  gtk_widget_ref (spells_widget);
  gtk_widget_show (spells_widget);
  gtk_container_set_border_width (GTK_CONTAINER (spells_widget), 2);

  vbox1 = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (vbox1);
  gtk_widget_show (vbox1);
  gtk_container_add (GTK_CONTAINER (spells_widget), vbox1);

#ifdef USE_GTK_2
  g_object_set_data_full(G_OBJECT(spells_widget), "vbox1", vbox1,
			 (GtkDestroyNotify) gtk_widget_unref);
#else
  gtk_object_set_data_full(GTK_OBJECT(spells_widget), "vbox1", vbox1,
			   (GtkDestroyNotify) gtk_widget_unref);
#endif

  frame4 = gtk_frame_new ("General");
  gtk_widget_ref (frame4);
  gtk_widget_show (frame4);
  gtk_box_pack_start (GTK_BOX (vbox1), frame4, FALSE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (frame4), 2);

#ifdef USE_GTK_2
  g_object_set_data_full(G_OBJECT(spells_widget), "frame4", frame4,
			 (GtkDestroyNotify) gtk_widget_unref);
#else
  gtk_object_set_data_full(GTK_OBJECT(spells_widget), "frame4", frame4,
			   (GtkDestroyNotify) gtk_widget_unref);
#endif

  GtkWidget * tmp_vbox = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(tmp_vbox);
  gtk_container_add(GTK_CONTAINER(frame4), tmp_vbox);
  gtk_container_set_border_width(GTK_CONTAINER(tmp_vbox), 2);
#ifdef USE_GTK_2
  g_object_set_data_full(G_OBJECT(spells_widget), "tmp_vbox", tmp_vbox,
			 (GtkDestroyNotify) gtk_widget_unref);
#else
  gtk_object_set_data_full(GTK_OBJECT(spells_widget), "tmp_vbox", tmp_vbox,
			   (GtkDestroyNotify) gtk_widget_unref);
#endif

  hbox1 = gtk_hbox_new (FALSE, 0);
  gtk_widget_ref (hbox1);
  gtk_widget_show (hbox1);
  gtk_container_add (GTK_CONTAINER (tmp_vbox), hbox1);
#ifdef USE_GTK_2
  g_object_set_data_full(G_OBJECT(spells_widget), "hbox1", hbox1,
			 (GtkDestroyNotify) gtk_widget_unref);
#else
  gtk_object_set_data_full(GTK_OBJECT(spells_widget), "hbox1", hbox1,
			   (GtkDestroyNotify) gtk_widget_unref);
#endif

  label1 = gtk_label_new ("Spell text pattern (regex): ");
  gtk_widget_ref (label1);
  gtk_widget_show (label1);
  gtk_box_pack_start (GTK_BOX (hbox1), label1, FALSE, FALSE, 0);
#ifdef USE_GTK_2
  g_object_set_data_full(G_OBJECT(spells_widget), "label1", label1,
			 (GtkDestroyNotify) gtk_widget_unref);
#else
  gtk_object_set_data_full(GTK_OBJECT(spells_widget), "label1", label1,
			   (GtkDestroyNotify) gtk_widget_unref);
#endif

  spells_regex = gtk_entry_new ();
  gtk_widget_ref (spells_regex);
  gtk_widget_show (spells_regex);
  gtk_box_pack_start (GTK_BOX (hbox1), spells_regex, TRUE, TRUE, 0);
#ifdef USE_GTK_2
  g_object_set_data_full(G_OBJECT(spells_widget), "spells_regex", spells_regex,
			 (GtkDestroyNotify) gtk_widget_unref);
#else
  gtk_object_set_data_full(GTK_OBJECT(spells_widget), "spells_regex", spells_regex,
			   (GtkDestroyNotify) gtk_widget_unref);
#endif

  frame5 = gtk_frame_new ("Spell Data");
  gtk_widget_ref (frame5);
  gtk_widget_show (frame5);
  gtk_box_pack_start (GTK_BOX (vbox1), frame5, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (frame5), 2);
#ifdef USE_GTK_2
  g_object_set_data_full(G_OBJECT(spells_widget), "frame5", frame5,
			 (GtkDestroyNotify) gtk_widget_unref);
#else
  gtk_object_set_data_full(GTK_OBJECT(spells_widget), "frame5", frame5,
			   (GtkDestroyNotify) gtk_widget_unref);
#endif

  vbox4 = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (vbox4);
  gtk_widget_show (vbox4);
  gtk_container_add (GTK_CONTAINER (frame5), vbox4);
#ifdef USE_GTK_2
  g_object_set_data_full(G_OBJECT(spells_widget), "vbox4", vbox4,
			 (GtkDestroyNotify) gtk_widget_unref);
#else
  gtk_object_set_data_full(GTK_OBJECT(spells_widget), "vbox4", vbox4,
			   (GtkDestroyNotify) gtk_widget_unref);
#endif

  vbox5 = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (vbox5);
  gtk_widget_show (vbox5);
  gtk_box_pack_start (GTK_BOX (vbox4), vbox5, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox5), 2);
#ifdef USE_GTK_2
  g_object_set_data_full(G_OBJECT(spells_widget), "vbox5", vbox5,
			 (GtkDestroyNotify) gtk_widget_unref);
#else
  gtk_object_set_data_full(GTK_OBJECT(spells_widget), "vbox5", vbox5,
			   (GtkDestroyNotify) gtk_widget_unref);
#endif

  spells_radio_text = gtk_radio_button_new_with_label (SpellsRadio_group, "Input spells in text box (one line at a time)");
#ifdef USE_GTK_2
  SpellsRadio_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (spells_radio_text));
#else
  SpellsRadio_group = gtk_radio_button_group (GTK_RADIO_BUTTON (spells_radio_text));
#endif
  gtk_widget_ref (spells_radio_text);
  gtk_widget_show (spells_radio_text);
  gtk_box_pack_start (GTK_BOX (vbox5), spells_radio_text, FALSE, FALSE, 0);
#ifdef USE_GTK_2
  g_object_set_data_full(G_OBJECT(spells_widget), "spells_radio_text", spells_radio_text,
			 (GtkDestroyNotify) gtk_widget_unref);
#else
  gtk_object_set_data_full(GTK_OBJECT(spells_widget), "spells_radio_text", spells_radio_text,
			   (GtkDestroyNotify) gtk_widget_unref);
#endif

  scrolledwindow2 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_ref (scrolledwindow2);
  gtk_widget_show (scrolledwindow2);
  gtk_box_pack_start (GTK_BOX (vbox5), scrolledwindow2, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow2), GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);
#ifdef USE_GTK_2
  g_object_set_data_full(G_OBJECT(spells_widget), "scrolledwindow2", scrolledwindow2,
			 (GtkDestroyNotify) gtk_widget_unref);
#else
  gtk_object_set_data_full(GTK_OBJECT(spells_widget), "scrolledwindow2", scrolledwindow2,
			   (GtkDestroyNotify) gtk_widget_unref);
#endif


#ifdef USE_GTK_2
  spells_buffer = gtk_text_buffer_new(NULL);
  spells_view = gtk_text_view_new_with_buffer(spells_buffer);

  gtk_widget_show(spells_view);
  gtk_container_add(GTK_CONTAINER(scrolledwindow2), spells_view);
  g_object_set_data_full(G_OBJECT(spells_widget), "spells_view", spells_view,
			 (GtkDestroyNotify) gtk_widget_unref);
#else
  spells_text = gtk_text_new (NULL, NULL);
  gtk_widget_ref (spells_text);
  gtk_widget_show (spells_text);
  gtk_container_add (GTK_CONTAINER (scrolledwindow2), spells_text);
  gtk_text_set_editable(GTK_TEXT(spells_text), true);
  gtk_object_set_data_full(GTK_OBJECT(spells_widget), "spells_text", spells_text,
			   (GtkDestroyNotify) gtk_widget_unref);
#endif

  vbox6 = gtk_vbox_new (FALSE, 0);
  gtk_widget_ref (vbox6);
  gtk_widget_show (vbox6);
  gtk_box_pack_start (GTK_BOX (vbox4), vbox6, FALSE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (vbox6), 2);
#ifdef USE_GTK_2
  g_object_set_data_full(G_OBJECT(spells_widget), "vbox6", vbox6,
			 (GtkDestroyNotify) gtk_widget_unref);
#else
  gtk_object_set_data_full(GTK_OBJECT(spells_widget), "vbox6", vbox6,
			   (GtkDestroyNotify) gtk_widget_unref);
#endif

  spells_radio_file = gtk_radio_button_new_with_label (SpellsRadio_group, "Read spells from file");
#ifdef USE_GTK_2
  SpellsRadio_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (spells_radio_file));
#else
  SpellsRadio_group = gtk_radio_button_group (GTK_RADIO_BUTTON (spells_radio_file));
#endif
  gtk_widget_ref (spells_radio_file);
  gtk_widget_show (spells_radio_file);
  gtk_box_pack_start (GTK_BOX (vbox6), spells_radio_file, FALSE, FALSE, 0);
#ifdef USE_GTK_2
  g_object_set_data_full(G_OBJECT(spells_widget), "spells_radio_file", spells_radio_file,
			 (GtkDestroyNotify) gtk_widget_unref);
#else
  gtk_object_set_data_full(GTK_OBJECT(spells_widget), "spells_radio_file", spells_radio_file,
			   (GtkDestroyNotify) gtk_widget_unref);
#endif

  // @@ Create a file entry widget so the user can select a file.

  if (mud && mud_get_preferences(mud)) {
	gchar * text = preferences_get_preference(mud_get_preferences(mud), SPELL_REGEX);
	if (text) {
      gtk_entry_set_text(GTK_ENTRY(spells_regex), text);
    }
    else
      gtk_entry_set_text(GTK_ENTRY(spells_regex), DEFAULT_SPELL_REGEX);
    
	text = preferences_get_preference(mud_get_preferences(mud), DEFAULT_SPELL_REGEX);
    if (text) {
      char * last = text;
      char * pc;
      while ((pc = strchr(last, '`'))) {
		*pc = '\n';
		last = pc + 1;
      }

#ifdef USE_GTK_2
      GtkTextIter start, end;
      gtk_text_buffer_get_bounds(spells_buffer, &start, &end);
      gtk_text_buffer_insert(spells_buffer, &end, text, -1);
#else
      gtk_text_insert(GTK_TEXT(spells_text), NULL, NULL, NULL, text, -1);
#endif
    }

	text = preferences_get_preference(mud_get_preferences(mud), SPELL_FILE);
    // @@ Set the default path into the file entry.

	text = preferences_get_preference(mud_get_preferences(mud), SPELL_ACTIVE);
    if (text && !strcmp(text, "1"))
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(spells_radio_text), TRUE);
    else
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(spells_radio_file), TRUE);

  } else {
	gchar * text = preferences_get_preference(get_global_preferences(), SPELL_REGEX);
    if (text) {
      gtk_entry_set_text(GTK_ENTRY(spells_regex), text);
    }
    else
      gtk_entry_set_text(GTK_ENTRY(spells_regex), DEFAULT_SPELL_REGEX);

	text = preferences_get_preference(get_global_preferences(), SPELL_TEXT);
    if (text) {      
      char * last = text;
      char * pc;
      while ((pc = strchr(last, '`'))) {
		*pc = '\n';
		last = pc + 1;
      }
#ifdef USE_GTK_2
      GtkTextIter start, end;
      gtk_text_buffer_get_bounds(spells_buffer, &start, &end);
      gtk_text_buffer_insert(spells_buffer, &end, text, -1);
#else
      gtk_text_insert(GTK_TEXT(spells_text), NULL, NULL, NULL, text, -1);
#endif
    }

	text = preferences_get_preference(get_global_preferences(), SPELL_ACTIVE);
    if (text && !strcmp(text, "1"))
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(spells_radio_text), TRUE);
    else
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(spells_radio_file), TRUE);
  }

  return spells_widget;
}

ListElement * Spells::create_new_entry(Connection * conn) {
  struct spell_data * data = (struct spell_data *)malloc(sizeof(struct spell_data));
  memset(data, 0, sizeof(struct spell_data));

  ListElement * ele = sList->newEntry(conn, NULL);
  ele->setData(data);
  return ele;
}

void Spells::getSpellData(char * buf, struct spell_data * data) {

  char * pc;
  char * last = buf;
  int current = 0;

  // Remove the entries already here.
  for (int i = 0; i < MAX_SPELLS; i++) {
    if (data->spell[i])
      free(data->spell[i]);
    data->spell[i] = NULL;
  }

  while ((pc = strchr(last, '`'))) {
    *pc = '\0';

    data->spell[current] = strdup(last);
    data->seen[current] = 0;
    current++;

    if (current >= MAX_SPELLS)
      return;

    last = pc + 1;
  }

  if (strlen(last) > 0) {
    data->spell[current] = strdup(last);
    data->seen[current] = 0;
    current++;
  }

  return;
}

void Spells::getSpellData(FILE * fp, struct spell_data * data) {
  char line[1024];
  int current = 0;

  // Remove the entries already here.
  for (int i = 0; i < MAX_SPELLS; i++) {
    if (data->spell[i])
      free(data->spell[i]);
    data->spell[i] = NULL;
  }

  // Read spells in.
  while (fgets(line, 1023, fp)) {
    data->spell[current] = strdup(line);
    data->seen[current] = 0;
    current++;

    if (current >= MAX_SPELLS)
      return;
  }

  return;
}

void Spells::onPrefsCancel(MUD * mud) {
  gtk_widget_hide(spells_widget);
  gtk_widget_destroy(spells_widget);
}

void Spells::onPrefsApply(MUD * mud) {
  if (!spells_regex)
    return;

  gchar * text = (gchar *)gtk_entry_get_text(GTK_ENTRY(spells_regex));
  if (!mud || !mud_get_preferences(mud))
	  preferences_set_preference(get_global_preferences(), SPELL_REGEX, text);
  else
	  preferences_set_preference(mud_get_preferences(mud), SPELL_REGEX, text);
  
  // @@ Figure out which radio button is selected.

  // @@ Get the settings from the gnome file entry.

#ifdef USE_GTK_2
  GtkTextIter start, end;
  gtk_text_buffer_get_bounds(GTK_TEXT_BUFFER(spells_buffer), &start, &end);
  text = (gchar *)gtk_text_buffer_get_text(spells_buffer, &start, &end, false);
#else
  text = (gchar *)gtk_editable_get_chars(GTK_EDITABLE(spells_text), 0, -1);
#endif
  char * last = text;
  char * pc;
  
  while ((pc = strchr(last, '\n'))) {
    *pc = '`';
    last = pc + 1;
  }
  if (!mud || !mud_get_preferences(mud))
	  preferences_set_preference(get_global_preferences(), SPELL_TEXT, text); 
  else
	  preferences_set_preference(mud_get_preferences(mud), SPELL_TEXT, text);

#ifdef USE_GTK_2
  g_free(text);
#endif
 
  if (GTK_TOGGLE_BUTTON(spells_radio_text)->active) {
    if (!mud || !mud_get_preferences(mud)) {
		preferences_set_preference(get_global_preferences(), SPELL_ACTIVE, "1");
    }
    else {
	  preferences_set_preference(mud_get_preferences(mud), SPELL_ACTIVE, "1");
    }
  } else {
    if (!mud || !mud_get_preferences(mud)) {
		preferences_set_preference(get_global_preferences(), SPELL_ACTIVE, "0");
    }
    else {
		preferences_set_preference(mud_get_preferences(mud), SPELL_ACTIVE, "0");
    }
  }


}

void Spells::onPrefsOk(MUD * mud) {
  onPrefsApply(mud);
  onPrefsCancel(mud);
}
